
"""Anagram game handler."""

from time import time
import random

from zope.interface import implements

from storm import locals as storm, uri

from mzbot import commands, ibot, ebot


class Word(object):

    """Single word for the anagram game."""

    __storm_table__ = 'anagram_word'

    id = storm.Int(primary=True)
    word = storm.Unicode()

    def __init__(self, word):
        self.word = word


class AnagramHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """Do anagram puzzle quiz thingies."""

    implements(ibot.IMessageHandler)

    def __init__(self, **kwargs):
        commands.CommandHandler.__init__(self)
        self.question = None
        self.hints = None
        self.solution = None
        self.playing = False
        self.set_path(u'None')
        self.channels = ''

    # XXX logic copied from corehandlers.ChanServOpHandler
    def get_channels(self):
        """List of auto-opped channels."""
        return u', '.join(channel.encode('string-escape').decode('ascii')
                          for channel in self.channels.split())

    def set_addchannel(self, args):
        """Add a channel to the autoop list."""
        channel, = commands.parseArgs(args, commands.bytestring)
        self.channels += ' ' + channel

    def set_removechannel(self, args):
        """Remove a channel from the autoop list."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        if channel not in channels:
            raise ebot.InvalidArgument('No such channel')
        channels.remove(channel)
        self.channels = ' '.join(channels)

    def set_path(self, args):
        path, = commands.parseArgs(args, commands.bytestring)
        if path == 'None':
            dburi = uri.URI('sqlite:?timeout=0')
        else:
            dburi = uri.URI('sqlite:%s?timeout=0' % (uri.escape(path),))

        db = storm.Store(storm.create_database(dburi))

        db.execute('CREATE TABLE IF NOT EXISTS [anagram_word] '
                   '([id] INTEGER PRIMARY KEY, [word] TEXT NOT NULL)')
        db.execute('CREATE UNIQUE INDEX IF NOT EXISTS [anagram_word_index] '
                   'ON [anagram_word] ([word])')

        self.store = db

    @commands.args()
    def handle_clear(self, user, source):
        self.store.find(Word).remove()
        self.store.flush()

    @commands.args(commands.bytestring)
    def handle_load(self, user, source, path):
        with open(path, 'rb') as f:
            try:
                for word in f:
                    try:
                        word = unicode(word, 'utf-8').strip()
                    except UnicodeDecodeError, e:
                        raise ebot.InvalidArgument(e)
                    else:
                        if word:
                            self.store.add(Word(word))
            except:
                self.store.rollback()
                raise
            else:
                self.store.flush()

    def message(self, user, source, message):
        """Take a single word as a guess."""
        # ignore stuff not on the game channel during a game
        if (not self.playing or
            user is source or
            source.name not in self.channels.split()):
            return
        # ignore ignored users
        for ignorer in user.bot.handlers.get(ibot.IIgnore, ()):
            if ignorer.ignored(user, source):
                return
        try:
            message = message.strip().decode('utf-8')
        except UnicodeDecodeError:
            return
        # ignore words that are not permutations of the solution
        if sorted(message) != sorted(self.solution):
            return
        # pass to guess
        self.handle_guess(user, source, message)

    @commands.args()
    def handle_anagram(self, user, source):
        """Start a new anagram."""
        if self.playing:
            source.msgTo(
                user, u"In progress: '%s' (remaining letters: '%s'). "
                u"If you're stuck, try 'hint' or 'surrender'." %
                (commands.bold(self.hints), commands.bold(self.question)))
            return
        self.playing = True
        self.start = time()
        self.tries = 0
        while True:
            word = self.store.find(Word).order_by('RANDOM()').first()
            if word is None:
                source.msgTo(user, u'no words')
                return
            self.solution = word.word
            self.hints = u'.' * len(self.solution)
            question = list(self.solution)
            random.shuffle(question)
            self.question = u''.join(question)
            if self.store.find(Word,
                               Word.word == self.question).one() is None:
                break
        source.msgTo(user, u"Unshuffle this for me please: '%s'" % (
                commands.bold(self.question),))

    @commands.args(commands.string)
    def handle_guess(self, user, source, guess):
        """Take a guess."""
        if not self.playing:
            source.msgTo(user, u'start a new anagram game first!')
            return
        self.tries += 1
        if guess == self.solution:
            source.msgTo(
                user, u'Correct! It took you %s tries and %s seconds' % (
                    commands.bold(u'%d' % (self.tries,)),
                    commands.bold(u'%d' % (int(time() - self.start),))))
            self.playing = False
        elif sorted(guess) == sorted(self.solution) and self.store.find(
            Word, Word.word == guess).one() is not None:
            source.msgTo(
                user, u"That'll do, but what I had in mind was '%s'" % (
                    commands.bold(self.solution),))
            self.playing = False
        else:
            source.msgTo(user, u'Wrong, try again!')

    @commands.args()
    def handle_hint(self, user, source):
        """Give the position of a random letter."""
        if not self.playing:
            source.msgTo(user, u'start a new anagram game first!')
            return
        # Pick a random unplaced character.
        questionIndex = random.randrange(len(self.question))
        questionChar = self.question[questionIndex]
        # Pick a random unknown spot where it fits
        hint = random.choice(list(
                i for i, char in enumerate(self.solution)
                if char == questionChar and self.hints[i] == u'.'))
        # Move the char from question to hints.
        self.hints = self.hints[:hint] + questionChar + self.hints[hint+1:]
        self.question = self.question[:questionIndex] + self.question[
            questionIndex+1:]
        if len(self.question) == 1:
            source.msgTo(user, u"The solution is '%s'" % (
                    commands.bold(self.solution),))
            self.playing = False
        else:
            source.msgTo(
                user, u"Hint: '%s' (remaining letters: '%s')" %
                (self.hints, self.question))

    @commands.args()
    def handle_surrender(self, user, source):
        """Give up."""
        if not self.playing:
            source.msgTo(user, u'start a new anagram game first!')
            return
        source.msgTo(user, u"The solution is '%s'" % (
                    commands.bold(self.solution),))
        self.playing = False
