
"""A bunch of stuff to help keep track of bans."""


import re
import fnmatch
import datetime

from zope.interface import implements

from twisted.python import log

from storm import locals as storm, uri

from mzbot import ibot, commands


def parseBan(mask):
    """Return a tuple (quiet, mask, forward)."""
    quiet = mask.startswith('%')
    if quiet:
        mask = mask[1:]
    index = mask.find('!', mask.index('!') + 1)
    if index == -1:
        return (quiet, mask, None)
    else:
        forward = mask[index+1:]
        mask = mask[:index]
        return (quiet, mask, forward)


def parseMask(mask):
    """Return a tuple (nick, user, host)."""
    nick, rest = mask.split('!')
    user, host = rest.split('@')
    return nick, user, host


class BanManagementHandler(commands.CommandHandler,
                           commands.ConfigurableMixin):

    """Manage bans."""

    implements(ibot.IChannelModeHandler, ibot.IJoinHandler)

    configurables = {
        u'announcechannel': (
            u'channel to announce on', 'announceChannel', u'%r',
            commands.bytestring),
        u'expiry': (
            u'default expiration time in seconds', 'defaultExpiry', None,
            commands.integer),
        }

    def __init__(self, defaultExpiry=2*7*24*60*60, announceChannel=None,
                 briefSeconds=10*60):
        commands.CommandHandler.__init__(self)
        self.defaultExpiry = defaultExpiry
        self.announceChannel = announceChannel
        self.briefSeconds = briefSeconds
        self.regexpCache = {}
        self.victims = None
        self.lastBan = None
        self.guessedIncident = None
        self.set_path(u'None')

    def set_path(self, args):
        path, = commands.parseArgs(args, commands.bytestring)
        if path == 'None':
            dburi = uri.URI('sqlite:?timeout=0')
        else:
            dburi = uri.URI('sqlite:%s?timeout=0' % (uri.escape(path),))

        db = storm.Store(storm.create_database(dburi))

        db.execute('CREATE TABLE IF NOT EXISTS [trackedchannel] '
                   '([id] INTEGER PRIMARY KEY, [channel] BLOB NOT NULL)')
        db.execute('CREATE UNIQUE INDEX IF NOT EXISTS [trackedchannel_name] '
                   'ON [trackedchannel] ([channel])')

        db.execute('CREATE TABLE IF NOT EXISTS [incident] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [name] TEXT NOT NULL,'
                   ' [description] TEXT)')

        db.execute('CREATE TABLE IF NOT EXISTS [offender] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [nickname] BLOB NOT NULL,'
                   ' [username] BLOB NOT NULL,'
                   ' [hostname] BLOB NOT NULL,'
                   ' [incident_id] INTEGER NOT NULL'
                   ' REFERENCES [incident] ([id]) ON DELETE CASCADE)')

        db.execute('CREATE TABLE IF NOT EXISTS [ban] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [channel] BLOB NOT NULL,'
                   ' [mask] BLOB NOT NULL,'
                   ' [forward] BLOB,'
                   ' [quiet] INTEGER NOT NULL,'
                   ' [setter] BLOB NOT NULL,'
                   ' [incident_id] INTEGER'
                   ' REFERENCES [incident] ([id]) ON DELETE SET NULL,'
                   ' [set] TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP,'
                   ' [expires] TEXT,'
                   ' [unset] TEXT)')

        self.store = db

    def getRegexp(self, pattern):
        """Return a compiled regexp, from our cache if possible."""
        pattern = fnmatch.translate(pattern)
        regexp = self.regexpCache.get(pattern)
        if regexp is None:
            regexp = self.regexpCache[pattern] = re.compile(pattern)
        return regexp

    def joined(self, channel):
        """Get some data on present users."""
        # TODO: figure out a nonsuck way to tell the bot we need this.
        # (we do not want to do multiple whos per channel)
        channel.who()

    def channelModeChanged(self, setter, channel, modeSet, mode, mask=None):
        """Add or remove Ban objects if we receive the relevant modechange."""
        if mode != 'b' or not mask:
            # We do not care.
            return
        folder = channel.bot.caseFolder
        if self.store.find(
            TrackedChannel,
            TrackedChannel.channel == folder(channel.name)).one() is None:
            # We do not care.
            return
        setterUser = ibot.IUser(setter, None)
        if setterUser is None:
            setterName = setter
        else:
            setterName = setter.nickname
        quiet, mask, forward = parseBan(channel.bot.caseFolder(mask))

        announceChannel = channel.bot.user.channels.get(self.announceChannel)
        if announceChannel is None:
            log.err('help! I am not on my announce channel!')
            announce = None
        else:
            announce = announceChannel.notice
        if modeSet:
            # Do stuff if this ban is new.
            try:
                """Possibly add it."""
                if self.store.find(Ban,
                                   Ban.channel == channel.name,
                                   Ban.mask == mask,
                                   Ban.forward == forward,
                                   Ban.quiet == quiet,
                                   Ban.unset == None).one() is not None:
                    # It exists, ignore it.
                    return
                self.lastBan = Ban()
                self.lastBan.channel = channel.name
                self.lastBan.mask = mask
                self.lastBan.forward = forward
                self.lastBan.quiet = quiet
                self.lastBan.expires = (
                    datetime.datetime.utcnow() +
                    datetime.timedelta(seconds=self.defaultExpiry))
                self.lastBan.setter = setterName
                self.store.add(self.lastBan)

                if announce is None:
                    # Crap.
                    return
                def strannounce(msg):
                    announce(msg.encode('string-escape').decode('ascii'))
                if forward:
                    strannounce('New forward: %s to %s' % (mask, forward))
                elif quiet:
                    strannounce('New quiet: %s' % (mask,))
                else:
                    strannounce('New ban: %s' % (mask,))
                regexp = self.getRegexp(mask)
                self.victims = list(
                    user for user in channel.bot.users.itervalues()
                    if regexp.match('%s!%s@%s' % (
                            user.nickname, user.username, user.hostname)
                                    ) is not None)
                # TODO this was added because axiom does not expose the
                # glob support sqlite has. Does storm?
                nick, user, host = tuple(
                    bit.replace('?', '_').replace('*', '%')
                    for bit in parseMask(mask))
                offenders = list(self.store.find(
                        Offender,
                        Offender.nickname.like(nick),
                        Offender.username.like(user),
                        Offender.hostname.like(host)))
                incidents = (set(off.incident for off in offenders) -
                             set([None]))
                self.guessedIncident = None
                if len(incidents) == 1:
                    incident = list(incidents)[0]
                    announce(u'Matches existing incident "%s", '
                             u'say "addtoincident" to add this ban to it' % (
                            incident.name,))
                    self.guessedIncident = incident
                elif incidents:
                    # XXX infinite spam for *!*@*
                    announce(u'Matches incidents "%s", say '
                             u'"addtoincident <incident>" to add this ban' % (
                            u'", "'.join(inc.name for inc in incidents),))
                if not self.victims:
                    string = ('No current users affected! '
                              'Say "newincident <incident>" to track as new, ')
                elif len(self.victims) == 1:
                    victim = self.victims[0]
                    string = ('Affected: %s!%s@%s. '
                              'Say "newincident [incident]" to track as new, '
                              % (victim.nickname, victim.username,
                                 victim.hostname))
                else:
                    string = ('%d affected. '
                              'Say "newincident <incident>" to track as new, '
                              % (len(self.victims),))
                announce(string.encode('string-escape').decode('ascii') +
                         u'"addoffender <incident> to add to existing, '
                         u'"forget" to prune from the db.')
            except:
                self.store.rollback()
                raise
            else:
                self.store.flush()
        else:
            # XXX this seems to miss tests and a flush() call...
            # Kill any active entries matching this ban
            ban = self.store.find(Ban,
                                  Ban.channel == channel.name,
                                  Ban.mask == mask,
                                  Ban.forward == forward,
                                  Ban.quiet == quiet,
                                  Ban.unset == None).one()
            if ban is not None:
                if datetime.datetime.utcnow() - ban.set < (
                    datetime.timedelta(seconds=self.briefSeconds)):
                    if announce is not None:
                        announce(u'Removing brief ban "%s"' % (
                                ban.mask.encode(
                                    'string-escape').decode('ascii'),))
                    self.store.remove(ban)
                else:
                    ban.unset = datetime.datetime()

    def _addVictims(self, announce, incident, victims, folder):
        """Helper function to add Offenders for a bunch of Users."""
        for victim in victims:
            offender = self.store.find(
                Offender,
                Offender.nickname == victim.nickname,
                Offender.username == victim.username,
                Offender.hostname == victim.hostname).one()
            if offender is None:
                offender = Offender()
                offender.nickname = victim.nickname
                offender.username = victim.username
                offender.hostname = victim.hostname
                offender.incident = incident
                self.store.add(offender)
                announce(u'Offender "%s" added to incident "%s"' % (
                        offender.nickname.encode(
                            'string-escape').decode('ascii'), incident.name))
            else:
                announce(
                    u'Offender "%s" not moved from incident "%s" to "%s"' % (
                        offender.nickname.encode(
                            'string-escape').decode('ascii'),
                        offender.incident.name,
                        incident.name))

    @commands.optargs(0, commands.string)
    def handle_newincident(self, user, source, name=None):
        """Log a new incident, guessing the name from last ban if missing."""
        if name is None:
            if self.victims is None:
                source.msgTo(user, u'Give me a name or ban someone.')
                return
            if len(self.victims) > 1:
                source.msgTo(user, u'Give me a name or ban exactly one client.')
                return
            name = self.victims[0].nickname.encode(
                'string-escape').decode('ascii')
            source.msgTo(user, u'Using name "%s" from last ban' % (name,))
        announceChannel = user.bot.user.channels.get(self.announceChannel)
        if announceChannel is None:
            source.msgTo(user, u'Help! I am not on my announce channel!')
            return
        announce = announceChannel.notice
        try:
            if self.store.find(Incident,
                               Incident.name == name).one() is not None:
                source.msgTo(user, u'An incident with that name exists.')
                return
            incident = Incident(name)
            announce(u'Incident "%s" created' % (incident.name,))
            if self.victims:
                self._addVictims(
                    announce, incident, self.victims, user.bot.caseFolder)
            self.victims = None
            if self.lastBan:
                self.lastBan.incident = incident
                announce(u'Ban "%s" added to incident "%s"' % (
                        self.lastBan.banmask(), incident.name))
                self.lastBan = None
        except:
            self.store.rollback()
            raise
        else:
            self.store.flush()

    @commands.optargs(0, commands.string)
    def handle_addtoincident(self, user, source, name=None):
        """Add the last-added ban to a new incident."""
        if name is None:
            if self.guessedIncident is None:
                source.msgTo(user, u'Give me a name '
                             u'or ban someone matching one incident')
                return
            incident = self.guessedIncident
        else:
            incident = self.store.find(Incident, Incident.name == name).one()
            if incident is None:
                source.msgTo(user, u'No incident with that name')
                return
        if self.lastBan is None:
            source.msgTo(user, u'No tracked ban')
            return
        announceChannel = user.bot.user.channels.get(self.announceChannel)
        if announceChannel is None:
            source.msgTo(user, u'Help! I am not on my announce channel!')
            return
        self.lastBan.incident = incident
        announceChannel.notice(u'Ban "%s" added to incident "%s"' % (
                self.lastBan.banmask(), incident.name))
        self.lastBan = None

    @commands.args(commands.string, commands.longstring)
    def handle_setdescription(self, user, source, name, description):
        """setdescription <incident> <description>: set longer description."""
        try:
            incident = self.store.find(Incident,
                                       Incident.name == name).one()
            if incident is None:
                source.msgTo(user, u'No incident with that name')
                return
            incident.description = description
            source.msgTo(user, u'Description set.')
            # TODO announce this?
        except:
            self.store.rollback()
            raise
        else:
            self.store.flush()

    @commands.args(commands.string)
    def handle_addoffender(self, user, source, name):
        """Add a new offender/offenders to an incident."""
        if not self.victims:
            source.msgTo(user, u'no offenders to add.')
            return
        announceChannel = user.bot.user.channels.get(self.announceChannel)
        if announceChannel is None:
            source.msgTo(user, u'Help! I am not on my announce channel!')
            return
        announce = announceChannel.notice
        try:
            incident = self.store.find(Incident, Incident.name == name).one()
            if incident is None:
                source.msgTo(user, u'No such incident!')
                return
            self._addVictims(
                announce, incident, self.victims, user.bot.caseFolder)
            self.victims = None
        except:
            self.store.rollback()
            raise
        else:
            self.store.flush()

    @commands.args(commands.bytestring)
    def handle_whybanned(self, user, source, query):
        """bans <query>: Lists channel bans (matching query)."""
        folder = user.bot.caseFolder
        query = folder(query)
        target = user.bot.users.get(query)
        if target is not None:
            source.msgTo(
                user,
                ('using mask %s!%s@%s from online user' % (
                        target.nickname, target.username, target.hostname)
                 ).encode('string-escape').decode('ascii'))
        # Look for matching Offenders.
        if target is None:
            likeQuery = '%%%s%%' % (query, )
            offenders = list(self.store.find(
                    Offender,
                    storm.Or(Offender.nickname.like(likeQuery),
                             Offender.username.like(likeQuery),
                             Offender.hostname.like(likeQuery))))
            bans = list(self.store.find(Ban,
                                        Ban.unset != None,
                                        Ban.mask.like(likeQuery)))
        else:
            offenders = list(self.store.find(
                    Offender,
                    Offender.nickname == folder(target.nickname),
                    Offender.username == folder(target.username),
                    Offender.hostname == folder(target.hostname)))
            mask = '%s!%s@%s' % (
                target.nickname, target.username, target.hostname)
            bans = list(
                ban for ban in self.store.find(Ban, Ban.unset == None)
                if self.getRegexp(ban.mask).match(mask) is not None)
        incidents = (set(offender.incident for offender in offenders) |
                     set(ban.incident for ban in bans)) - set((None,))
        for incident in incidents:
            if incident.description:
                message = u'Incident: "%s" (%s)' % (
                    incident.name, incident.description)
            else:
                message = u'Incident: "%s"' % (incident.name,)
            source.msgTo(user, message)
        if offenders:
            source.msgTo(user, u'Offenders: %s' % (u', '.join(
                        offender.hostmask() for offender in offenders[:5])))
            if len(offenders) > 5:
                source.msgTo(user, u'(%s out of %s)' % (len(offenders),))
        if bans:
            for ban in bans[:5]:
                source.msgTo(user, u'Ban: %s (set by %s %s ago)' % (
                        ban.banmask(),
                        ban.setter.encode('string-escape').decode('ascii'),
                        commands.formatTime(datetime.datetime.utcnow()
                                            - ban.set)))
            if len(bans) > 5:
                source.msgTo(user, u'(5 out of %s' % (len(bans,)))
        if not bans and not offenders and not incidents:
            source.msgTo(user, u'Nothing found!')

    @commands.args()
    def handle_unknownbans(self, user, source):
        """Check for unknown bans."""
        source.msgTo(user, u'You may have to run this one twice.')
        for channelObj in self.store.find(TrackedChannel):
            channelName = channelObj.channel
            channel = user.bot.user.channels.get(channelName)
            if channel is None:
                source.msgTo(u'skipping %s (I am not on it)' % (channelName,))
                continue
            channel.mode(True, 'b')
            # (setter and timestamp unused) pylint: disable-msg=W0612
            for ban, (setter, timestamp) in channel.modes['b'].iteritems():
                quiet, mask, forward = parseBan(user.bot.caseFolder(ban))
                if self.store.find(Ban,
                                   Ban.channel == channelName,
                                   Ban.mask == mask,
                                   Ban.forward == forward, Ban.quiet == quiet,
                                   Ban.unset == None).one() is None:
                    source.msgTo(
                        user,
                        ('%s: %s' % (channelName, mask)).encode(
                            'string-escape').decode('ascii'))

    def get_expiry(self):
        """Get default expiration time."""
        delta = datetime.timedelta(seconds=self.defaultExpiry)
        return unicode(commands.formatTime(delta))

    def get_channels(self):
        """Return the list of channels we listen on."""
        return u', '.join(
            channel.channel.encode('string-escape').decode('ascii')
            for channel in self.store.find(TrackedChannel))

    def set_addchannel(self, args):
        """Add a channel to track."""
        channel, = commands.parseArgs(args, commands.bytestring)
        if self.store.find(TrackedChannel,
                           TrackedChannel.channel == channel).one() is None:
            self.store.add(TrackedChannel(channel))
            self.store.flush()

    def set_removechannel(self, args):
        """Remove a channel from the tracked list."""
        channel, = commands.parseArgs(args, commands.bytestring)
        self.store.find(TrackedChannel,
                        TrackedChannel.channel == channel).remove()


class TrackedChannel(object):

    """A channel tracked by a BanManagementHandler."""

    __storm_table__ = 'trackedchannel'

    id = storm.Int(primary=True)
    channel = storm.RawStr()

    def __init__(self, channel):
        self.channel = channel


class Incident(object):

    """Represents an incident/offender (mainly for other stuff to refer to)."""

    __storm_table__ = 'incident'

    id = storm.Int(primary=True)
    name = storm.Unicode()
    description = storm.Unicode()

    def __init__(self, name):
        self.name = name


class Offender(object):

    """Represents a connection/client from an offender."""

    __storm_table__ = 'offender'

    id = storm.Int(primary=True)
    nickname = storm.RawStr()
    username = storm.RawStr()
    hostname = storm.RawStr()
    incident_id = storm.Int()
    incident = storm.Reference(incident_id, Incident.id)

    def hostmask(self):
        """Return a hostmask as unicode."""
        return ('%s!%s@%s' % (self.nickname, self.username, self.hostname)
                ).encode('string-escape').decode('ascii')


class Ban(object):

    """Represents a single tracked ban."""

    __storm_table__ = 'ban'

    id = storm.Int(primary=True)
    channel = storm.RawStr()
    mask = storm.RawStr()
    forward = storm.RawStr()
    quiet = storm.Bool()
    setter = storm.RawStr()
    incident_id = storm.Int()
    incident = storm.Reference(incident_id, Incident.id)
    set = storm.DateTime()
    expires = storm.DateTime()
    unset = storm.DateTime()

    def banmask(self):
        """Return unicode like what you could give to a +b modechange."""
        mask = self.mask
        if self.quiet:
            mask = '%' + mask
        if self.forward:
            mask = '%s!%s' % (mask, self.forward)
        return mask.encode('string-escape').decode('ascii')
