
"""Connect the bot to a buildbot.

The change notification currently needs a hacked up buildmaster
(buildbot's RemoteChange makes no sense). Getting a RemoteChange that
makes sense and probably also a RemoteControl into buildbot is
planned, but do not hold your breath.
"""


import collections

from zope.interface import implements

from twisted.spread import pb
from twisted.cred import credentials
from twisted.python import log
from twisted.internet import reactor, defer

from mzbot import commands, ibot


# Random copy from builder.py from buildbot
RESULTS = ["success", "warnings", "failure", "skipped", "exception"]


class PBConnectMixin(commands.ConfigurableMixin):

    """Helper class that keeps a pb connection alive."""

    configurables = {
        'host': (
            'the remote host', 'host', u'%s', commands.bytestring),
        'port': (
            'the remote port', 'port', u'%s', commands.rangedInteger(0)),
        'username': (
            'the username used for connecting',
            'name', u'%s', commands.bytestring),
        'password': (
            'the password used for connecting',
            'password', None, commands.bytestring),
        }

    def _pbConnect(self, existing=None):
        self.remote = None
        creds = credentials.UsernamePassword(self.name, self.password)
        factory = pb.PBClientFactory()
        d = factory.login(creds)
        d.addCallbacks(self._pbConnected, self._pbConnectionFailed)
        d.addErrback(log.err)
        reactor.connectTCP(self.host, self.port, factory)
        return d

    def _pbConnected(self, remote):
        self.remote = remote
        remote.notifyOnDisconnect(self._pbConnect)

    def _pbConnectionFailed(self, fail):
        log.err(fail)
        reactor.callLater(60, self._pbConnect)

    @commands.args()
    def handle_remotestatus(self, user, source):
        """Check if our buildbot connection is up."""
        if self.remote is None:
            source.msgTo(user, 'disconnected')
        else:
            source.msgTo(user, 'connected')


class BuildBotSender(commands.CommandHandler, PBConnectMixin):

    """Send commands to a buildbot."""

    def __init__(self, host='localhost', port=9989,
                 name='change', password='changepw'):
        """Initialize."""
        commands.CommandHandler.__init__(self)
        self.host = host
        self.port = port
        self.name = name
        self.password = password
        self._pbConnect()

    @commands.args(commands.bytestring, commands.bytestring,
                   commands.longbytestring)
    def handle_forcechange(self, user, source, branch, filename, comments):
        """forcechange <branch> <filename> <comments>: force/fake a change."""
        if self.remote is None:
            source.msgTo(user, 'Not currently connected. Try again later.')
            return
        d = self.remote.callRemote(
            'addChange', dict(who=user.nickname, branch=branch,
                              files=[filename], comments=comments))
        d.addCallbacks(self.changeSent, self.changeFailed,
                       callbackArgs=(user, source), errbackArgs=(user, source))
        return d

    def changeSent(self, result, user, source):
        """addChange callback."""
        source.msgTo(user, 'Change sent')

    def changeFailed(self, fail, user, source):
        """addChange errback."""
        source.msgTo(user, 'Change failed (%s)!' % (fail,))


class BuildBotStatus(commands.CommandHandler,
                     PBConnectMixin,
                     pb.Referenceable):

    """Receive status notifications from a buildbot."""

    implements(ibot.ILoginHandler, ibot.IDisconnectHandler)

    default_changes_len = 100

    def __init__(self, host='localhost', port=9989,
                 name='statusClient', password='clientpw', channels=''):
        """Initialize."""
        commands.CommandHandler.__init__(self)
        self.host = host
        self.port = port
        self.name = name
        self.password = password
        self.channels = channels
        self.bot = None
        # Hack. We do not get any kind of change notification or scheduler
        # activity notification, just build notification. But we do get a
        # "reason" for every build, including changes that went into it.
        # We can use that to print changes, but if we do that "blindly" we
        # do duplicate announcements if one change triggers multiple
        # builds. So we keep a backlog of a few recent changes.
        self.changes = collections.deque()
        # This is an instance attribute, not a class attribute, because we
        # have to grow it if we see a single batch longer than the queue.
        self.changes_len = self.default_changes_len
        self._pbConnect()

    def loggedIn(self, bot):
        """Remember this bot as announce target."""
        # XXX this misbehaves when someone hooks this up to multiple bots.
        # assert self.bot is None
        self.bot = bot
        return defer.succeed(True)

    def disconnected(self, bot, reason=None):
        """Drop the reference to our announce target bot."""
        # assert self.bot is bot
        self.bot = None

    def _pbConnected(self, remote):
        PBConnectMixin._pbConnected(self, remote)
        remote.callRemote('subscribe', 'builds', 5, self)

    def _announce(self, msg):
        if self.bot is not None:
            for channel in self.channels.split():
                chanobj = self.bot.user.channels.get(channel)
                if chanobj is not None:
                    chanobj.msg(msg)

    def remote_builderAdded(self, builderName, builder):
        """A builder connected."""
        self._announce('builder %r added' % (builderName,))
        # For the "real" IStatusReceiver we should return self here,
        # but that is not necessary for this remote version.

    def remote_builderRemoved(self, builderName):
        """A builder disconnected."""
        self._announce('builder %r removed' % (builderName,))

    def remote_builderChangedState(self, builderName, state, eta=None):
        """A builder changed state.

        The final arg is deprecated but currently required.
        """
        # No need to do anything here, we get separate calls for build
        # start/finish (and do not really have to announce if the
        # builder is idle or offline)

    def remote_buildStarted(self, builderName, build):
        """A builder started building."""
        d = build.callRemote('getChanges')
        d.addCallback(self._announceChanges)
        d.addErrback(log.err)

    def _announceChanges(self, changes):
        """Announce any changes that are new."""
        # Boo, have to request all of these separately.
        d = defer.gatherResults(list(
                defer.gatherResults(list(
                        change.callRemote(method) for change in changes))
                for method in ['getNumber', 'getComments', 'getWho']))
        d.addCallback(self._reallyAnnounceChanges)
        d.addErrback(log.err)

    def _reallyAnnounceChanges(self, changes):
        toAnnounce = []
        for number, comment, who in zip(*changes):
            if number in self.changes:
                # Already seen this one.
                continue
            self.changes.append(number)
            toAnnounce.append((comment, who))
        # TODO unhardcode this.
        if len(toAnnounce) > 10:
            authors = set(zip(*toAnnounce)[1])
            self._announce('A ton of changes by %s' % (', '.join(authors),))
        else:
            # Perfectly valid to have nothing toAnnounce (already seen
            # everything in changes).
            for comment, who in toAnnounce:
                self._announce('Change by %s: %s' % (who, comment.strip()))
        self.changes_len = max(self.changes_len, len(changes[0]))
        while len(self.changes) > self.changes_len:
            self.changes.popleft()

    def remote_buildFinished(self, builderName, build, results):
        """A builder stopped building."""
        self._announce('builder %r finished: %s' % (builderName,
                                                    RESULTS[results]))

    def get_channels(self):
        """Get a list of channels we search logs for and users on."""
        return u', '.join(sorted(self.channels.split()))

    def set_addchannel(self, args):
        """Add a channel to the list of searched channels."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        channels.add(channel)
        self.channels = ' '.join(channels)

    def set_removechannel(self, args):
        """Remove a channel from the list of searched channels."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        channels.discard(channel)
        self.channels = ' '.join(channels)
