
"""Classes and functions to aid in writing CommandHandlers."""


import inspect

from zope.interface import implements

from twisted.python import util

from mzbot import ibot, ebot


class ConfigurableMixin(object):

    """Mixin to help implementing IConfigurable.

    Provide get_foo() and return the result of get(u'foo'),
    set_bar(value) which is called from set(u'bar', value).

    set_ should take and get_ should return unicode.

    For sufficiently simple attributes, provide a configurables
    attribute like:

    {u'floodtime': (
        u'Number of seconds the flood protection looks at',
        'floodTime', u'%.0f seconds', commands.number),
    }

    The get_ and set_ methods override entries in this dict. Setting
    the format string or function to None disables the getter or
    setter respectively.
    """

    implements(ibot.IConfigurable)

    configurables = {}

    def gettables(self):
        """Return a dict based on get_methods and docstrings."""
        d = dict(
            (name, doc)
            for name, (doc, attr, format, func)
            in self.configurables.iteritems()
            if format is not None)
        d.update(
            (name[4:].decode('ascii'),
             getattr(self, name).__doc__.split('\n', 1)[0])
            for name in dir(self)
            if (name.startswith('get_') and len(name) > 4 and
                inspect.ismethod(getattr(self, name))))
        return d

    def settables(self):
        """Return a dict based on set_methods and docstrings."""
        d = dict(
            (name, doc)
            for name, (doc, attr, format, func)
            in self.configurables.iteritems()
            if func is not None)
        d.update(
            (name[4:].decode('ascii'),
             getattr(self, name).__doc__.split('\n', 1)[0])
            for name in dir(self)
            if (name.startswith('set_') and len(name) > 4 and
                inspect.ismethod(getattr(self, name))))
        return d

    def get(self, key):
        """Return the result from a get_method."""
        try:
            func = getattr(self, 'get_%s' % (key.encode('ascii'),), None)
        except UnicodeError:
            func = None

        if func is not None:
            return func()
        else:
            doc, attr, format, func = self.configurables[key]
            return format % (getattr(self, attr),)

    def set(self, key, value):
        """Call a set_method."""
        try:
            func = getattr(self, 'set_%s' % (key.encode('ascii'),), None)
        except UnicodeError:
            func = None

        if func is not None:
            return func(value)
        else:
            doc, attr, format, func = self.configurables[key]
            parsed_value, = parseArgs(value, func)
            setattr(self, attr, parsed_value)


class CommandHandler(object):

    """The base CommandHandler.

    You'll want to subclass, override commands and add handle_foo(
    user, channel, args) commands.

    Make sure you call the __init__ method this class has.
    """

    implements(ibot.ICommandHandler, ibot.ICommandHelpProvider)

    def __init__(self):
        self.commandset = set()
        for name in dir(self):
            if (name.startswith('handle_') and len(name) > 7 and
                inspect.ismethod(getattr(self, name))):
                self.commandset.add(name[7:].decode('ascii'))

    def handle(self, user, source, command, myArgs):
        """Handle an incoming command. We may assume it's in 'commands'."""
        handler = getattr(self, 'handle_' + command.encode('ascii'))
        return handler(user, source, myArgs)

    def getCommandDescription(self, command):
        """Use our handle_ docstrings as description."""
        command = command.encode('ascii')
        doc = getattr(self, 'handle_' + command).__doc__
        if not doc:
            return 'undocumented'
        doc = doc.split('\n', 1)[0]
        if doc.startswith(command) and ':' in doc:
            doc = doc[len(command):].lstrip()
        return doc


def notInPrivmsg(func):
    """Decorator for a handle_ that doesn't want privmsgs."""
    def new_f(self, user, source, *myArgs):
        """Reply if it's in a privmsg, pass on to func otherwise."""
        if user is source:
            source.msgTo(user, u"doesn't work in a privmsg")
        else:
            return func(self, user, source, *myArgs)
    return util.mergeFunctionMetadata(func, new_f)

def onlyInPrivmsg(func):
    """Decorator for a handle_ that only wants privmsgs."""
    def new_f(self, user, source, *myArgs):
        """Reply if it's not in a privmsg, pass on to func otherwise."""
        if user is not source:
            source.msgTo(user, u'do this in a privmsg')
        else:
            return func(self, user, source, *myArgs)
    return util.mergeFunctionMetadata(func, new_f)


def maybeLoggedIn(interface):
    """Decorator for a handle_ that uses the user's avatar.

    interface is the interface the avatar is adapted to or IAvatar for
    the unadapted avatar object.

    Converts the signature from (self, user, source, *args) to
    (self, user, source, avatar, *args). avatar is the avatar or None.
    """
    def _decorator(func):
        def newFunc(self, user, source, *myArgs):
            """Get the avatar."""
            avatar = user.data.get(ibot.IAvatar)
            if avatar is not None and interface is not ibot.IAvatar:
                avatar = interface(avatar, None)
            return func(self, user, source, avatar, *myArgs)
        return util.mergeFunctionMetadata(func, newFunc)
    return _decorator

def loggedIn(interface):
    """Like maybeLoggedIn, but avatar is never None."""
    def _decorator(func):
        def newFunc(self, user, source, *myArgs):
            """Error reply if no avatar."""
            avatar = user.data.get(ibot.IAvatar)
            if avatar is None:
                source.msgTo(user, u'you are not logged in!')
                return
            if interface is not ibot.IAvatar:
                avatar = interface(avatar, None)
            if avatar is None:
                source.msgTo(
                    user,
                    u'you are logged in but do not have the needed access!')
            else:
                return func(self, user, source, avatar, *myArgs)
        return util.mergeFunctionMetadata(func, newFunc)
    return _decorator


def kwargs(*options):
    """Decorator for a handle_ that has multiple acceptable ways to call it.

    Arguments are iterables of (name, filter function) tuples. They
    are tried in order, and the first one that matches the supplied
    args is used in a kwargs call to your function.

    This function is tricky to use. It might be more convenient to
    just do your own argument parsing.

    Example:

    @kwargs([("number", integer), ("text", string)],
            [("text", string), ("number", integer)],
            [("number", integer)],
            [("text", string)])
    def handle_foo(self, user, source, number=None, text=None):
        pass

    this will allow the following to work:
    bot: foo 42
    bot: foo bar
    bot: foo 42 bar
    bot: foo bar 42

    Notice the order of iterables often matters: since "42" passes
    both the string and integer filter, if the single-integer option
    was listed after the single-string one, "bot: foo 42" would result
    in a call to handle_foo(text="42")!
    """
    def _decorator(func):
        def _new_f(self, user, source, origArgs):
            try:
                result = parseKwargs(origArgs, *options)
            except ebot.InvalidArgument, e:
                source.msgTo(user, unicode(e))
            else:
                return func(self, user, source, **result)
        return util.mergeFunctionMetadata(func, _new_f)
    return _decorator

def parseKwargs(origArgs, *options):
    """Non-decorator function that does kwargs parsing.

    See the docstring for 'kwargs' for what options should be.
    origArgs is the args to parse. Returns a dict if origArgs are
    valid, raises ebot.InvalidArgument if they are invalid.
    """
    failures = set()
    for filters in options:
        argsLeft = origArgs
        result = {}
        for name, filterFunc in filters:
            if not argsLeft:
                failures.add(u'not enough arguments')
                break
            try:
                newOpt, argsLeft = filterFunc(argsLeft)
            except ebot.InvalidArgument, e:
                failures.add(unicode(e))
                break
            result[name] = newOpt
        else:
            if argsLeft:
                failures.add(u'too many arguments')
            else:
                return result
    raise ebot.InvalidArgument(u'one of the following: %s' % (
            u', '.join(sorted(failures)),))


def optargs(required, *filters):
    """Decorator for a handle_ that has optional arguments.

    the first arg is the amount of *required* arguments. Other args
    are filter functions.

    Example:

    @optargs(1, integer, string)
    def handle_foo(self, user, source, number, text=None):
        pass

    This can be called as:
    bot: foo 42 text
    bot: foo 42

    Notice you can achieve the same thing with:
    @kwargs([("number", integer), ("text", string)],
            [("number", integer)])

    However, optargs gives better error messages when incorrect
    arguments are supplied and is easier to read in code.
    """
    def _decorator(func):
        def _new_f(self, user, source, origArgs):
            try:
                result = parseOptArgs(origArgs, required, *filters)
            except ebot.InvalidArgument, e:
                source.msgTo(user, unicode(e))
            else:
                return func(self, user, source, *result)
        return util.mergeFunctionMetadata(func, _new_f)
    return _decorator

def parseOptArgs(origArgs, required, *filters):
    """Non-decorator function that does optargs parsing.

    See the docstring for 'optargs' for what 'required' and 'filters'
    should be. origArgs is the args to parse. Returns a list if
    origArgs are valid, raises ebot.InvalidArgument if they are invalid.
    """
    argsLeft = origArgs
    result = []
    for filterFunc in filters:
        if not argsLeft:
            if len(result) >= required:
                return result
            else:
                raise ebot.InvalidArgument(u'not enough arguments')
        newOpt, argsLeft = filterFunc(argsLeft)
        result.append(newOpt)
    if argsLeft:
        raise ebot.InvalidArgument(u'too many arguments')
    return result

def args(*filters):
    """Decorator for a handle_ that takes some args 'filtered' by a function.

    arguments are filter functions. These are tried in order.

    Example:

    @args(integer, string)
    def handle_foo(self, user, source, number, text):
        pass
    """
    return optargs(len(filters), *filters)

def parseArgs(origArgs, *filters):
    """Exactly like parseOptArgs but all args are required."""
    return parseOptArgs(origArgs, len(filters), *filters)


def string(myArgs):
    """Filter out a single string (unicode object).

    Supports unquoted, single-quoted and double-quoted strings.
    """
    # This used to shlex but I switched it to something homegrown because
    # I do not understand the parsing rules and api shlex uses fully
    # and want to do this on unicode.
    args = myArgs.lstrip()
    if not args:
        raise ebot.InvalidArgument(u'argument is empty')

    if args[0] not in u'"\'':
        # Unquoted string, just split.
        argseq = args.split(None, 1)
        if len(argseq) == 1:
            return argseq[0], u''
        return argseq[0], argseq[1].lstrip()

    # Quoted string.
    quote, args = args[0], args[1:]
    escaped = False
    out = []
    # This scales terribly, but that does not matter to me.
    while args:
        c, args = args[0], args[1:]
        if not escaped:
            if c == u'\\':
                escaped = True
            elif c == quote:
                break
            else:
                out.append(c)
        else:
            if c in u'"\'\\':
                out.append(c)
            else:
                out.append(u'\\' + c)
            escaped = False
    else:
        raise ebot.InvalidArgument(u'"%s" is not properly quoted' % (myArgs,))

    rest = args.lstrip()
    if rest and rest == args:
        raise ebot.InvalidArgument(u'"%s" is not properly quoted' % (myArgs,))

    return u''.join(out), args.lstrip()

def bytestring(myArgs):
    """Return a bytestring (str object)."""
    result, rest = string(myArgs)
    try:
        return result.encode('ascii').decode('string-escape'), rest
    except UnicodeError:
        raise ebot.InvalidArgument(u'"%s" is not ascii' % (myArgs,))

def longbytestring(myArgs):
    """Filter all remaining arguments as a single bytestring (str object)."""
    # TODO: error handling?
    myArgs = myArgs.encode('ascii').decode('string-escape')
    return myArgs, u''

def longstring(myArgs):
    """Filter all remaining arguments as a single string (unicode object)."""
    return myArgs, u''

def stringlist(rest):
    """Filter out all strings and return them as a list of unicode."""
    result = []
    while rest:
        token, rest = string(rest)
        result.append(token)
    return result, u''

def bytestringlist(rest):
    """Filter out all strings and return them as a list of bytestrings."""
    result = []
    while rest:
        token, rest = bytestring(rest)
        result.append(token)
    return result, u''

def integer(myArgs):
    """Filter any positive or negative integer.

    This uses python's 'int' for conversion, it accepts anything that does.
    """
    arg, rest = string(myArgs)
    try:
        arg = int(arg)
    except ValueError:
        raise ebot.InvalidArgument(u'"%s" is not an integer' % (arg,))
    return arg, rest

def rangedInteger(minimum=None, maximum=None):
    """Filter an integer between certain bounds."""
    def _parse(myArgs):
        arg, rest = integer(myArgs)
        if ((minimum is not None and arg < minimum) or
            (maximum is not None and arg > maximum)):
            raise ebot.InvalidArgument(u'out of range')
        return arg, rest
    return _parse

def number(myArgs):
    """Filter out a floating-point value."""
    arg, rest = string(myArgs)
    try:
        arg = float(arg)
    except ValueError:
        raise ebot.InvalidArgument(u'"%s" is not a number' % (arg,))
    return arg, rest


COLORS = {
    'white': 0,
    'black': 1,
    'blue': 2,
    'green': 3,
    'red': 4,
    'darkred': 5,
    'purple': 6,
    'darkyellow': 7,
    'yellow': 8,
    'brightgreen': 9,
    'darkaqua': 10,
    'aqua': 11,
    'lightblue': 12,
    'brightpurple': 13,
    'darkgrey': 14,
    'lightgrey': 15,
    }

def colorize(text, foreground, background=None):
    """Return text colorized.

    foreground is the foreground color number or a key of colors.
    background is a color like foreground or None for default background.
    """
    try:
        color = u'%d' % (COLORS[foreground],)
    except KeyError:
        try:
            color = u'%d' % (int(foreground),)
        except ValueError:
            raise ValueError('invalid foreground color %r' % foreground)
    if background is not None:
        try:
            background = u'%d' % (COLORS[background],)
        except KeyError:
            try:
                background = u'%d' % (int(background),)
            except ValueError:
                raise ValueError('invalid background color %r' % background)
        color = u'%s,%s' % (color, background)
    return u'\003%s%s\017' % (color, text)

# Note this is copied directly from Supybot:
def bold(text):
    """Returns the string s, bolded."""
    return u'\x02%s\x02' % (text,)

def reverse(text):
    """Returns the string s, reverse-videoed."""
    return u'\x16%s\x16' % (text,)

def underline(text):
    """Returns the string s, underlined."""
    return u'\x1F%s\x1F' % (text,)

def resetstyle():
    """Returns a reset style code."""
    return u'\x0F'


def formatTime(timedelta):
    """Convert a datetime.timedelta to a human-readable string."""
    pieces = []
    divname = [(60, u'second'),
               (60, u'minute'),
               (24, u'hour'),
               (None, u'day')]
    timespan = timedelta.seconds
    for div, name in divname:
        if div is None:
            toFormat = timedelta.days
        else:
            timespan, toFormat = divmod(timespan, div)
        if toFormat:
            format = u'%d %s' % (toFormat, name)
            if toFormat > 1:
                format += u's'
            pieces.insert(0, format)
    if not pieces:
        return u'0 seconds'
    return u', '.join(pieces[:-2] + [u' and '.join(pieces[-2:])])
