
"""Highly experimental config file-driven bot creation."""


from __future__ import absolute_import, with_statement, division

import json
import difflib
from cStringIO import StringIO

from twisted.python import usage, reflect, log
try:
    from twisted.internet import ssl
    ssl # shut up pyflakes
except ImportError:
    ssl = None
from twisted.application import service, internet

from mzbot import irc, ibot, plugin, corehandlers, coreauth, commands

try:
    from pymeta.runtime import ParseError, EOFError
    from mzbot.authparser import Parser as AuthParser
    AuthParser, ParseError # shut up pyflakes
except ImportError:
    AuthParser = None

    class ParseError(ValueError):
        """Raised by parseAuth."""

    def parseAuth(handlers, string):
        try:
            return handlers[string][1]
        except KeyError:
            raise ParseError()
else:
    def parseAuth(handlers, string):
        def named(name):
            return handlers[name][1]
        def anded(handlers):
            return coreauth.AuthenticatorStack(handlers, True)
        def ored(handlers):
            return coreauth.AuthenticatorStack(handlers, False)
        globals_ = dict(named=named, anded=anded, ored=ored)
        # Based on the terml example in the pymeta source.
        parser = AuthParser(string, globals_)
        result, error = parser.apply('top')
        try:
            parser.input.head()
        except EOFError:
            return result
        else:
            raise error


class AdminHandler(corehandlers.AdminHandlerBase):

    def __init__(self, confpath, confdict, handlers):
        corehandlers.AdminHandlerBase.__init__(self)
        self.confpath = confpath
        self.confdict = confdict
        self.handlers = handlers

    def _getConfigurable(self, user, source, handlername):
        try:
            return ibot.IConfigurable(self.handlers[handlername][1])
        except KeyError:
            source.msgTo(user, 'no such handler')
        except TypeError:
            source.msgTo(user, 'not configurable')
        return None

    @commands.args()
    def handle_confdifflog(self, user, source):
        """Log a diff between the saved and actual config."""
        if self.confpath is None:
            source.msgTo(user, 'no confpath known')
            return

        import yaml
        with open(self.confpath, 'rb') as f:
            saved = list(f)
        handlerconf = {}
        confdict = dict(networks=self.confdict[u'networks'],
                        handlers=handlerconf)
        for name, (plugname, handler) in self.handlers.iteritems():
            handler = ibot.ISaveable(handler, None)
            if handler is not None:
                handlerconf[name] = confseq = [plugname]
                confseq.extend(list(t) for t in handler.save())
        actual = [line + '\n' for line in yaml.dump(confdict).split('\n')]
        for line in difflib.unified_diff(saved, actual):
            log.msg(line.rstrip('\n'))


LOADERS = {
    'json': 'json.load',
    'yaml': 'mzbot.yaml.load',
    }


def _coerceLoader(name):
    try:
        classname = LOADERS[name]
    except KeyError:
        raise usage.UsageError('%s does not exist (try %s)' %
                               (name, ', '.join(LOADERS)))

    try:
        return reflect.namedAny(classname)
    except ImportError, e:
        raise usage.UsageError('%s is unavailable (%s)' % (name, e))


class Options(usage.Options):

    synopsis = 'conffile'

    optFlags = [
        ('arg', 'a', 'Use the argument as data instead of a filename'),
        ]

    optParameters = [
        ('loader', 'l', json.load,
         'Config file format to use.', _coerceLoader),
        ]

    def parseArgs(self, conffile):
        self['conffile'] = conffile


def _buildHandler(name):
    plug = plugin.getConfigPluginByName(name)
    if plug is None:
        raise KeyError(name)
    return plug.buildHandler()


class HandlerDict(dict):

    def __missing__(self, key):
        handler = _buildHandler(key)
        self[key] = (key, handler)
        return key, handler


def makeService(config):

    if config['arg']:
        conffile = None
        confdict = config['loader'](StringIO(config['conffile']))
    else:
        try:
            conffile = config['conffile']
            with open(conffile, 'rb') as f:
                confdict = config['loader'](f)
        except EnvironmentError:
            # TODO: do something useful.
            raise

    #from pprint import pprint;pprint(confdict)
    #import yaml;print yaml.dump(confdict)

    handlers = HandlerDict()
    for handlername, data in confdict[u'handlers'].iteritems():
        it = iter(data)
        name = next(it)
        handler = _buildHandler(name)
        handlers[handlername] = (name, handler)
        # This is a format HACK: having these be two-elem sequences
        # instead of a single string we split means the "data" sequence
        # contains inner sequences, so it gets serialized like this:
        # autojoin:
        # - autojoin
        # - [add, stuff]
        # instead of like this:
        # autojoin: [autojoin, add stuff]
        for line in it:
            handler.set(*(s.encode('utf-8') for s in line))

    ms = service.MultiService()

    for name, settings in confdict[u'networks'].iteritems():
        # Shallow copy we can mutate:
        settings = dict(settings)

        host = settings.pop(u'hostname')
        use_ssl = settings.pop(u'ssl', False)
        port = settings.pop(u'port', 6697 if use_ssl else 6667)
        handlernames = settings.pop(u'handlers')

        holder = corehandlers.CommandHandlerHolder()
        for data in settings.pop(u'commands', []):
            handlername, handler = handlers[data[u'handler']]
            if u'channels' in data:
                for channel in data[u'channels']:
                    holder.addHandler(data[u'name'], handler, channel)
            else:
                holder.addHandler(data[u'name'], handler)

        if settings.pop(u'help', True):
            helpauth = settings.pop(u'helpauth', None)
            if helpauth is not None:
                helpauth = parseAuth(handlers, helpauth)
            holder.addHandler(u'help',
                              corehandlers.HelpHandler(holder, helpauth))
        if settings.pop(u'admin', True):
            admin = AdminHandler(conffile, confdict, handlers)
            holder.addHandler(u'admin', admin)

        for data in settings.pop(u'commandauth', []):
            string = data[u'handler']
            auth = parseAuth(handlers, string)
            if u'channels' in data:
                for channel in data[u'channels']:
                    if u'subcommands' in data:
                        holder.addAuth(data[u'name'], auth,
                                       data[u'subcommands'], channel=channel)
                    else:
                        holder.addAuth(data[u'name'], auth, channel=channel)
            else:
                if u'subcommands' in data:
                    holder.addAuth(data[u'name'], auth, data[u'subcommands'])
                else:
                    holder.addAuth(data[u'name'], auth)

        # XXX incomplete
        for k in [u'nickname', u'username', u'hostname']:
            if k in settings:
                settings[k] = settings[k].encode('ascii')

        # XXX yuck.
        if u'saslMechanisms' in settings:
            settings[u'saslMechanisms'] = [
                s.encode('ascii') for s in settings[u'saslMechanisms']]

        factory = irc.IRCBotFactory(**dict((k.encode('ascii'), v)
                                           for k, v in settings.iteritems()))
        for name in handlernames:
            factory.addHandler(handlers[name][1])
        factory.addHandler(holder)

        if use_ssl:
            # TODO: shiny error message if ssl unavailable?
            # XXX I suspect this accepts self-signed certs etc.
            contextFactory = ssl.ClientContextFactory()
            client = internet.SSLClient(host, port, factory, contextFactory)
        else:
            client = internet.TCPClient(host, port, factory)

        #from twisted.internet import reactor;reactor.callLater(
        #    0, admin.handle_confdifflog, None, None, '')

        client.setName(name)
        client.setServiceParent(ms)

    return ms
