
"""A bunch of possibly useful authenticators."""

import re
import fnmatch

from zope.interface import implements

from twisted.internet import defer

from mzbot import ibot, commands


class ModeAuthenticator(object):

    """I authorize if the user has a certain mode in one of my channels."""

    implements(ibot.IAuthenticator)

    def __init__(self, mode, channellist):
        self.mode = mode
        self.channellist = channellist

    def auth(self, user):
        """Return True if the user has the mode."""
        for channel in self.channellist:
            if channel in user.chanmodes[self.mode]:
                return defer.succeed(True)
        return defer.succeed(False)

    def description(self):
        """human-readable description."""
        return u'mode +%s in one of %s' % (
            self.mode.decode('ascii'),
            u', '.join(c.decode('ascii', 'replace')
                       for c in self.channellist))


class AuthenticatorStack(object):

    """Authorize if one or all of my children do.

    The children are tried in order, so put the "fast" ones first.
    """

    implements(ibot.IAuthenticator)

    def __init__(self, children, all=True):
        self.children = children
        self.all = all

    @defer.inlineCallbacks
    def auth(self, user):
        """Try all children in sequence, return as soon as possible."""
        for auth in self.children:
            result = yield auth.auth(user)
            if result and not self.all:
                defer.returnValue(True)
            if self.all and not result:
                defer.returnValue(False)
        defer.returnValue(self.all)

    def description(self):
        """human-readable description."""
        if self.all:
            word = u'all'
        else:
            word = u'one'
        return (u'%s of (%s)' %
                (word, u", ".join(i.description() for i in self.children)))


class AlwaysTheSameAuthenticator(object):

    """Always return (de)auth'd."""

    implements(ibot.IAuthenticator)

    def __init__(self, status):
        self.status = status

    def auth(self, user):
        """Return hardcoded value."""
        return defer.succeed(self.status)

    def description(self):
        """human-readable description."""
        if self.status:
            return u'always authenticated'
        else:
            return u'always denied'


class BaseNameAuthenticator(object):

    """Return auth'd if the user has a certain name."""

    implements(ibot.IAuthenticator)

    def __init__(self, globs):
        self.globs = globs
        self._compile(globs)

    def _compile(self, globs):
        self.res = list(re.compile(fnmatch.translate(pat)) for pat in globs)

    def _check(self, hostmask):
        mask = '%s!%s@%s' % hostmask
        for regexp in self.res:
            if regexp.match(mask) is not None:
                return True
        return False

    def auth(self, user):
        """Compare to our list."""
        return user.hostmask().addCallback(self._check)

    def description(self):
        """human-readable description."""
        # XXX gah, encoding...
        return u'hostmask matches one of %s' % ', '.join(
            mask.decode('ascii', 'replace') for mask in self.globs)


class NameAuthenticator(BaseNameAuthenticator, commands.ConfigurableMixin):

    def __init__(self):
        self.globs = []
        self.res = []

    def get_patterns(self):
        """Return a list of accepted globs."""
        return u' '.join(mask.decode('ascii', 'replace')
                         for mask in self.globs)

    def set_add(self, arg):
        """Add a glob to match against."""
        name, = commands.parseArgs(arg, commands.bytestring)
        self.globs.append(name)
        self._compile(self.globs)

    def set_remove(self, arg):
        """Remove a glob from our list (does nothing if not listed)."""
        name, = commands.parseArgs(arg, commands.bytestring)
        self.globs = [glob for glob in self.globs if glob != name]
        self._compile(self.globs)


class IdentifiedAuthenticator(object):

    """Auth based on nickserv identified status."""

    implements(ibot.IAuthenticator)

    def __init__(self, allowUnknown=False):
        self.allowUnknown = allowUnknown

    def auth(self, user):
        """Check the user object for identified status."""
        if user.identified is None and self.allowUnknown:
            return defer.succeed(True)
        return defer.succeed(bool(user.identified))

    def description(self):
        """human-readable description."""
        if self.allowUnknown:
            return u"Either identified or 'unknown'"
        else:
            return u"an identified user."


class ChanServAuthenticator(object):

    """Authenticate based on minimal access level to a channel.

    Don't forget to add this to factory.handlers too.

    WARNING: this doesn't check linked nicks or hostmasks!

    If there are multiple ChanServAuthenticators in one bot they
    should share a lock. If there is only one, don't pass in a lock
    and a private one will be created.

    We assume others we share the lock with release it in time. If
    they don't auth will deadlock.
    """

    implements(ibot.IAuthenticator, ibot.INoticeHandler)

    levelRegexp = re.compile(r"\d+ +(\d+) +(\S+)")

    def __init__(self, channel, level, maxQueryTime=5.0, clock=None):
        self.channel = channel
        self.level = level
        self.maxQueryTime = maxQueryTime
        self.lock = defer.DeferredLock()
        self.timeoutCall = None
        self.deferred = None
        self.hasLock = False
        self.nick = None
        if clock is None:
            from twisted.internet import reactor as clock
        self.clock = clock

    def description(self):
        """human-readable description."""
        return u'at least access level %d in channel %s' % (
            self.level, self.channel.decode('ascii', 'replace'))

    def auth(self, user):
        """Queue a chanserv query."""
        d = defer.Deferred()
        self.lock.acquire().addCallback(self._lockAcquiredCB, user.bot, d)
        self.nick = user.bot.caseFolder(user.nickname)
        return d

    def _lockAcquiredCB(self, lock, bot, d):
        self.timeoutCall = self.clock.callLater(self.maxQueryTime,
                                                self.timeout)
        self.hasLock = True
        assert self.deferred is None
        self.deferred = d
        bot.getUser("ChanServ").msg("access %s list" % self.channel)

    def timeout(self):
        """Called back when the query timed out."""
        self.hasLock = False
        self.lock.release()
        if self.deferred is not None:
            # deal with callback running code synchronously
            d = self.deferred
            self.deferred = None
            d.callback(False)

    def notice(self, user, source, message):
        """Callback on notice: handle it if it's from chanserv."""
        if not self.hasLock or user.nickname != "ChanServ":
            return
        match = self.levelRegexp.match(message)
        if match is not None:
            level, nick = match.groups()
            if (self.deferred is not None and
                user.bot.caseFolder(nick) == self.nick):
                # got our user.
                # deal with callback running code synchronously
                d = self.deferred
                self.deferred = None
                d.callback(int(level) >= self.level)
                return
        if message.strip() == "-- End of list --":
            # we've been through the list without hitting the nick.
            # make sure we release the lock.
            self.lock.release()
            self.hasLock = False
            self.timeoutCall.cancel()
            if self.deferred is not None:
                # deal with callback running code synchronously
                d = self.deferred
                self.deferred = None
                d.callback(False)
