
"""Essential handlers."""

import re
import datetime
import collections

from zope.interface import implements

from twisted.python import log
from twisted.internet import defer

from mzbot import ibot, ebot, commands, __version__


class BotHandler(object):

    """Very basic bot behaviour."""

    implements(ibot.IUnknownHandler,
               ibot.ISystemMessageHandler,
               ibot.ICTCPQueryHandler,
               ibot.ILoginNickInUseHandler)

    ctcpVersion = None

    def unknown(self, command, args):
        """Ignore some common numerics, log the rest."""
        log.err("received unknown numeric %r (args: %r)" % (command, args))

    def systemMessage(self, bot, user, channel, message):
        """Ignore what I hope are common useless system messages."""

    systemNotice = systemMessage

    def ctcpQuery(self, user, channel, tag, data):
        """Callback for receiving an unknown ctcp query."""
        if tag == 'VERSION':
            if data is not None:
                log.msg("Why did %s send %r with a VERSION query?"
                        % (user.nickname, data))
            if self.ctcpVersion is None:
                ctcpVersion = 'Twisted Mess %s' % __version__
            else:
                ctcpVersion = self.ctcpVersion
            user.ctcpReply('VERSION', ctcpVersion)
        elif tag != 'ACTION':
            log.msg(
                "ctcp: unknown query %r from %r channel %r data %r" %
                (tag, user.nickname, channel, data))

    def loginNickInUse(self, bot, nickname):
        """Append _ to our nickname until it is no longer in use."""
        return nickname + '_'


class WaitForIdentifiedHandler(commands.ConfigurableMixin):

    """Waits for a +e modechange after logging in."""

    implements(ibot.ILoginHandler,
               ibot.IDisconnectHandler,
               ibot.IUserModeHandler)

    configurables = {
        u'mode': (
            u'mode to wait for', 'mode', u'%s', commands.bytestring),
        }

    def __init__(self):
        self.deferred = None
        self.mode = 'e'

    def loggedIn(self, bot):
        """Returns a deferred that fires when we are identified."""
        if self.deferred is not None: # pragma: no cover
            log.err("logged in again before previous was accepted!")
            # We could call it back here but that would probably
            # confuse the bot.
        self.deferred = defer.Deferred()
        return self.deferred

    def userModeChanged(self, bot, modeSet, mode):
        """Continue login if we get mode +e set."""
        if self.deferred and modeSet and mode == self.mode:
            self.deferred.callback(True)
            self.deferred = None

    def disconnected(self, bot, reason=None):
        """Drop our login deferred if we disconnect before getting id'd."""
        # We could call it back here but that would probably confuse
        # the bot that just disconnected.
        self.deferred = None


class WaitForCloakHandler(object):

    """Waits for a cloak to be set after logging in.

    This does not currently care what the cloak actually is.
    """

    implements(ibot.ILoginHandler,
               ibot.IDisconnectHandler,
               ibot.IHostHiddenHandler)

    def __init__(self):
        self.deferred = None

    def loggedIn(self, bot):
        """Returns a deferred that fires when we are identified."""
        if self.deferred is not None: # pragma: no cover
            log.err("logged in again before previous was accepted!")
            # We could call it back here but that would probably
            # confuse the bot.
        self.deferred = defer.Deferred()
        return self.deferred

    def hostHidden(self, bot, vhost):
        """Continue login if we get cloaked."""
        self.deferred.callback(True)
        self.deferred = None

    def disconnected(self, bot, reason=None):
        """Drop our login deferred if we disconnect before getting id'd."""
        # We could call it back here but that would probably confuse
        # the bot that just disconnected.
        self.deferred = None


class SetUserModeHandler(commands.ConfigurableMixin):

    """Set usermode after connecting."""

    implements(ibot.ILoginHandler)

    configurables = {
        u'enabledmodes': (
            u'list of enabled modes', 'enabledModes',
            u'%r', commands.bytestring),
        u'disabledmodes': (
            u'list of disabled modes', 'disabledModes',
            u'%r', commands.bytestring),
        }

    def __init__(self, enabledModes='', disabledModes=''):
        self.enabledModes = enabledModes
        self.disabledModes = disabledModes

    def loggedIn(self, bot):
        """Twiddle umode."""
        bot.mode(self.enabledModes, self.disabledModes)
        return defer.succeed(True)


class AutoJoinHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """Joins channels on signon, and lets you join/part channels on demand.

    Doesn't auto-rejoin on kick, to prevent loops.
    """

    implements(ibot.ILoginHandler, ibot.ISaveable)

    def __init__(self):
        commands.CommandHandler.__init__(self)
        # A dict mapping channel names to keys seems far more obvious.
        # This just follows the approach from the axiom-based original.
        self.channels = []

    def _addChannel(self, channel, key):
        self.channels.append((channel, key))

    def _removeChannel(self, channel):
        self.channels = [(chan, key) for chan, key in self.channels
                         if chan != channel]

    def loggedIn(self, bot):
        """Callback for logging on to the network: autojoin channels."""
        for channel, key in self.channels:
            bot.join(channel, key)
        return defer.succeed(True)

    @commands.optargs(1, commands.bytestring, commands.bytestring)
    def handle_join(self, user, source, channel, key=None):
        """join <channel> [key]: join a channel."""
        self._addChannel(channel, key)
        source.bot.join(channel, key)

    @commands.optargs(0, commands.bytestring)
    def handle_part(self, user, source, channel=None):
        """part <channel>: leave a channel."""
        if channel is None:
            if user is source:
                source.msgTo(user,
                             u'give me an argument or do this in channel')
                return
            channel = source.name
        self._removeChannel(channel)
        channel = source.bot.user.channels.get(channel)
        if channel is None:
            source.msgTo(user, u"I'm not in that channel!")
        else:
            channel.part()

    def get_channels(self):
        """Return a list of autojoined channels."""
        return u', '.join(channel.encode('string-escape').decode('ascii')
                          for channel, key in self.channels)

    def set_add(self, args):
        """add <channel> [key]: Add a channel to autojoin."""
        result = commands.parseOptArgs(args, 1, commands.bytestring,
                                       commands.bytestring)
        if len(result) == 2:
            key = result[1]
        else:
            key = None
        self._addChannel(result[0], key)

    def set_remove(self, args):
        """remove <channel>: no longer autojoin a channel."""
        channel, = commands.parseArgs(args, commands.bytestring)
        self._removeChannel(channel)

    def save(self):
        for channel, key in self.channels:
            if key is None:
                yield 'add', channel
            else:
                yield 'add', '%s %s' % (channel, key)


class PrivMsgLogger(object):

    """Logger for privmsg directed at the bot."""

    implements(ibot.IMessageHandler, ibot.INoticeHandler, ibot.IActionHandler)

    def message(self, user, source, msg):
        """Callback: Log this."""
        if user is source:
            log.msg("user '%s' said '%s'" % (user.nickname, msg))

    notice = message
    action = message


class FloodIgnore(commands.ConfigurableMixin):

    """Ignore users who are flooding us."""

    implements(ibot.IIgnore)

    # time in seconds the flood protection looks at
    floodTime = 10
    # max number of messages from a user in floodTime seconds
    floodMessages = 5

    # seconds a user is ignored after triggering the flood protect
    ignoreTime = 30

    configurables = {
        u'floodtime': (
            u'Number of seconds the flood protection looks at.',
            'floodTime', u'%.0f seconds', commands.number),
        u'floodmessages': (
            u'Number of messages allowed in floodtime seconds.',
            'floodMessages', u'%d messages', commands.integer),
        u'ignoretime': (
            u'Seconds a user triggering the flood protection is ignored.',
            'ignoreTime', u'%.0f seconds', commands.number),
        }

    def __init__(self, _clock=None):
        if _clock is None: # pragma: no cover
            from twisted.internet import reactor as _clock
        self.clock = _clock

    def ignored(self, user, source=None):
        """Run flood checks on a user object."""
        now = self.clock.seconds()
        if (self, 'ignored') in user.data:
            if user.data[self, 'ignored'] > now:
                return True
            del user.data[self, 'ignored']
        emptyQueue = collections.deque(self.floodMessages * [0])
        recent = user.data.setdefault((self, "recent"), emptyQueue)
        while len(recent) > self.floodMessages:
            recent.popleft()
        while len(recent) < self.floodMessages:
            recent.appendleft(0)
        recent.append(now)
        if now - recent.popleft() < self.floodTime:
            if source is not None:
                source.msgTo(
                    user,
                    u"you're talking too much, I'm ignoring you for a bit")
            user.data[self, "ignored"] = now + self.ignoreTime
            return True
        return False


class CommandHandlerHolderBase(commands.ConfigurableMixin):

    """This glues ICommandHandlers to IMessageHandler.

    Subclass and implement the get{Global,Restricted}{Commands,Auths} methods
    that define how commands/auths are hooked up.

    Public vars:

    globalCommands: nested dict for global CommandHandler configuration.
    restrictedCommands: same per-channel.
    globalAuths: same for global auths.
    restrictedAuths: same per-channel.
    prefix: single char like '!' for '!seen'-like behaviour, or None.
    """

    implements(ibot.IMessageHandler)

    # character used for talking to the bot without using nick (or None)
    prefix = None

    def __init__(self):
        object.__init__(self)
        self.globalCommands = {}
        self.restrictedCommands = {}
        self.globalAuths = {}
        self.restrictedAuths = {}

    def getGlobalCommands(self):
        """Return an iterator of (name, handler)."""
        return ()

    def getRestrictedCommands(self):
        """Return an iterator of (channel, name, handler)."""
        return ()

    def getGlobalAuths(self):
        """Return an iterator of (name, commandname, authenticator)."""
        return ()

    def getRestrictedAuths(self):
        """Return an iterator of (channel, name, commandname, auth)."""
        return ()

    def updateCommands(self):
        """Call when the handler ruleset has changed."""
        self.globalCommands = dict(self.getGlobalCommands())
        self.restrictedCommands = {}
        for channel, name, handler in self.getRestrictedCommands():
            try:
                commandDict = self.restrictedCommands[channel]
            except KeyError:
                commandDict = self.globalCommands.copy()
                self.restrictedCommands[channel] = commandDict
            if handler is None:
                commandDict.pop(name, None)
            else:
                commandDict[name] = handler

    def updateAuths(self):
        """Call when the auth ruleset has changed."""
        self.globalAuths = {}
        for name, command, auth in self.getGlobalAuths():
            self.globalAuths[name, command] = auth
        self.restrictedAuths = {}
        for channel, name, command, auth in self.getRestrictedAuths():
            try:
                auths = self.restrictedAuths[channel]
            except KeyError:
                auths = self.restrictedAuths[channel] = self.globalAuths.copy()
            if auth is None:
                auths.pop((name, command), None)
            else:
                auths[name, command] = auth

    def handle(self, user, source, handler, handlerName, command, myArgs):
        """Handle a command."""
        if user is source:
            auth = self.restrictedAuths.get(None, self.globalAuths)
        else:
            auth = self.restrictedAuths.get(source.name, self.globalAuths)
        # check for auths for this handler
        authDeferreds = []
        for level in (None, command):
            # authenticate for both handler and command
            authenticator = auth.get((handlerName, level))
            if authenticator is not None:
                authDeferreds.append(authenticator.auth(user))
        authed = defer.gatherResults(authDeferreds)
        def actuallyHandle(authed):
            """Run the actual handler if authed."""
            for result in authed:
                if not result:
                    return source.msgTo(
                        user, u'Sorry, you are not allowed to do that.')
            return handler.handle(user, source, command, myArgs)
        def logFailure(err):
            """Tell the user about the failure and log an error."""
            source.msgTo(
                user, u'That made me traceback, tell my owner to fix me')
            log.err(err)
        authed.addCallback(actuallyHandle)
        authed.addErrback(logFailure)
        return authed

    def message(self, user, source, command):
        """Check if a message is meant for us and handle it."""
        if user is not source:
            # ignore stuff not targeted at us (/msg or "us:"-prefixed)
            if self.prefix is None:
                pattern = r'^%s[,.;:] *(.*)' % (user.bot.user.nickname,)
            else:
                pattern = r'^(?:%s[,.;:] *|%s)(.*)' % (
                    user.bot.user.nickname, self.prefix)
            match = re.match(pattern, command)
            if match is None:
                return
            command = match.group(1)
            # sourceName is what we check our handler and auth dicts for
            sourceName = source.name
        else:
            if self.prefix is not None and command.startswith(self.prefix):
                command = command[len(self.prefix):]
            sourceName = None
        for ignorer in user.bot.handlers.get(ibot.IIgnore, ()):
            if ignorer.ignored(user, source):
                return

        # TODO: allow customizing the encoding per source?
        try:
            command = command.decode('utf-8')
        except UnicodeError:
            source.msgTo(user, u'%r is not utf-8' % (command,))
            return

        # command is the first word, myArgs is the rest or ""
        command, sep, myArgs = command.partition(u' ')
        command = command.lower()
        handlers = self.restrictedCommands.get(sourceName, self.globalCommands)

        # see if it's directly targeted at a handler
        handler = handlers.get(command)
        if handler is not None:
            # directedCommand is the second word, directedArgs the rest or ""
            directedCommand, sep, directedArgs = myArgs.partition(u' ')
            directedCommand = directedCommand.lower()
            if directedCommand in handler.commandset:
                # it was aimed at this one, check auth
                return self.handle(user, source, handler,
                                   command, directedCommand, directedArgs)

        candidates = [
            (name, handler)
            for name, handler in handlers.iteritems()
            if handler is not None and command in handler.commandset]
        if len(candidates) == 1:
            name, handler = candidates[0]
            # it was aimed at this one, check auth
            return self.handle(user, source, handler, name, command, myArgs)
        elif len(candidates) > 1:
            source.msgTo(
                user, u"multiple matches for '%s', try one of %s" %
                (command, ", ".join("'%s %s'" % (name, command)
                                    for name, handler in candidates)))
        # ignore if there's no handler at all (to prevent message loops)

    def get_prefix(self):
        """Get the command shortcut."""
        if self.prefix is None:
            return u'None'
        else:
            return unicode(self.prefix.encode('string-escape'))

    def set_prefix(self, args):
        """Set the command shortcut. Set to 'None' for no prefix."""
        prefix, = commands.parseArgs(args, commands.bytestring)
        if prefix.lower() == 'none':
            self.prefix = None
        else:
            self.prefix = prefix


class CommandHandlerHolder(CommandHandlerHolderBase):

    """Master handler for privmsg's. Plug commandhandlers into me."""

    def __init__(self):
        CommandHandlerHolderBase.__init__(self)
        self._globalCommands = {}
        self._restrictedCommands = {}
        self._globalAuths = {}
        self._restrictedAuths = {}

    _missing = object()

    def addHandler(self, name, handler, channel=_missing, auth=_missing):
        """Add an ICommandHandler.

        'name' is the string name used for calling this handler.
        'handler' is an ICommandHandler.
        'channel' is the name of the channel this handler is restricted to.
        'auth' is an IAuthenticator guarding the entire handler.

        If 'channel' is omitted, the handler is available on all channels.
        If 'channel' is None, it is available only in privmsg.
        If 'auth' is missing, the handler doesn't need authentication.
        """
        if auth is not self._missing:
            self.addAuth(name, auth, channel=channel)
        if channel is self._missing:
            self._globalCommands[name] = handler
        else:
            self._restrictedCommands.setdefault(channel, {})[name] = handler
        self.updateCommands()

    def addAuth(self, name, auth, commands=(None,), channel=_missing):
        """Add an IAuthenticator.

        'name' is the name of the handler protected.
        'auth' is the IAuthenticator.
        'commands' is a sequence of specific commands to protect.
        'channel' is the name of the channel to protect on.

        If 'commands' is missing, all commands are protected.
        If 'channel' is missing, all channels are protected.
        """
        if channel is self._missing:
            target = self._globalAuths
        else:
            target = self._restrictedAuths.setdefault(channel, {})
        for command in commands:
            target.setdefault(name, {})[command] = auth
        self.updateAuths()

    def getGlobalCommands(self):
        """Return base rules."""
        return self._globalCommands.iteritems()

    def getRestrictedCommands(self):
        for channel, overrides in self._restrictedCommands.iteritems():
            for name, handler in overrides.iteritems():
                yield channel, name, handler

    def getGlobalAuths(self):
        for name, auths in self._globalAuths.iteritems():
            for command, auth in auths.iteritems():
                yield name, command, auth

    def getRestrictedAuths(self):
        for channel, overrides in self._restrictedAuths.iteritems():
            for name, auths in overrides.iteritems():
                for command, auth in auths.iteritems():
                    yield channel, name, command, auth


class HelpHandler(commands.CommandHandler):

    """Give info on available commands."""

    def __init__(self, commandHandlerHolder, authenticator=None):
        commands.CommandHandler.__init__(self)
        self.holder = commandHandlerHolder
        self.auth = authenticator

    def handle_help(self, user, source, module):
        """no arguments: list modules. Else list commands in a module."""
        if user is not source:
            sourceName = source.name
        else:
            sourceName = None
        handlers = self.holder.restrictedCommands.get(
            sourceName, self.holder.globalCommands)
        if not module:
            return self._moduleList(user, source, handlers)
        handler = handlers.get(module)
        if handler is None:
            # try looking up a module offering this command and using that.
            modules = [
                (mod, handler) for mod, handler in handlers.iteritems()
                if handler is not None and module in handler.commandset]
            if not modules:
                source.msgTo(
                    user, u"I don't have that module or command, try 'help'")
                return
            if len(modules) > 1:
                source.msgTo(
                    user,
                    u"that command is offered by the following modules: %s. "
                    u"You can ask for further help on those." %
                    u", ".join(sorted(mod for mod, handler in modules)))
                return
            module, handler = modules[0]
        # check if we're going to document auth-protected stuff
        if self.auth is None:
            return self._moduleHelp(user, source, module, handler, False)
        else:
            d = self.auth.auth(user)
            @d.addCallback
            def _(result):
                return self._moduleHelp(user, source, module, handler, result)
            return d

    def _moduleList(self, user, source, handlers):
        """Send a module list."""
        if user is not source:
            source.msgTo(user, u"module list is being /msg'd")
        user.msg(u"available modules:")
        for name, module in sorted(handlers.iteritems()):
            if module is not None:
                if module.__doc__:
                    user.msg(u'%s: %s' % (
                            commands.bold(name),
                            module.__doc__.split("\n")[0]))
                else:
                    user.msg(commands.bold(name))
        user.msg(u"for further help, 'help modulename'")

    def _moduleHelp(self, user, source, module, handler, authed):
        """Send help on a single module."""
        if user is not source:
            source.msgTo(
                user, u"help on module '%s' is being /msg'd" % module)
            auth = self.holder.restrictedAuths.get(
                source.name, self.holder.globalAuths)
        else:
            auth = self.holder.restrictedAuths.get(
                None, self.holder.globalAuths)
        helpProvider = ibot.ICommandHelpProvider(handler, None)
        if helpProvider is None:
            user.msg(
                u'undocumented, commands: %s' % (
                    u', '.join(handler.commandset),))
        else:
            for command in sorted(handler.commandset):
                reply = '%s: %s' % (
                    commands.bold(command),
                    helpProvider.getCommandDescription(command))
                if authed:
                    # attempt to document the authenticator if there is one
                    authenticator = auth.get((module, command))
                    if authenticator is not None:
                        reply = (u'%s (authentication: %s)' %
                                 (reply, authenticator.description()))
                user.msg(reply)
        # document entire handler's authenticator if available:
        authenticator = auth.get((module, None))
        if authed and authenticator is not None:
            user.msg(u"entire module needs authentication: %s" %
                     authenticator.description())


class PasswordAuthenticatorHandler(commands.CommandHandler):

    """Authenticator allowing username/password based logins."""

    implements(ibot.IAuthenticator)

    def __init__(self, passwords):
        """Users is a dict mapping username to password."""
        commands.CommandHandler.__init__(self)
        self.passwords = passwords

    def auth(self, user):
        """Return true if the user successfully logged in."""
        return defer.succeed(self in user.data)

    def description(self):
        """Return a human-readable description."""
        return "authenticated based on a list"

    @commands.onlyInPrivmsg
    @commands.args(commands.string, commands.string)
    def handle_identify(self, user, source, username, password):
        """identify username password: make yourself known."""
        if username in self.passwords and password == self.passwords[username]:
            source.msgTo(user, u"you are now identified")
            user.data[self] = True
        else:
            source.msgTo(user, u"Authentication failed!")


class AdminHandlerBase(commands.CommandHandler):

    """Add and remove handlers for an ircbot."""

    def _getConfigurable(self, user, source, handlername): # pragma: no cover
        """Helper function to get an IConfigurable by name."""
        raise NotImplementedError(self._getConfigurable)

    @commands.args(commands.string)
    def handle_gettables(self, user, source, handlername):
        """gettables handler: list available configuration settings."""
        configurable = self._getConfigurable(user, source, handlername)
        if configurable is not None:
            source.msgTo(user, u', '.join(
                    '%s: %s' % entry
                    for entry in configurable.gettables().iteritems()))

    @commands.args(commands.string)
    def handle_settables(self, user, source, handlername):
        """settables handler: list available configuration settings."""
        configurable = self._getConfigurable(user, source, handlername)
        if configurable is not None:
            source.msgTo(user, u', '.join(
                    '%s: %s' % entry
                    for entry in configurable.settables().iteritems()))

    @commands.args(commands.string, commands.string)
    def handle_getconfig(self, user, source, handlername, key):
        """getconfig handler key: get a configuration setting."""
        configurable = self._getConfigurable(user, source, handlername)
        if configurable is not None:
            if key not in configurable.gettables():
                source.msgTo(user, u'no such config key')
            else:
                source.msgTo(user, configurable.get(key))

    @commands.args(commands.string, commands.string, commands.longstring)
    def handle_setconfig(self, user, source, handlername, key, value):
        """setconfig handler key ...: set a configuration setting."""
        configurable = self._getConfigurable(user, source, handlername)
        if configurable is not None:
            if key not in configurable.settables():
                source.msgTo(user, u'no such config key')
            else:
                try:
                    configurable.set(key, value)
                except ebot.InvalidArgument, e:
                    source.msgTo(user, unicode(e))
                else:
                    source.msgTo(user, u'set')


class RawLogFile(commands.ConfigurableMixin):

    implements(ibot.IRawLogHandler)

    configurables = {
        u'filename': (
            u'filename', 'filename', u'%s', commands.bytestring),
        }

    def __init__(self):
        self.file = None

    def set_filename(self, args):
        """set <filename>: Set the path to write to."""
        filename, = commands.parseArgs(args, commands.bytestring)
        if self.file is not None:
            self.file.close()
        self.file = open(filename, 'a')

    def rawlog(self, bot, out, line):
        if self.file is not None:
            self.file.write('%s %s %r\n' % (datetime.datetime.utcnow(),
                                            '>>>' if out else '<<<',
                                            line))
            self.file.flush()


class SpecialFeaturesHandler(commands.CommandHandler):

    """Provides some special features."""

    @commands.args(commands.bytestring, commands.longstring)
    def handle_say(self, user, source, target, message):
        """say <channel> <message>: say something."""
        target = user.bot.user.channels.get(target)
        if target is None:
            source.msgTo(user, u'no such channel')
            return
        target.msg(message)

    @commands.args(commands.bytestring, commands.longstring)
    def handle_do(self, user, source, target, message):
        """do <channel> <message>: do something (using /me)."""
        target = user.bot.user.channels.get(target)
        if target is None:
            source.msgTo(user, u'no such channel')
            return
        target.me(message)

    @commands.args(commands.bytestring, commands.bytestring,
                   commands.longbytestring)
    def handle_remusr(self, user, source, target, username, message):
        """remusr <channel> <nick> <reason>: remove a user from the channel."""
        targetchannel = user.bot.user.channels.get(target)
        if targetchannel is None:
            source.msgTo(user, u'no such channel')
            return
        targetuser = targetchannel.users.get(username)
        if targetuser is None:
            source.msgTo(user, u'no such user')
            return
        targetchannel.remove(targetuser, message)


class ChanServOpHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """This has a list of channels to op on join, and offers op commands."""

    implements(ibot.IJoinHandler)

    # XXX this wants to use bytes when sending to chanserv, not unicode

    def __init__(self, channels=''):
        commands.CommandHandler.__init__(self)
        self.channels = channels

    def joined(self, channel):
        """Callback: auto-op if requested."""
        if channel.name in self.channels.split():
            channel.bot.getUser("ChanServ").msg(u"op %s" % channel.name)

    @commands.notInPrivmsg
    @commands.optargs(0, commands.bytestring)
    def handle_op(self, user, channel, target=None):
        """Ask chanserv for +o."""
        if target is None:
            channel.bot.getUser("ChanServ").msg(u"op %s" % (channel.name,))
        else:
            channel.bot.getUser("ChanServ").msg(
                u"op %s %s" % (channel.name, target))

    @commands.notInPrivmsg
    @commands.args()
    def handle_deop(self, user, channel):
        """Ask chanserv for -o."""
        channel.bot.getUser("ChanServ").msg(u"op %s -%s" % (
                channel.name, channel.bot.user.nickname))

    @commands.notInPrivmsg
    @commands.args(commands.bytestring)
    def handle_giveop(self, user, channel, target):
        """giveop <target>: sets mode +o on target."""
        channel.mode(True, 'o', target)

    def get_channels(self):
        """List of auto-opped channels."""
        return u', '.join(channel.encode('string-escape').decode('ascii')
                          for channel in self.channels.split())

    def set_addchannel(self, args):
        """Add a channel to the autoop list."""
        channel, = commands.parseArgs(args, commands.bytestring)
        self.channels += ' ' + channel

    def set_removechannel(self, args):
        """Remove a channel from the autoop list."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        if channel not in channels:
            raise ebot.InvalidArgument('No such channel')
        channels.remove(channel)
        self.channels = ' '.join(channels)


class IgnoreCommandHandler(commands.CommandHandler,
                           commands.ConfigurableMixin):

    """Ignore people when told to."""

    implements(ibot.IIgnore)

    configurables = {
        u'ignoretime': (
            u'the number of seconds we ignore.', 'ignoreTime',
            u'%d seconds', commands.integer),
        }

    def __init__(self, ignoreTime=600, _clock=None):
        commands.CommandHandler.__init__(self)
        self.ignoreTime = ignoreTime
        if _clock is None: # pragma: no cover
            from twisted.internet import reactor as _clock
        self.clock = _clock

    @commands.args(commands.bytestring)
    def handle_ignore(self, user, source, targetName):
        """ignore <target>: Ignore a user."""
        target = user.bot.users.get(targetName)
        if target is None:
            source.msgTo(user, u'I do not see that user!')
            return
        target.data[self] = self.clock.seconds() + self.ignoreTime
        source.msgTo(user, u'done.')

    @commands.args(commands.bytestring)
    def handle_unignore(self, user, source, targetName):
        """unignore <target>: Unignore a user."""
        target = user.bot.users.get(targetName)
        if target is None:
            source.msgTo(user, u'I do not see that user!')
            return
        ignore = target.data.pop(self, None)
        if ignore is None:
            source.msgTo(user, u'I was not ignoring that user!')
        else:
            source.msgTo(user, u'done.')

    def ignored(self, user, source=None):
        """Check if we were told to ignore this one."""
        ignore = user.data.get(self)
        return ignore is not None and ignore > self.clock.seconds()
