
"""query various game servers."""

import struct

from twisted.internet import protocol, defer, reactor
from twisted.python import log

from mzbot import commands


ET_QUERY_PORT = 27960

UT_QUERY_PORT = 7778
UT_QUERY_VERSION = "\x80"


class ColoredString(object):

    """Representation of a colored string returned from an ut server."""

    def __init__(self, string):
        self.string = string

    def __repr__(self):
        return "ColoredString <%s>" % self.string


class EnemyTerritoryProtocol(protocol.DatagramProtocol):

    """Enemy Territory proto deciphered from the qstat source."""

    def __init__(self, address, port=ET_QUERY_PORT, timeout=5):
        self.address = address
        self.port = port
        self.timeoutSecs = timeout
        self.deferreds = []
        self.status = self.info = self.timeout = None
        self.players = []

    def startProtocol(self):
        """Start listening."""
        self.transport.connect(self.address, self.port)

    def query(self):
        """Run a query, return a deferred firing with results."""
        if not self.deferreds:
            # do a query
            self.status = self.info = None
            self.players = []
            self.transport.write('\xff\xff\xff\xffgetinfo\n')
            self.transport.write('\xff\xff\xff\xffgetstatus\n')
            self.timeout = reactor.callLater(self.timeoutSecs, self.errback)
        d = defer.Deferred()
        self.deferreds.append(d)
        return d

    def datagramReceived(self, data, address):
        """Parse data from the server."""
        infoHeader = '\xff\xff\xff\xffinfoResponse\n\\'
        statusHeader = '\xff\xff\xff\xffstatusResponse\n\\'
        if data.startswith(infoHeader):
            data = data[len(infoHeader):].split('\n')
            if len(data) != 1:
                log.msg('huh, len(%r) != 1' % data)
            data = data[0].split('\\')
            self.info = result = {}
            while data:
                if len(data) < 2:
                    log.msg('odd number of entries')
                    break
                key = data.pop(0)
                val = data.pop(0)
                if key in result:
                    log.msg('second entry for key %r ignored' % key)
                    continue
                result[key] = val
        elif data.startswith(statusHeader):
            data = data[len(statusHeader):].split('\n')
            players = data[1:]
            data = data[0].split('\\')
            self.status = result = {}
            while data:
                if len(data) < 2:
                    log.msg('odd number of entries')
                    break
                key = data.pop(0)
                val = data.pop(0)
                if key in result:
                    log.msg('second entry for key %r ignored' % key)
                    continue
                result[key] = val
            for player in players:
                if player:
                    score, ping, name = player.split(' ', 2)
                    strippedName = []
                    jump = False
                    for char in name:
                        if jump:
                            jump = False
                        elif char == '^':
                            jump = True
                        else:
                            strippedName.append(char)
                    self.players.append({
                            'name': ''.join(strippedName),
                            'ping': int(ping),
                            'score': int(score),
                            })
        else:
            log.err('weird data %r' % data)

        if self.status is not None and self.info is not None:
            for d in self.deferreds:
                d.callback((self.info, self.status, self.players))
            if self.timeout is not None:
                self.timeout.cancel()
            self.deferreds = []

    def errback(self):
        """Timeout all running queries."""
        deferreds = self.deferreds
        self.deferreds = []
        for d in deferreds:
            defer.timeout(d)


class EnemyTerritoryHandler(commands.CommandHandler):

    """Query an enemy territory server."""

    def __init__(self, host, port=ET_QUERY_PORT, _reactor=None):
        commands.CommandHandler.__init__(self)
        self.host = host
        self.port = port
        self.server = {}
        self.game = {}
        self.players = []
        self.timeouts = 0
        if _reactor is None:
            from twisted.internet import reactor as _reactor

        def _connect(address):
            self.proto = EnemyTerritoryProtocol(address, self.port)
            _reactor.listenUDP(0, self.proto)
        _reactor.resolve(self.host).addCallback(_connect)

    @defer.inlineCallbacks
    def query(self):
        """Update our copy of the game state."""
        try:
            server, game, players = yield self.proto.query()
        except defer.TimeoutError:
            self.timeouts += 1
            return
        self.timeouts = 0
        for player in players:
            player['name'] = unicode(
                player['name'], 'ascii', 'replace').encode('ascii', 'replace')
        self.server = server
        self.game = game
        self.players = players

    @defer.inlineCallbacks
    def tryQuery(self, user, source):
        """Try to update our game state, provide feedback."""
        probablyDown = self.timeouts >= 3
        if probablyDown:
            source.msgTo(user, 'server down, I think')
        yield self.query()
        if self.timeouts >= 3:
            defer.returnValue(False)
        else:
            if probablyDown:
                source.msgTo(user, 'I was wrong, the server seems to be up')
            defer.returnValue(True)

    @commands.args()
    @defer.inlineCallbacks
    def handle_et(self, user, source):
        """Get some general statistics."""
        result = yield self.tryQuery(user, source)
        if not result:
            return
        counterHi = None
        terrHi = None
        if (self.game['Players_Allies'] == '(None)' or
            self.game['Players_Axis'] == '(None)'):
            source.msgTo(user, 'not enough players')
            return
        counter = [int(i) for i in self.game['Players_Allies'].split()]
        terr = [int(i) for i in self.game['Players_Axis'].split()]
        for i, player in enumerate(self.players):
            if i in counter and (
                counterHi is None or player['score'] > counterHi['score']):
                counterHi = player
            elif i in terr and (
                terrHi is None or player['score'] > terrHi['score']):
                terrHi = player
        source.msgTo(
            user, 'currently %s/%s players are playing on %s' % (
                self.server['clients'],
                self.server['sv_maxclients'],
                self.server['mapname']))
        if counterHi is not None:
            source.msgTo(
                user, 'counter-terrorist leader: %s with %s' %
                (counterHi['name'], counterHi['score']))
        if terrHi is not None:
            source.msgTo(
                user, 'terrorist leader: %s with %s' %
                (terrHi['name'], terrHi['score']))

    @commands.args()
    @defer.inlineCallbacks
    def handle_players(self, user, source):
        """Get a list of players currently online."""
        result = yield self.tryQuery(user, source)
        if not result:
            return
        if self.players:
            source.msgTo(user, ', '.join(
                    player['name'] for player in self.players))
        else:
            source.msgTo(user, 'no players')


class UnrealTournamentProtocol(protocol.DatagramProtocol):

    """Unreal Tournament 2k4 protocol."""

    # list of list of tuples (isStr, desc), if it's not a str it's an int
    serverInfoTypes = [
        (False, "id"),
        (True, "ip"),
        (False, "gameport"),
        (False, "queryport"),
        (True, "name"),
        (True, "map"),
        (True, "type"),
        (False, "players"),
        (False, "maxplayers"),
        (False, "ping"),
        (False, "flags"),
        (True, "skill"),
        ]

    playerInfoTypes = [
        (False, "id"),
        (True, "name"),
        (False, "ping"),
        (False, "score"),
        (False, "stats"),
        ]

    gameInfoList = set(("Mutator",))

    def __init__(
        self, address, port=UT_QUERY_PORT, version=UT_QUERY_VERSION,
        timeout=5):
        self.address = address
        self.port = port
        self.initialVersion = version
        self.timeoutSecs = timeout
        self.serverVersion = None
        self.clearResults()

    def clearResults(self):
        """Throw away our game state."""
        self.serverInfo = {}
        self.gameInfo = {}
        self.playerInfo = []
        self.deferreds = []
        self.timeout = None

    def startProtocol(self):
        """Start listening."""
        self.transport.connect(self.address, self.port)

    def query(self):
        """Ask the server for game state. Return a deferred for results."""
        if not self.deferreds:
            # do a query
            if self.serverVersion is None:
                version = self.initialVersion
            else:
                version = self.serverVersion
            self.transport.write(version + 4*"\0")
            self.timeout = reactor.callLater(self.timeoutSecs, self.errback)
        d = defer.Deferred()
        self.deferreds.append(d)
        return d

    def parseString(self, data):
        """returns a tuple result, remainingdata."""
        length = struct.unpack('B', data[0])[0]
        color = length > 128
        if color:
            length -= 128
            length *= 2
        if len(data) <= length:
            raise ValueError
        # we don't need the null terminator
        string = data[1:length]
        if color:
            string = ColoredString(string)
        return string, data[length + 1:]

    def datagramReceived(self, data, address):
        """Parse server data."""
        if self.serverVersion not in (None, data[0]):
            log.msg(
                'version mismatch (%r, %r)' % (self.serverVersion, data[0]))
        self.serverVersion = data[0]
        if data[1:4] != "\0\0\0":
            log.err("weird data %r" % data[1:4])
        dataType = struct.unpack('B', data[4])[0]
        data = data[5:]

        if not data:
            log.err("no data (dataType %s)" % dataType)
            return

        if dataType == 0:
            # server info
            for isStr, desc in self.serverInfoTypes:
                if isStr:
                    try:
                        self.serverInfo[desc], data = self.parseString(data)
                    except ValueError:
                        log.err()
                        break
                else:
                    if len(data) < 4:
                        log.err("ran out of data")
                        break
                    self.serverInfo[desc] = struct.unpack("<i", data[:4])[0]
                    data = data[4:]

                # huh, \0\0 happens...
                if data == '\0\0' or not data:
                    break
            else:
                log.err("more data than I can handle (%r)" % data)
            # query for more stuff
            self.transport.write(self.serverVersion + 3*"\0" + "\x03")
        elif dataType == 1:
            # game info
            while True:
                key, data = self.parseString(data)
                value, data = self.parseString(data)
                if key not in self.gameInfoList:
                    if key in self.gameInfo:
                        log.err("duplicate %s (%s, %s)" %
                                (key, value, self.gameInfo[key]))
                    self.gameInfo[key] = value
                else:
                    self.gameInfo.setdefault(key, []).append(value)
                if not data:
                    break
            # if the playerinfo is already there or it's not expected:
            if self.playerInfo or not self.serverInfo.get('players', 0):
                self.callback()
        elif dataType == 2:
            if 'players' not in self.serverInfo:
                log.err('no player count')
                self.serverInfo['players'] = 8
            for i in range(self.serverInfo['players']+1):
                if not data:
                    break
                result = {}
                self.playerInfo.append(result)
                for isStr, desc in self.playerInfoTypes:
                    if isStr:
                        try:
                            result[desc], data = self.parseString(data)
                        except ValueError:
                            log.err()
                            break
                    else:
                        if len(data) < 4:
                            log.err("ran out of data")
                            break
                        result[desc] = struct.unpack("<i", data[:4])[0]
                        data = data[4:]
            if self.gameInfo:
                self.callback()
        else:
            log.err("don't know how to handle response type %s" % dataType)

    def callback(self):
        """Update outstanding game status requests."""
        for d in self.deferreds:
            d.callback((self.serverInfo, self.gameInfo, self.playerInfo))
        if self.timeout is not None:
            self.timeout.cancel()
        self.clearResults()

    def errback(self):
        """Timeout all running queries."""
        deferreds = self.deferreds
        self.deferreds = []
        for d in deferreds:
            defer.timeout(d)
        self.clearResults()


def utIsOnRed(player):
    """Given a player dict return True if the player is on the red team."""
    return bool(player['stats'] & 1 << 29)

def utIsOnBlue(player):
    """Given a player dict return True if the player is on the blue team."""
    return bool(player['stats'] & 1 << 30)


class UnrealTournamentHandler(commands.CommandHandler):

    """Query an ut2k4 server."""

    def __init__(self, host, port=UT_QUERY_PORT, _reactor=None):
        commands.CommandHandler.__init__(self)
        self.host = host
        self.port = port
        self.server = {}
        self.game = {}
        self.players = []
        self.timeouts = 0
        if _reactor is None:
            from twisted.internet import reactor as _reactor
        def _connect(address):
            self.proto = UnrealTournamentProtocol(
                address, self.port, UT_QUERY_VERSION)
            _reactor.listenUDP(0, self.proto)
        _reactor.resolve(self.host).addCallback(_connect)

    @defer.inlineCallbacks
    def query(self):
        """Update our copy of the game state."""
        try:
            server, game, players = yield self.proto.query()
        except defer.TimeoutError:
            self.timeouts += 1
            return
        self.timeouts = 0
        for player in players:
            player['name'] = unicode(
                player['name'], 'ascii', 'replace').encode('ascii', 'replace')
        self.server = server
        self.game = game
        self.players = players

    @defer.inlineCallbacks
    def tryQuery(self, user, source):
        """Try to update our game state, provide feedback."""
        probablyDown = self.timeouts >= 3 or 'players' not in self.server
        if probablyDown:
            source.msgTo(user, 'server down, I think')
        yield self.query()
        if self.timeouts >= 3 or 'players' not in self.server:
            defer.returnValue(False)
        else:
            if probablyDown:
                source.msgTo(user, 'I was wrong, the server seems to be up')
            defer.returnValue(True)

    @commands.args()
    @defer.inlineCallbacks
    def handle_ut(self, user, source):
        """Get some general statistics."""
        result = yield self.tryQuery(user, source)
        if not result:
            return
        redHi = None
        blueHi = None
        for player in self.players:
            if utIsOnRed(player) and (
                redHi is None or player['score'] > redHi['score']):
                redHi = player
            elif utIsOnBlue(player) and (
                blueHi is None or player['score'] > blueHi['score']):
                blueHi = player
            else:
                log.err("unknown color in %r" % player)
        source.msgTo(
            user, 'currently %s/%s players are playing on %s' %
            (self.server['players'],
             self.server['maxplayers'],
             self.server['map']))
        if redHi is not None:
            source.msgTo(
                user, 'red leader: %s with %s' %
                (redHi['name'], redHi['score']))
        if blueHi is not None:
            source.msgTo(
                user, 'blue leader: %s with %s' %
                (blueHi['name'], blueHi['score']))

    @commands.args()
    @defer.inlineCallbacks
    def handle_players(self, user, source):
        """Get a list of players currently online."""
        result = yield self.tryQuery(user, source)
        if not result:
            return
        if self.players:
            result = []
            for player in self.players:
                name = player['name']
                if utIsOnRed(player):
                    name = commands.colorize(name, 'red')
                elif utIsOnBlue(player):
                    name = commands.colorize(name, 'blue')
                result.append(name)
            source.msgTo(user, ', '.join(result))
        else:
            source.msgTo(user, 'no players')

    @commands.args(commands.bytestring)
    @defer.inlineCallbacks
    def handle_playerinfo(self, user, source, query):
        """playerinfo <player>: Get info on <player>"""
        result = yield self.tryQuery(user, source)
        if not result:
            return
        query = query.lower()
        for player in self.players:
            if player['name'].lower() == query:
                if utIsOnRed(player):
                    player['color'] = "red"
                elif utIsOnBlue(player):
                    player['color'] = "blue"
                else:
                    player['color'] = "unknown"
                source.msgTo(
                    user, '%(name)s is on team %(color)s, score %(score)s, '
                    'ping %(ping)s' % player)
                break
        else:
            source.msgTo(user, 'No such player')
