
"""A couple of handlers that are only useful for some gentoo channels."""


import os
import re
import fnmatch
import time
import datetime

from zope.interface import implements

from twisted.python import log
from twisted.internet import defer, reactor
from twisted.web import client

from xml.etree import cElementTree as etree

import BeautifulSoup

from mzbot import commands, ibot


class InviteHandler(commands.CommandHandler):

    """Invite people selectively."""

    # TODO: cancel doesn't do the right thing if hostmask changes after the +I

    implements(ibot.IUserJoinedHandler)

    def __init__(self, target='#gentoo', delay=60.0):
        """target: channel to invite into, delay: seconds to keep the mode."""
        commands.CommandHandler.__init__(self)
        self.target = target
        self.delay = delay

    @commands.optargs(0, commands.string)
    def handle_invite(self, user, source, target=None):
        """invite <nick>: invite <nick> into the target channel."""
        if target is None:
            target = user
        else:
            target = user.bot.users.get(target)
            if target is None:
                source.msgTo(user, "I don't see that user!")
                return
        mode = '%s!*@*' % (target.nickname,)
        user.bot.user.channels[self.target].mode(True, 'I', mode)
        target.data[self] = reactor.callLater(
            self.delay, self.cancel, target, mode)
        source.msgTo(target, 'now do: "/join %s"' % (self.target,))

    def userJoined(self, user, channel):
        """remove modes and delayedCall if we invited user into channel."""
        if channel.name != self.target:
            return
        delayedCall = user.data.get(self)
        if delayedCall is None:
            return
        # user joined the channel we invited it into within the limit
        delayedCall.cancel()
        self.cancel(user, '%s!*@*' % (user.nickname,))

    def cancel(self, user, mode):
        """Cancel the invitation of user."""
        try:
            del user.data[self]
        except KeyError:
            log.err("huh, tried to cancel an invite we don't remember")
        user.bot.user.channels[self.target].mode(False, 'I', mode)


class SpamHandler(object):

    """Talk to people whenever they say anything at all."""

    implements(ibot.IMessageHandler)

    # XXX TODO make this configurable

    def __init__(self, channel=None, spam=None):
        self.channel = channel
        self.spam = spam

    def message(self, user, source, message):
        """Figure out if we should spam."""
        if (user is source or source.name != self.channel or
            self.channel in user.chanmodes['o'] or
            self.channel in user.chanmodes['v'] or
            self in user.data):
            return
        # Remember we already spammed this user.
        user.data[self] = True
        # XXX ick, ignore potential CommandHandler triggers
        if re.match(
            '^(%s[,.;:] ?|!)(.*)' % (
                user.bot.user.nickname,), message) is not None:
            return
        source.msgTo(user, self.spam)


class TorHandler(object):

    """Notice a channel when tor users join another channel."""

    implements(ibot.IUserJoinedHandler)

    def __init__(self, mask='*!*@tor/session/*', joinChannel='#gentoo',
                 noticeChannel='#gentoo-ops',
                 message=u'tor user who might need voice'):
        self.mask = mask
        self.joinChannel = joinChannel
        self.noticeChannel = noticeChannel
        self.message = message
        self.regexp = re.compile(fnmatch.translate(self.mask))

    def userJoined(self, user, channel):
        """Schedule a check if needed."""
        if channel.name != self.joinChannel:
            return
        reactor.callLater(5, self.check, user)

    @defer.inlineCallbacks
    def check(self, user):
        """Check if we should announce."""
        channel = user.channels.get(self.joinChannel)
        # no need to warn about a user that already parted or is voiced
        if (channel is None or self.joinChannel in user.chanmodes['v'] or
            self.joinChannel in user.chanmodes['o']):
            return
        mask = yield user.hostmask()
        if self.regexp.match('%s!%s@%s' % mask):
            target = user.bot.user.channels[self.noticeChannel]
            target.notice('%s: %s' % (user.nickname, self.message))


class JeevesWannabeHandler(commands.CommandHandler):

    """Unfinished clone of some of the stuff jeeves does."""

    def __init__(self, cachedir, cachetime=60):
        commands.CommandHandler.__init__(self)
        # must exist because marienz is lazy
        self.cachedir = cachedir
        # in seconds
        self.cachetime = cachetime

    @defer.inlineCallbacks
    def _cache(self, url, filename):
        filename = os.path.join(self.cachedir, filename)
        if (not os.path.exists(filename) or
            (time.time() - os.path.getmtime(filename) > self.cachetime)):
            # update the cached file
            yield client.downloadPage(url, filename)
        defer.returnValue(open(filename))

    @commands.args()
    @defer.inlineCallbacks
    def handle_herdlist(self, user, source):
        """Return a list of all known herds."""
        herdsxml = yield self._cache(
            'http://www.gentoo.org/proj/en/metastructure/herds/herds.xml'
            '?passthru=1', 'herds.xml')
        herds = [
            herd.text.strip()
            for herd in etree.parse(herdsxml).findall('herd/name')]
        herds.sort()
        while herds:
            chunk, herds = herds[:40], herds[40:]
            source.msgTo(user, ', '.join(chunk))

    @commands.args(commands.string)
    @defer.inlineCallbacks
    def handle_herd(self, user, source, herdname):
        """herd <name>: return info on a herd."""
        herdsxml = yield self._cache(
            'http://www.gentoo.org/proj/en/metastructure/herds/herds.xml'
            '?passthru=1', 'herds.xml')

        emails = []

        for herd in etree.parse(herdsxml).findall('herd'):
            if herd.findtext('name') == herdname:
                for dev in herd.findall('maintainer'):
                    role = dev.findtext('role')
                    email = dev.findtext('email').split('@')[0]
                    if role is not None:
                        email = '%s (%s)' % (email, role)
                    emails.append(email)

                projects = list(herd.findall('maintainingproject'))
                if len(projects) > 1:
                    log.msg(
                        "I don't like multiple maintainingprojects "
                        "per herd, Please fix me.")
                if projects:
                    projectxml = yield self._cache(
                        'http://www.gentoo.org/%s?passthru=1' %
                        projects[0].text, '%s.xml' % herdname)
                    for dev in etree.parse(projectxml).findall('dev'):
                        emails.append(dev.text.strip())

                source.msgTo(user, ', '.join(sorted(emails)))
                return

    @commands.optargs(1, commands.string, commands.longstring)
    @defer.inlineCallbacks
    def handle_ping(self, user, source, herdname, reason='ping'):
        """ping <herd> [reason]: ping herd members."""
        herdsxml = yield self._cache(
            'http://www.gentoo.org/proj/en/metastructure/herds/herds.xml'
            '?passthru=1', 'herds.xml')

        emails = []

        for herd in etree.parse(herdsxml).findall('herd'):
            if herd.findtext('name') == herdname:
                for dev in herd.findall('maintainer'):
                    email = dev.findtext('email').split('@')[0]
                    emails.append(email)

                projects = list(herd.findall('maintainingproject'))
                if len(projects) > 1:
                    log.msg(
                        "I don't like multiple maintainingprojects "
                        "per herd, Please fix me.")
                if projects:
                    projectxml = yield self._cache(
                        'http://www.gentoo.org/%s?passthru=1' %
                        projects[0].text, '%s.xml' % herdname)
                    for dev in etree.parse(projectxml).findall('dev'):
                        emails.append(dev.text.strip())

                source.msg(
                    '%s: %s (%s)`' % (
                        ', '.join(sorted(emails)), reason, user.nickname))
                break
        else:
            source.msgTo(user, 'no such herd')


class RegExpBan(commands.ConfigurableMixin):

    """Ban when text matching a regexp is mentioned."""

    implements(ibot.IMessageHandler, ibot.IUserPartedHandler)

    configurables = {
        'target': (
            'the destination channel.', 'target', u'%s', commands.bytestring),
        }

    def __init__(self, target='#gentoo-ops'):
        self.target = target
        self.patterns = {}

    def message(self, user, source, message):
        """Possibly respond to a message."""
        if user is source or source.name == self.target:
            return
        for pattern, regexp in self.patterns.iteritems():
            if regexp.match(message) is not None:
                source.bot.user.channels[self.target].notice(
                    'ban on %s of %s for matching %r' % (
                        source.name, user.nickname, pattern))
                d = user.hostmask()
                @d.addCallback
                def kickban(hostmask):
                    """Use the hostmask to kickban."""
                    nickname, username, hostname = hostmask
                    source.mode(
                        True, 'b', '*!*@%s!%s' % (hostname, self.target))
                    # user may have left by now
                    if source.users.get(nickname) is user:
                        source.kick(user, 'go away')
                d.addErrback(log.err)
                return d

    userParted = message

    def get_patterns(self):
        """Get a list of configured patterns."""
        return u', '.join(unicode(repr(pattern)) for pattern in self.patterns)

    def set_add(self, args):
        """Add a pattern."""
        pattern, = commands.parseArgs(args, commands.longbytestring)
        self.patterns[pattern] = re.compile(pattern)

    def set_remove(self, args):
        """Remove a pattern."""
        pattern, = commands.parseArgs(args, commands.longbytestring)
        try:
            del self.patterns[pattern]
        except KeyError:
            pass # report?


class BanListHandler(commands.CommandHandler):

    """Query a banlist."""

    implements(ibot.IJoinHandler)

    # XXX TODO probably make maxResults configurable?

    def __init__(self, maxResults=5):
        commands.CommandHandler.__init__(self)
        self.maxResults = maxResults
        # XXX this could in theory get annoyingly large
        self.regexpCache = {}

    def joined(self, channel):
        """Update banlist."""
        channel.mode(True, 'b')

    @commands.args(commands.bytestring, commands.bytestring)
    def handle_bans(self, user, source, channel, query):
        """bans <channel> <query>: Lists channel bans (matching query)."""
        try:
            channel = user.bot.user.channels[channel]
        except KeyError:
            source.msgTo(user, 'No such channel!')
            return
        if user.nickname not in channel.users:
            # This is a lightweight way to make sure a banned user
            # cannot use the bot to look up the mask(s) banning him.
            source.msgTo(user, 'You can only ask about channels you are in')
            return
        query = user.bot.caseFolder(query)
        target = user.bot.users.get(query)
        if target is not None:
            mask = '%s!%s@%s' % (
                target.nickname, target.username, target.hostname)
            source.msgTo(user, 'using mask %s' % (mask, ))
        else:
            mask = None
        bans = channel.modes['b']
        count = 0
        sent = 0
        for ban, (setter, timestamp) in bans.iteritems():
            line = '%s (set by %s %s ago)' % (
                ban, setter, commands.formatTime(
                    datetime.datetime.now() -
                    datetime.datetime.fromtimestamp(int(timestamp))))
            matched = False
            ban = user.bot.caseFolder(ban)
            if query in ban or query in user.bot.caseFolder(setter):
                matched = True
            elif mask is not None:
                # If it is a forward we do not want to match on the destination
                index = ban.find('!', ban.index('!') + 1)
                if index != -1:
                    ban = ban[:index]
                # If it is a +q we do not want to include the %
                if ban.startswith('%'):
                    ban = ban[1:]
                regexp = self.regexpCache.get(ban)
                if regexp is None:
                    regexp = self.regexpCache[ban] = re.compile(
                        fnmatch.translate(ban))
                if regexp.match(mask) is not None:
                    matched = True
            if matched:
                count += 1
                if count < self.maxResults:
                    source.msgTo(user, line)
                    sent += 1
        if count:
            source.msgTo(
                user, 'Returned %s out of %s results' % (sent, count))
        else:
            source.msgTo(user, 'Nothing found')


class DocScraperHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """Parse gentoo's documentation listing and search it."""

    configurables = {
        'maxage': (
            'Time in seconds after which the list is refreshed.',
            'maxAge', u'%s', commands.integer),
        }

    def __init__(self, maxAge=60*10):
        commands.CommandHandler.__init__(self)
        self.maxAge = maxAge
        self.docs = []
        self.lastUpdate = 0
        self._updateCall = None

    def _updateMessage(self, user, source):
        """Called to spam the channel with the reason we're so slow."""
        source.msgTo(user, 'Updating document listing cache...')
        self._updateCall = None

    @commands.args(commands.longstring)
    def handle_gentoodoc(self, user, source, query):
        """Search through the gentoo doc listing."""
        if time.time() > self.lastUpdate + self.maxAge:
            if self._updateCall is not None:
                source.msgTo(
                    user, "Doc update in progress, try again in a few seconds")
                return
            # If we take more than 10 seconds we say what we're doing.
            self._updateCall = reactor.callLater(
                10, self._updateMessage, user, source)
            d = self._updateCache()
        else:
            d = defer.succeed(self.docs)
        d.addCallback(self._query, user, source, query)
        return d

    def _updateCache(self):
        """Update our cache."""
        page = client.getPage('http://www.gentoo.org/doc/en/list.xml')
        page.addCallback(self._preParse)
        return page

    def _preParse(self, page):
        """Get the relevant tagsoup part."""
        self.lastUpdate = time.time()
        # Note: ElementSoup uses Null instead of None if it cannot find things
        # so this really should be "if not ...", not "if ... is not None",
        # and "blindly" calling findNext is actually safe.
        soup = BeautifulSoup.BeautifulSoup(page)
        lists = soup.first('p', attrs={'class': 'chaphead'}).findNext('ul')
        if not lists:
            raise ValueError('no lists')
        self.docs = list(self._parse([], lists))
        if self._updateCall.active():
            self._updateCall.cancel()
        self._updateCall = None
        return self.docs

    def _parse(self, path, soup):
        """Turn a tagsoup into a bunch of doc entries."""
        for thing in soup.fetch('li', recursive=False):
            title = thing.first('b', recursive=False)
            link = thing.first('a', recursive=False)
            if not bool(title) ^ bool(link):
                raise ValueError('not one of title and link (%r)' % (thing,))
            if title:
                childLists = thing.fetch('ul', recursive=False)
                if not childLists:
                    raise ValueError('title but no childlist')
                for childList in childLists:
                    for child in self._parse(path + [title.string], childList):
                        yield child
            else:
                yield path, link.get('href'), (link.string or '')

    def _query(self, soup, user, source, query):
        """Search through the soup."""
        query = query.lower()
        # Some urls occur more than once. Pick one at random.
        results = dict((href, (path + [name]))
                       for path, href, name in self.docs
                       if query in name.lower())
        if results:
            if len(results) > 3:
                source.msgTo(user, 'only giving you the first 3 matches')
            for href, path in results.items()[:3]:
                source.msgTo(user, 'http://www.gentoo.org%s (%s)' % (
                        href, '/'.join(path)))
        else:
            source.msgTo(user, 'no match')
