
"""Google search integration.

This uses the "ajax" search api, which includes a simple "RESTful"
interface: a simple http GET request returning JSON data.

This interface is obviously designed for web use. I think what I do
here is not against the TOS, but the TOS was written with web use in
mind. I am sending the bot's homepage as "referer". This is nuts, but
the documentation strongly recommends sending a referer, and using my
homepage is what the "cringer" bot also does. I'm also sending an api
key generated for the same url (not included in the source since it is
tied to my google account).

This provides the same interface as a previous handler using the
(dead) SOAP api.
"""

from urllib import urlencode
import json

from zope.interface import implements

from twisted.python import log
from twisted.internet import defer
from twisted.web import client

from mzbot import commands, ebot, ibot


class GoogleHandler(commands.ConfigurableMixin):

    """Do google feeling-lucky searches."""

    implements(ibot.ICommandHandler, ibot.ICommandHelpProvider)

    configurables = {
        'key': (
            'key', 'key', u'%s', commands.string,
            ),
        }

    def __init__(self):
        self.commandset = set()
        self.additions = {}
        self.key = ''

    def get_additions(self):
        """List registered additions."""
        return u', '.join(u'%s: %r' % pair
                          for pair in self.additions.iteritems())

    def set_add(self, value):
        """add <command> [addition]: Add a command."""
        args = commands.parseOptArgs(
            value, 1, commands.string, commands.longbytestring)
        command = args[0]
        if command in self.additions:
            raise ebot.InvalidArgument('command exists')
        if command.endswith(u'list'):
            # Need to prevent clashes. We could deal with someone
            # registering "foolist" (getting "foolistlist" as its list
            # command), but not with someone registering "foo" as well
            # as "foolist" as regular commands (in either order). Just
            # forbid this outright instead of trying to allow
            # non-clashing use.
            raise ebot.InvalidArgument('must not end in "list"')
        if len(args) == 1:
            addition = ''
        else:
            addition = args[1]

        self.additions[command] = addition
        self.commandset.add(command)
        self.commandset.add(command + u'list')

    def set_remove(self, value):
        """remove <command>: remove a command."""
        command, = commands.parseArgs(value, commands.string)

        try:
            del self.additions[command]
        except KeyError:
            raise ebot.InvalidArgument('no such command')

        self.commandset.remove(command)
        self.commandset.remove(command + u'list')

    def _getPage(self, url, headers): # pragma: no cover
        """Actually do an http request. Overridden in tests."""
        # This uses the old api instead of the new client.Agent one.
        # "It should be significantly better than anything above",
        # but requires a *lot* of boilerplate to just grab a page.
        return client.getPage(url, headers=headers)

    @defer.inlineCallbacks
    def handle(self, user, source, command, args):
        """Do the requested search."""
        # This only works because we forbid regular additions ending
        # in "list" above.
        wantList = command.endswith(u'list')
        if wantList:
            command = command[:-4]

        addition = self.additions[command]

        # TODO: this is probably cleaner using a local arg parser that
        # wraps bytestring and returns users (or fails). Avoids the
        # "add it back to query" code below.
        try:
            args = commands.parseKwargs(args,
                                        [('target', commands.bytestring),
                                         ('query', commands.longbytestring)],
                                        [('query', commands.longbytestring)])
        except ebot.InvalidArgument, e:
            source.msgTo(user, str(e))
        else:
            query = args['query']
            target = args.get('target', user)
            if target is not user:
                # if first arg is a valid user, use as one
                target = user.bot.users.get(target, None)
                if target is None:
                    # not a valid user, add it back to the query
                    target = user
                    query = '%s %s' % (args['target'], query)
            if addition:
                query = '%s %s' % (addition, query)

            # I could send a "userip" for uncloaked users, but that
            # is probably not worth the hassle.
            # rsz='small' is currently the default, but let's be explicit.
            # We might want to make hl configurable at some point.
            qdict = dict(q=query, v='1.0', rsz='small', hl='en')
            if self.key:
                qdict['key'] = self.key

            try:
                page = yield self._getPage(
                    'http://ajax.googleapis.com/ajax/services/search/web?' +
                    urlencode(qdict), headers=dict(
                        referer='http://mzz.mine.nu/m/mzbot.html'))
            except Exception: # XXX try to be more specific
                log.err()
                source.msgTo('Google http request failed!')
                return

            # This is only slightly paranoid in dealing with data from google.
            # I'm hoping this will suffice in practice, and failures here
            # are dealt with gracefully enough higher up.
            response = json.loads(page)
            details = response.get(u'responseDetails', None)
            status = response.get(u'responseStatus', 0)
            if details is not None:
                # Truncate since I have no clue how long these get.
                source.msgTo(user, 'Google said: %r' % (details[:100],))
                log.msg('Google api status %d response %r' %
                        (status, details))
                return

            if status != 200:
                source.msgTo(user, 'Google response status %d' % (status,))
                log.msg('Google api status %d' % (status,))
                return

            results = response.get(u'responseData', {}).get(u'results', [])
            if not results:
                source.msgTo(user, 'Google had no results!')
                return

            if wantList:
                source.msgTo(user, u' - '.join(result.get(u'url', u'???')
                                               for result in results))
            else:
                result = results[0]
                source.msgTo(user, u'%s (%s)' % (
                        result.get(u'url', u'???'),
                        result.get(u'titleNoFormatting')))

    def getCommandDescription(self, command):
        """Use the addition for command help."""
        # This only works because we forbid regular additions ending
        # in "list" above.
        if command.endswith(u'list'):
            command = command[:-4]
            reply = ('[target] <query>: do a "%s" google query, '
                     'returning the first few results')
        else:
            reply = '[target] <query>: do a "%s" google query'
        addition = self.additions[command]
        if addition is None:
            return '[target] <query>: do a google query'
        else:
            return reply % (addition,)
