# do not check for methods with no argument, without self as first arg,
# missing __init__, missing docstring
# pylint: disable-msg=E0211, E0213, W0232, C0111

"""The ircbot's interfaces."""

from zope.interface import Interface, Attribute


class NoSuchUser(Exception):
    """Raised if a method was called on a User the server doesn't know."""


class IBotHandler(Interface):
    """Interfaces responding directly to IBot should subclass this."""

class IMessageHandler(IBotHandler):
    def message(user, source, message):
        """called with an IUser, IChannel or IUser, and string."""

class IActionHandler(IBotHandler):
    def action(user, source, action):
        """called with an IUser, IChannel or IUser, and string."""

class INoticeHandler(IBotHandler):
    def notice(user, source, message):
        """called with an IUser, IChannel or IUser, and string."""

class IUserModeHandler(IBotHandler):
    def userModeChanged(bot, modeSet, mode):
        """Called with an IBot, a bool, and a char."""

class IChannelModeHandler(IBotHandler):
    def channelModeChanged(setter, channel, modeSet, mode, arg=None):
        """Called when a channel mode changed.

        setter is an IUser or string.
        channel is an IChannel.
        modeSet is a bool.
        mode is a single-character string.
        arg is a string or None.

        This is called before the bot handles the modechange itself.
        """

class IKickedHandler(IBotHandler):
    def kicked(kicker, channel, kicked, message):
        """Called with an IUser, IChannel, IUser and bytestring.

        Some user (possibly us) was kicked from the channel.
        """

class ILoginHandler(IBotHandler):
    def loggedIn(bot):
        """Called with an IBot that just logged in.

        This must return a deferred. The bot will wait until this
        fires. If it returns False the bot will disconnect, if it
        returns True remaining handlers will run.
        """

class IDisconnectHandler(IBotHandler):
    def disconnected(bot, reason=None):
        """Called with an IBot that just disconnected.

        This is a bit special: obviously most of the methods on this
        bot and the users and channels reachable through it just
        became invalid. This should only be used for cleanup.

        The reason argument is the same one passed to connectionLost
        by twisted.
        """

class INickInUseHandler(IBotHandler):
    def nickInUse(bot, nickname):
        """An attempt to change to nickname failed: it was in use."""

class ILoginNickInUseHandler(IBotHandler):
    def loginNickInUse(bot, nickname):
        """Called with an IBot that just connected, but nickname was in use.

        Should return either an alternate nickname or None.
        """

class IHostHiddenHandler(IBotHandler):
    def hostHidden(bot, vhost):
        """Called when our hostname is being changed (vhost/cloak applied).

        This runs before bot.user.hostname is updated.
        """

class INewUserHandler(IBotHandler):
    def newUser(user):
        """Called with an IUser that was just created.

        This might be just a 'transient' user created to send a msg
        to, or a user that just joined a channel. This handler is
        called before a join is processed.
        """

class IJoinHandler(IBotHandler):
    def joined(channel):
        """Called with an IChannel that we just joined.

        The userlist will be up to date, but the modelist will not be.
        """

class IUserJoinedHandler(IBotHandler):
    def userJoined(user, channel):
        """Called when a user joined a channel."""

class ILeftHandler(IBotHandler):
    def left(channel):
        """Called with an IChannel that we just left."""

class IUserLeftHandler(IBotHandler):
    def userLeft(user, channel):
        """Called when a user leaves a channel, for any reason.

        This is also called for all channels the user was in when they
        quit.
        """

class IUnknownHandler(IBotHandler):
    def unknown(command, params):
        """Called with a numeric and params the bot can't handle."""

class ISystemMessageHandler(IBotHandler):

    """Handle messages and notices not from a user or channel."""

    def systemMessage(bot, user, channel, message):
        """Called with an ibot and 3 strings for a message not from a user."""

    def systemNotice(bot, user, channel, message):
        """Called with an ibot and 3 strings for a notice not from a user."""


class ICTCPReplyHandler(IBotHandler):
    def ctcpReply(user, channel, tag, data):
        """Called with an IUser, an ISource and two strings."""

class ICTCPQueryHandler(IBotHandler):
    def ctcpQuery(user, channel, tag, data):
        """Called with an IUser, an ISource and two strings."""

class IUserQuitHandler(IBotHandler):
    def userQuit(user, reason):
        """Called with an IUser and a string."""

class IUserPartedHandler(IBotHandler):
    def userParted(user, channel, reason):
        """Called with an IUser, IChannel and string."""

class ITopicHandler(IBotHandler):

    """Channel topic-related information."""

    def topicChanged(user, channnel, topic):
        """Called right before the internal channel topic change.

        user is a bytestring or IUser.
        channel is an IChannel.
        topic is a bytestring.
        """

    def topicState(channel, topic):
        """Called when the current topic state becomes known.

        channel is an IChannel.
        topic is a bytestring.
        """

    def topicStateExtended(channel, setter, timestamp):
        """Called with extra information about the current topic.

        channel is an IChannel.
        setter is a bytestring.
        timestamp is a timestamp (integer).

        Do not rely on this, it is not standardized.
        """


class IMOTDHandler(IBotHandler):

    """'message of the day' stuff."""

    def welcome(line):
        """Friendly welcome message."""

    def motdStart(line):
        """Called with the first motd line."""

    def motd(line):
        """Called with a line of motd."""

    def motdEnd():
        """Called when the motd is complete."""

class IStatisticsHandler(IBotHandler):

    """Various server statistics."""

    def myinfo(servername, version, usermodes, channelmodes):
        """Four strings describing the server name, version and features."""

    def yourhost(line):
        """Human-readable description of the server."""

    def created(line):
        """Human-readable description of server creation datetime."""

    def luserclient(line):
        """human-readable description of network-connected client count."""

    def luserop(count):
        """Number of ircops."""

    def luserunknown(count):
        """Number of "unknown" connections."""

    def luserchannels(count):
        """Number of channels."""

    def luserme(line):
        """Human-readable description of client count on this server."""

    def maxConnectionCount(line):
        """Human-readable highest connection count."""

    def localUsers(line):
        """Human-readable local user count."""

    def globalUsers(line):
        """Human-readable global user count."""


class ICommandHandler(Interface):

    """Plugin for the CommandHandler."""

    commandset = Attribute("set of commands (unicode) we accept")

    def handle(user, channel, command, args):
        """Handle a command.

        user is an IUser, channel an ISource, command and args are unicode.
        """

class ICommandHelpProvider(Interface):

    """CommandHandlers are adapted to this to get help."""

    def getCommandDescription(command):
        """Return a string with help for a command.

        This should be a bytestring like '<requiredarg> [optionalarg]: blah.'
        It should *not* start with the command name.
        """


class IAuthenticator(Interface):

    """Class that authenticates an IUser."""

    def auth(user):
        """return a deferred returning if an IUser is authenticated or not."""

    def description():
        """return a string describing this authenticator."""


class ISource(Interface):

    """A source for a message that can be replied to."""

    bot = Attribute("the IBot that knows this source.")
    name = Attribute('source name bytestring')
    uname = Attribute('channel name unicode (may not be unique)')
    data = Attribute("a dict for storing stuff that shouldn't be persisted")

    def notice(message):
        """Send a notice to this user."""

    def msg(message):
        """Send a msg to this channel."""

    def msgTo(target, message):
        """Send a message to the target user on this channel.

        target should provide IUser.
        """

    def me(message):
        """Do an action."""

    def ctcpReply(tag, data=None):
        """Send one extended message as a CTCP reply.

        Arguments are bytestrings.
        """

    def ctcpQuery(messages):
        """Send one extended message as a CTCP query.

        Arguments are bytestrings.
        """


class IChannel(ISource):

    """A channel the bot is on.

    Mappings may be case insensitive if the protocol is.

    The modes attribute maps the mode character to mode-specific data:
    - If the mode takes no argument the data is None.
    - If the mode takes a bytestring the data is that bytestring.
    - If the mode is applied to users on the channel the data is a map
      nickname -> IUser (like our users attribute).
    - If the mode is a list of masks the data is a map of
      mask -> (settermask, time())
    """

    topic = Attribute('channel topic or None if unknown')
    users = Attribute('mapping of nicks to IUser objects')
    modes = Attribute('mapping of mode characters to mode data.')

    def setTopic(topic):
        """Set the topic."""

    def mode(mode, enable, arg=None):
        """Change a channel mode."""

    def kick(user, reason=None):
        """Kick a user from this channel."""

    def remove(user, reason=None):
        """Remove a user from this channel."""

    def part(reason=None):
        """Leave a channel."""


class IUser(ISource):

    """A user on irc."""

    nickname = Attribute("The user's nickname (alias for the name property).")
    username = Attribute("The user's username or None.")
    hostname = Attribute("The user's hostname or None.")
    realname = Attribute("The user's realname or None.")
    server = Attribute("The user's server name or None.")
    chanmodes = Attribute(
        '''Map from channel mode letters to a set of channel names

        The current irc implementation uses a dict, but you are encouraged
        not to rely on that (only rely on "in" or "not in" checks).''')

    channels = Attribute("A dict of channels the user is on.")
    identified = Attribute(
        "True if identified, False if not identified, None if unknown.")

    def who():
        """Returns a deferred firing self after setting server and realname."""

    def hostmask():
        """Returns a deferred (nickname, username, hostname)."""


class IBot(Interface):

    """The bot itself."""

    users = Attribute("mapping from nicknames to IUsers")
    user = Attribute('the IUser representing the bot itself.')
    handlers = Attribute(
        'mapping from IBotHandler subclass to sequence of handlers')

    def away(message=''):
        """Set our away message."""

    def mode(enabledModes='', disabledModes=''):
        """Set user modes. Arguments are (byte)strings."""

    def setNick(nickname):
        """Change our nick."""

    def quit(message=''):
        """Disconnect."""

    def join(channel, key=None):
        """Join a channel."""

    def getUser(nickname):
        """Return an IUser for a nickname."""

    def caseFolder(string):
        """Convert string to lowercase based on the ircd's casemapping."""


class IBotFactory(Interface):

    """Where bots are made."""

    nickname = Attribute("our nick")
    bot = Attribute("the current Ibot, or None if there isn't one.")
    handlers = Attribute("list of handlers")

    def buildProtocol(addr):
        """Return a protocol"""


class IBotPlugin(Interface):

    """Build a handler."""

    name = Attribute("name of this handler, should not contain spaces")
    doc = Attribute('Docstring-style description of how to call buildHandler')

    def buildHandler(store, args):
        """Make an IBotHandler and/or ICommandHandler from a store and string.

        Raises ebot.InvalidArgument if the string is invalid.
        """


class IBotConfigPlugin(Interface):

    """Definition of a handler builder."""

    name = Attribute("name of this handler, should not contain spaces")
    doc = Attribute('Docstring-style description.')

    def buildHandler():
        """Build an IBotHandler and/or ICommandHandler."""


class ILogger(IBotHandler):

    """Keeps logs in a database."""

    store = Attribute('the storm Store with the actual logs.')


class IConfigurable(Interface):

    """A configurable handler.

    Notice gettables and settables may return completely different
    things. You can for example use two 'settable' commands to add and
    remove items from a list and provide a 'gettable' command to view
    the list.
    """

    def gettables():
        """Get a dict mapping strings accepted by get to docstrings."""

    def get(key):
        """Get a configurable setting. Return a unicode object."""

    def settables():
        """Get a dict mapping strings accepted by set to docstrings."""

    def set(key, value):
        """Set a configurable setting. value should be unicode.

        May raise ebot.InvalidArgument.
        """

class ISaveable(IConfigurable):

    """IConfigurable subclass that can save its interesting state."""

    def save():
        """Describe how to call set to reinitialize us from scratch.

        Yields (key, value) pairs.
        """


class IAvatar(Interface):
    """Marker interface for an object adaptable to various others."""


class IIgnore(IBotHandler):
    def ignored(user, source=None):
        """Check if we are ignoring this IUser on an optional ISource."""


class IRawLogHandler(IBotHandler):
    def rawlog(bot, out, message):
        """Log a single message going over the network.

        out is True for outgoing, False for incoming.

        Because of when this handler runs calling back into the bot is unwise.
        """
