
"""Implementation of ibot interfaces using the irc protocol.

All you need from this module is the IRCBotFactory. Use that to spawn
a bot and add handlers, look at ibot for what handlers can do.

This module tries to follow rfc2812 for irc itself.

It uses a simpler method than the one in
http://www.irchelp.org/irchelp/rfc/ctcpspec.html for ctcp: every
message has exactly one ctcp message or one normal message.

Misparsed messages or protocol violations will trigger exceptions
which will (usually) just be logged.
"""

import sys
import string
import textwrap
import datetime
import collections
from base64 import b64decode, b64encode
from time import time

from zope.interface import implements, providedBy

from twisted.python import log
from twisted.protocols import basic
from twisted.internet import defer, reactor, protocol
from twisted.persisted import styles
from twisted.words.protocols.irc import (
    lowDequote, lowQuote, X_DELIM, ctcpDequote, ctcpQuote,
    numeric_to_symbolic as NUMERICS, ServerSupportedFeatures, _intOrDefault)
# NUMERICS actually differs from the RFC on one point:
# RPL_BOUNCE in rfc2812 is 005, but NUMERICS has 010 for RPL_BOUNCE
# and 005 for RPL_ISUPPORT, which matches the real world.
# It also has a bug: RPL_ADMINLOC1 missing (collision, must file bug).


from mzbot import ibot, ebot, util


class SASLMechanism(object):

    def __init__(self, nick, password):
        self.nick = nick
        self.password = password


class Plain(SASLMechanism):

    def received(self, data):
        if data:
            raise ValueError('did not expect %r' % (data,))

        return '\0%s\0%s' % (self.nick, self.password)


# TODO: DH-BLOWFISH support? Issues include:
# - Cannot find docs other than source (dh-blowfish.c and cap_sasl.pl)
#   Not a huge deal, but makes this a bit more time-consuming.
# - I cannot find a convenient Python implementation or binding to do
#   the DH part of this. Blowfish is in pycrypto, but DH is where?
#   m2crypto? paramiko and conch have something but I'm not sure it applies.
# This is low-priority for me because with ssl support enabled I do not think
# DH-BLOWFISH buys me anything over PLAIN.

SASL_MECHANISMS = {
    'PLAIN': Plain,
    }


CASE_FOLDERS = {}
for folderName, mapping in [
    ('rfc1459', string.maketrans(string.ascii_uppercase + r'\[]~',
                                 string.ascii_lowercase + r'|{}^')),
    ('strict-rfc1459', string.maketrans(string.ascii_uppercase + r'\[]',
                                        string.ascii_lowercase + r'|{}')),
    ('ascii', string.maketrans(string.ascii_uppercase,
                               string.ascii_lowercase)),
    ]:
    def folder(text, mapfunc=mapping):
        """Return the lowercase version of 'text'."""
        return text.translate(mapfunc)
    CASE_FOLDERS[folderName] = folder
del folderName, mapping


CAP_MODIFIERS = {'-': 'disabled',
                 '=': 'sticky',
                 '~': 'ack',
                 }

def _parseCap(args):
    # This is fine according to the draft rfc.
    if not args:
        args = ['']

    if len(args) == 1:
        more = False
        caps = args[0]
    elif len(args) == 2:
        star, caps = args
        if star != '*':
            raise ValueError('unsupported CAP args %r' % (args,))
        more = True
    else:
        raise ValueError('unsupported CAP LS %r' % (args,))

    result = {}
    for cap in caps.split():
        props = set()
        while cap[0] in CAP_MODIFIERS:
            mod, cap = cap[0], cap[1:]
            props.add(CAP_MODIFIERS[mod])
        result[cap] = props

    return more, result


def _isupportFlag(name, params):
    if params != ['']:
        log.msg('isupport %s did not expect params %r' % (name, params))
        # but we set it anyway
    return True


class SupportedFeatures(ServerSupportedFeatures):

    def __init__(self):
        ServerSupportedFeatures.__init__(self)
        self._features['KEYLEN'] = 23 # (from rfc2812)
        # TODO: sanity-check this default.
        self._features['CHANMODES'] = {
            'addressModes': 'bdeIq',
            'param': 'k',
            'setParam': 'fJl',
            'noParam': 'cgijLmnpPQrRstz',
            }

    def isupport_unknown(self, command, params):
        log.msg('Unknown ISUPPORT %s %r' % (command, params))
        return ServerSupportedFeatures.isupport_unknown(self, command, params)

    def isupport_CAPAB(self, params):
        # Sent by hyperion.
        return _isupportFlag('CAPAB', params)

    def isupport_KNOCK(self, params):
        """Parse if the KNOCK command is available."""
        # Not in the draft isupport rfc, hyperion and seven send this.
        return _isupportFlag('KNOCK', params)

    def isupport_ETRACE(self, params):
        # Just guessing...
        return _isupportFlag('ETRACE', params)

    def isupport_CPRIVMSG(self, params):
        # Just guessing...
        return _isupportFlag('CPRIVMSG', params)

    def isupport_CNOTICE(self, params):
        # Just guessing...
        return _isupportFlag('CNOTICE', params)

    def isupport_FNC(self, params):
        # Just guessing...
        return _isupportFlag('FNC', params)

    def isupport_WHOX(self, params):
        # Just guessing...
        return _isupportFlag('WHOX', params)

    # Fairly certain these are sufficiently similar.
    isupport_IRCD = ServerSupportedFeatures.isupport_NETWORK
    # This is not really documented. The ircd changelog just says
    # "Add CLIENTVER=3.0 to indicate presence of 3.0 core capabilities
    # (such as CAP, extended WHO, etc) as requested by LifeIsPain."
    isupport_CLIENTVER = ServerSupportedFeatures.isupport_NETWORK

    def isupport_CHARSET(self, params):
        # This was removed from the spec but the ircd still sends it. Ignore.
        return None

    def isupport_CASEMAPPING(self, params):
        return params[0]

    def isupport_ELIST(self, params):
        # Supports "extensions for the LIST command".
        if len(params) > 1:
            log.msg('unsupported ISUPPORT ELIST %r' % (params,))
        return params[0]

    def isupport_HOSTLEN(self, params):
        """Parse max username length (in masks)."""
        # Not in the draft isupport rfc, hyperion sends this.
        return _intOrDefault(params[0])

    def isupport_KEYLEN(self, params):
        """Parse maximum length of channel key."""
        # Not in the draft isupport rfc, hyperion sends this.
        return _intOrDefault(params[0], self.getFeature('KEYLEN'))

    def isupport_USERLEN(self, params):
        """Parse max username length (in masks)."""
        # Not in the draft isupport rfc, hyperion sends this.
        return _intOrDefault(params[0])

    def isupport_SILENCE(self, params):
        """Max length of the SILENCE list."""
        # Not in the draft isupport rfc, hyperion sends this.
        return _intOrDefault(params[0])

    def isupport_CALLERID(self, params):
        # Not in the draft isupport rfc, seven sends this.
        # Indicates a umode.
        return params[0]

    def isupport_DEAF(self, params):
        # Not in the draft isupport rfc, seven sends this.
        # Indicates a umode.
        return params[0]

    def isupport_MONITOR(self, params):
        # I cannot find any documentation on this.
        # This treats no parameter as "support an infinite amount"
        # (as opposed to completely missing meaning "not supported")
        return _intOrDefault(params[0], sys.maxint)

    def isupport_EXTBAN(self, params):
        # Made-up parsing rules.
        if len(params) < 2:
            log.msg('unsupported ISUPPORT EXTBAN %r' % (params,))
            return

        if len(params) > 2:
            log.msg('extra ISUPPORT EXTBAN args ignored from %r' % (params,))

        return tuple(params[:2])

    # TODO: isupport_STD


class Source(object, styles.Ephemeral):

    """internal base class for Channel and User.

    Note this doesn't implement ISource, it's missing msgTo.
    """

    def __init__(self, bot, name):
        object.__init__(self)
        self.bot = bot
        self.name = name
        self.data = {}

    def notice(self, message, maySplit=True):
        """Send a notice to this source."""
        self._send('NOTICE %s :' % (self.name,), message, maySplit)

    def msg(self, message, maySplit=True):
        """Send a privmsg to this source."""
        self._send('PRIVMSG %s :' % (self.name,), message, maySplit)

    def _send(self, prefix, message, maySplit):
        """Split up message and send it."""
        if isinstance(message, unicode):
            message = message.encode('utf-8')
        message = lowQuote(message)
        if maySplit:
            lines = textwrap.wrap(
                message, self.bot.maxMessageLength - len(prefix),
                expand_tabs=False, replace_whitespace=False)
            for line in lines:
                self.bot.sendMessage(self.name, '%s%s' % (prefix, line))
        else:
            self.bot.sendMessage(self.name, '%s%s' % (prefix, message))

    def me(self, message): # naming convention #pylint: disable-msg=C0103
        """Do a ctcp action."""
        if isinstance(message, unicode):
            message = message.encode('utf-8')
        self.ctcpQuery('ACTION', message)

    def ctcpReply(self, tag, data=None):
        """Send a CTCP reply."""
        if data is not None:
            tag = '%s %s' % (tag, data)
        self.notice('%s%s%s' % (X_DELIM, ctcpQuote(tag), X_DELIM), False)

    def ctcpQuery(self, tag, data=None):
        """Send a CTCP query."""
        if data is not None:
            tag = '%s %s' % (tag, data)
        self.msg('%s%s%s' % (X_DELIM, ctcpQuote(tag), X_DELIM), False)

    @property
    def uname(self):
        # XXX lame, improve
        return self.name.decode('latin-1')

class Channel(Source):

    """I represent a channel and the people (or bots) in it."""

    implements(ibot.IChannel)

    def __init__(self, bot, name):
        Source.__init__(self, bot, name)
        self.users = util.PreservingFoldingDict(bot.caseFolder)
        self.topic = None
        self.modes = {}
        for mode in bot.supported.getFeature('CHANMODES')['addressModes']:
            self.modes[mode] = util.PreservingFoldingDict(bot.caseFolder)
        self.synced = False

    # ISource

    def msgTo(self, user, message):
        """Send a privmsg to the target user on this channel."""
        self.msg("%s: %s" % (user.nickname, message))

    # IChannel

    def setTopic(self, topic):
        """Change the topic."""
        self.bot.sendLine('TOPIC %s :%s' % (self.name, topic))

    def mode(self, enable, mode, *args):
        """Change exactly one user mode.

        enable is True for a '+' change and False for '-'.
        mode is the mode to change.
        args is the argument, if needed.
        """
        # TODO: support for multiple types of modechanges?
        if enable:
            tag = '+'
        else:
            tag = '-'
        if not args:
            self.bot.sendLine('MODE %s %s%s' % (self.name, tag, mode))
        else:
            modesPerMessage = self.bot.supported.getFeature('MODES')
            while args:
                batch = args[:modesPerMessage]
                args = args[modesPerMessage:]
                self.bot.sendLine(
                    'MODE %s %s%s %s' %
                    (self.name, tag, mode * len(batch), ' '.join(batch)))

    def kick(self, user, reason=None):
        """Kick user from channel with optional reason."""
        if reason is not None:
            self.bot.sendLine(
                'KICK %s %s :%s' % (self.name, user.nickname, reason))
        else:
            self.bot.sendLine('KICK %s %s' % (self.name, user.nickname))

    def remove(self, user, reason=None):
        """Remove user from channel with optional reason."""
        if reason is not None:
            self.bot.sendLine(
                'REMOVE %s %s :%s' % (self.name, user.nickname, reason))
        else:
            self.bot.sendLine('REMOVE %s %s' % (self.name, user.nickname))

    def part(self, reason=None):
        """Leave a channel."""
        if reason is not None:
            self.bot.sendLine('PART %s :%s' % (self.name, reason))
        else:
            self.bot.sendLine('PART %s' % (self.name,))

    def who(self):
        """Return a deferred that calls back when the who has completed."""
        d = self.bot.sendLine("WHO " + self.name)
        who = defer.Deferred()
        def _queueWho(result):
            self.bot._whos.append((self.name, who, self))
        d.addCallback(_queueWho)
        d.addErrback(log.err)
        return who


class User(Source):

    """I represent a person we can see in a channel."""

    implements(ibot.IUser)

    def __init__(self, bot, name):
        Source.__init__(self, bot, name)
        self.channels = util.PreservingFoldingDict(bot.caseFolder)
        self.hostname = None
        self.username = None
        self.realname = None
        self.server = None
        self.identified = None
        self.away = None
        self.awayMessage = None
        self.chanmodes = {}
        for mode in bot.supported.getFeature('PREFIX'):
            self.chanmodes[mode] = util.PreservingFoldingDict(
                bot.caseFolder)

    # ISource

    def msgTo(self, user, message):
        """Send a privmsg to the target user."""
        if user is self:
            self.msg(message)
        else:
            raise ValueError('this is deprecated now')

    # IUser

    def who(self):
        """Return a deferred that fires when our info is updated."""
        d = self.bot.sendLine("WHO " + self.nickname)
        who = defer.Deferred()
        def _queueWho(result):
            self.bot._whos.append((self.nickname, who, self))
        d.addCallback(_queueWho)
        d.addErrback(log.err)
        return who

    def hostmask(self):
        """Return a deferred returning a tuple (nick, username, hostname)."""
        if self.hostname is not None:
            # Return the cached data.
            return defer.succeed((self.nickname, self.username, self.hostname))
        return self.who().addCallback(User.hostmask)

    def whois(self):
        """Return a deferred returning a dict of whois data."""
        d = defer.Deferred()
        self.bot._whois.setdefault(self.nickname, []).append(d)
        self.bot.sendLine('WHOIS %s' % self.nickname)
        return d

    def _getNickName(self):
        return self.name

    def _setNickName(self, nickname):
        self.name = nickname

    nickname = property(_getNickName, _setNickName, doc='mirror of name.')


class OutQueue(object):

    """Object responsible for ratelimiting.

    This gets fed every message at the same time it is actually sent to
    the server, and gets a chance to veto sending before it happens.

    This implementation assumes the ircd processes messages using a
    timer like the one described in rfc1459, we do not need to worry
    about latency between us and the ircd (fixed latency), and we need
    to keep the number of unprocessed messages as well as the number
    of unprocessed bytes below certain limits.
    """

    # Maximum size of the server receive queue (bytes).
    maxQueueBytes = 555
    # Maximum number of messages in the server queue.
    maxQueueMessages = 10
    # Penalty time for one line (seconds).
    lineTime = 2
    # Starting time to fill (seconds).
    buildupTime = 10
    # Idle time (seconds) after sending a message after which the
    # server queue should be really empty (and we can send a ping for
    # measuring latency and flushing the buffer).
    # (this has a little slack).
    pingAfterSendTime = int(buildupTime + 2 * lineTime)

    def __init__(self, sendLines, tags):
        self._sendLines = sendLines
        self._tags = tags
        self._queue = collections.deque()
        # in bytes
        self._queueLength = 0
        self._queueEmptyCall = None
        self._messageTimer = 0

    def send(self, message, tag):
        if (self._queueLength + len(message) > self.maxQueueBytes or
            len(self._queue) >= self.maxQueueMessages):
            return False

        self._queueLength += len(message)
        self._queue.append((time(), len(message), tag))

        # Schedule emptying the outbound queue (if not already done).
        # XXX there was a comment claiming "Delay is lineTime because
        # that doubles as the delay before the virtual server "receives"
        # it" but I think that comment is confused and the delay should be
        # 0 here.
        if self._queueEmptyCall is None:
            self._queueEmptyCall = reactor.callLater(self.lineTime,
                                                     self._emptyQueue)

        return True

    def emptyQueue(self):
        if self._queueEmptyCall is not None:
            self._queueEmptyCall.cancel()
        self._emptyQueue()

    def _emptyQueue(self):
        """Pull lines out of our virtual server queue and maybe send more."""
        self._queueEmptyCall = None
        now = time()
        if self._messageTimer < now:
            self._messageTimer = now

        # "receive" lines as long as we're less than buildupTime
        # seconds "ahead", penalize lineTime seconds per line.
        while (self._queue and
               self._messageTimer <= now + self.buildupTime and
               self._queue[0][0] + self.lineTime < now):
            length, tag = self._queue[0][1:]
            if tag is not None:
                # tagged message.
                tagQueue = self._tags[tag]
                replyReceived = tagQueue[0]
                if replyReceived is None:
                    # Wait for the reply.
                    break
                tagQueue.popleft()
                # do not keep tons of empty queues around
                if not tagQueue:
                    del self._tags[tag]

            # Pop the event.
            self._queue.popleft()
            self._queueLength -= length
            # XXX this sort of matches what hyperion does but may be
            # waaay off for other ircds
            self._messageTimer += self.lineTime + length // 120
        else:
            if self._queue:
                # Resume after 1 second because that's the resolution
                # of our _messageTimer bumps.
                self._queueEmptyCall = reactor.callLater(1, self._emptyQueue)
        # potentially send more stuff.
        self._sendLines()

    def disconnected(self):
        if self._queueEmptyCall is not None:
            self._queueEmptyCall.cancel()
            self._queueEmptyCall = None


class IRCBot(basic.LineOnlyReceiver):

    """The core class of the ircbot, calls handlers and maintains lists."""

    implements(ibot.IBot, ibot.ICTCPQueryHandler, ibot.ICTCPReplyHandler)

    # Time in seconds between pings if we are idle.
    idlePingTime = 5 * 60

    maxMessageLength = 510 # Excludes the trailing CR-LF

    supportedCap = frozenset(['identify-msg', 'multi-prefix', 'sasl'])

    # Keep in sync with the default PREFIX in ServerSupportedFeatures
    _prefixToMode = {'@': 'o', '+': 'v'}

    def __init__(self):
        self.whoisResult = None
        self._highPriorityQueue = collections.deque()
        self._lowPriorityQueue = collections.deque()
        self._messageQueue = {}
        self._tags = {}
        self._outQueue = OutQueue(self._sendLines, self._tags)
        self.serverCap = {}
        self.activeCap = {}
        self.clientCapab = {}
        self.caseFolder = CASE_FOLDERS['rfc1459']
        self.users = util.PreservingFoldingDict(self.caseFolder)
        self._whois = util.PreservingFoldingDict(self.caseFolder)
        self._whos = collections.deque()
        self.modes = set()
        self.user = None

    ## core internal methods

    def connectionMade(self):
        """Called when the bot connected to the server."""
        self._idlePingCall = reactor.callLater(
            self.idlePingTime, self._sendIdlePing)
        self.handlers = self.factory.handlers
        self.supported = SupportedFeatures()
        self.loggedIn = False
        if self.factory.cap_support:
            self.sendLine('CAP LS')
        if self.factory.password is not None:
            self.sendLine("PASS %s" % self.factory.password)
        self.user = User(self, self.factory.nickname)
        self.user.username = self.factory.username
        self.user.realname = self.factory.realname
        self.saslMechanisms = list(self.factory.saslMechanisms)
        self.setNick(self.user.nickname)
        # The "*" args here are unused in rfc1459 and the real world.
        # In rfc2812 the first one influences initial mode.
        self.sendLine(
            "USER %s * * :%s" % (self.user.username, self.user.realname))
        # Receiving the 001/WELCOME numeric triggers the rest of login.

    def lineReceived(self, line):
        """Parse a line."""
        # rfc2812 states "Empty messages are silently ignored"
        if not line:
            return

        self.runHandlers(ibot.IRawLogHandler['rawlog'], self, False, line)

        line = lowDequote(line)

        prefix = None
        if line[0] == ':':
            prefix, line = line[1:].split(' ', 1)
            try:
                nick, rest = prefix.split('!')
                username, hostname = rest.split("@")
            except ValueError:
                # must be a servername or something, pass as-is
                pass
            else:
                prefix = self.getUser(nick)
                prefix.username, prefix.hostname = username, hostname

        if ' :' in line:
            line, trailing = line.split(' :', 1)
            args = line.split() + [trailing]
        else:
            args = line.split()
        command = args.pop(0)
        # determine if it is a numeric
        if len(command) == 3 and command.isdigit():
            nickname = args.pop(0)
            # If we are not logged in we do not have a nick to check.
            if (self.loggedIn and self.caseFolder(nickname) !=
                self.caseFolder(self.user.nickname)):
                log.err('numeric %r: our nick %r != %r' % (
                        command, self.user.nickname, nickname))
                return
            command = NUMERICS.get(command, command)
        method = getattr(self, 'irc_%s' % (command,), None)
        if method is not None:
            method(prefix, *args)
        else:
            self.runHandlers(ibot.IUnknownHandler['unknown'], command, args)
            if self.whoisResult is not None:
                self.whoisResult.setdefault('unknown', []).append(
                    (command, args))

    def connectionLost(self, reason=None):
        """Kill our pending timed call(s) when the connection dies."""
        # XXX perhaps our outqueue should just be added as a handler?
        self._outQueue.disconnected()
        self._idlePingCall.cancel()
        self._idlePingCall = None
        self.runHandlers(ibot.IDisconnectHandler['disconnected'],
                         self, reason)

    def _checkMessage(self, message):
        """Sanity-check a message."""
        if len(message) > self.maxMessageLength:
            raise ebot.MessageTooLong(message)
        if '\0' in message or '\n' in message or '\r' in message:
            raise ebot.InvalidCharacter(message)

    def sendLine(self, message, tag=None):
        """Send a raw line, high-priority, using an optional tag."""
        self._checkMessage(message)
        if tag is not None:
            self._applyTag(tag)
        d = defer.Deferred()
        self._highPriorityQueue.append((message, d, tag))
        self._sendLines()
        return d

    def sendMessage(self, target, line, tag=None):
        """Send a raw line over irc, low-priority.

        'target' can be any object. There is a queue per object,
        messages are sent per-queue in order.

        Tag is a string used for matching a reply to this message.
        """
        self._checkMessage(line)
        if tag is not None:
            self._applyTag(tag)
        d = defer.Deferred()
        queue = self._messageQueue.setdefault(target, collections.deque())
        queue.append((line, d, tag))
        self._sendLines()
        return d

    def _applyTag(self, tag):
        """Try to put a pending tag in the list of tags."""
        queue = self._tags.setdefault(tag, collections.deque())
        # It is an error to put in a new tag before an old one is received
        for replyReceived in queue:
            if replyReceived is None:
                raise ValueError('A tagged %r is pending' % (tag,))
        queue.append(None)

    def _sendLines(self):
        """Call this whenever you added something to a send queue."""
        # This should be safe to call when all queues are empty too
        # (so that _queueEmpty can call it).

        while True:
            # Figure out what queue to send out of.
            if self._highPriorityQueue:
                queue = self._highPriorityQueue
            elif self._lowPriorityQueue:
                queue = self._lowPriorityQueue
            elif self._messageQueue:
                queue = self._lowPriorityQueue
                # no iter, since I'm not sure it's safe when we drop keys
                for target, targetQueue in self._messageQueue.items():
                    queue.append(targetQueue.popleft())
                    # this makes sure we don't leak the keys, and makes
                    # sure that a non-empty _messageQueue actually has
                    # messages, not just empty queues.
                    if not targetQueue:
                        del self._messageQueue[target]
            else:
                # nothing to send
                break

            text, deferred, tag = queue[0]

            if not self._outQueue.send(text, tag):
                # Message does not fit on the queue, abort
                break

            queue.popleft()
            self.runHandlers(ibot.IRawLogHandler['rawlog'], self, True, text)
            basic.LineOnlyReceiver.sendLine(self, text)
            self._idlePingCall.reset(self._outQueue.pingAfterSendTime or
                                     self.idlePingTime)
            # Be careful: the following can recurse back to here!
            deferred.callback(None)

    def _sendIdlePing(self):
        self._idlePingCall = reactor.callLater(self.idlePingTime,
                                               self._sendIdlePing)
        if not self.loggedIn:
            # Just noop this out.
            # _sendLines assumes _idlePingCall is a LoopingCall,
            # so we do want it set already in connectionMade.
            # And we do get here if identd is firewalled
            # (connecting takes long enough then).
            return

        self.sendLine('PING %s' % (self.user.server,), tag='ping')
        # sendLine probably reset our idlePingCall to
        # pingAfterSendTime. Just re-reset it here:
        self._idlePingCall.reset(self.idlePingTime)
        # TODO: match the pong to this ping and use it to calculate
        # latency.

    def _tagReplyReceived(self, tag):
        """Call when receiving the server reply matching a tagged message.

        Returns a bool indicating if the tag reply was set correctly.
        """
        try:
            queue = self._tags[tag]
        except KeyError:
            return False
        for i, replyReceived in enumerate(queue):
            if replyReceived is None:
                queue[i] = time()
                # We may be able to send pending outbound messages now:
                self._outQueue.emptyQueue()
                return True
        return False

    ## handlers for messages from the ircd

    def irc_ERR_NICKNAMEINUSE(self, prefix, nickname, message):
        """A nick we changed or our initial nick is taken."""
        # last arg is a human-readable message
        if self.loggedIn:
            self.runHandlers(ibot.INickInUseHandler['nickInUse'],
                             self, nickname)
        else:
            for handler in self.handlers.get(ibot.ILoginNickInUseHandler, ()):
                try:
                    newNick = handler.loginNickInUse(self, nickname)
                except KeyboardInterrupt:
                    raise
                except Exception:
                    log.err()
                    newNick = None
                if newNick is not None:
                    self.user.nickname = newNick
                    self.setNick(newNick)
                    break
            else:
                self.quit('no nick available')

    def irc_JOIN(self, prefix, channel):
        """Someone, either we or another user, joined."""
        # System join messages are not accepted.
        user = ibot.IUser(prefix)
        if user is self.user:
            # rfc2812 says our nick MUST be in RPL_NAMREPLY so do
            # further bookkeeping there.
            chanObj = self.user.channels[channel] = Channel(self, channel)
            self.runHandlers(ibot.IJoinHandler['joined'], chanObj)
        else:
            channel = self.user.channels[channel]
            nick = user.nickname
            # This user is now "seen" by us
            self.users[nick] = user
            user.channels[channel.name] = channel
            channel.users[nick] = user
            self.runHandlers(ibot.IUserJoinedHandler['userJoined'],
                             user, channel)

    def irc_PART(self, prefix, channel, reason=None):
        """Someone, either we or another user, parted."""
        # System part messages are not accepted.
        user = ibot.IUser(prefix)
        channel = user.channels[channel]
        if user is self.user:
            self._weLeft(channel)
        else:
            self._userLeft(user, channel)
            self.runHandlers(ibot.IUserPartedHandler['userParted'],
                             user, channel, reason)

    def irc_QUIT(self, prefix, reason):
        """Callback for a user quitting irc."""
        # must be from a user
        user = ibot.IUser(prefix)
        if user is self.user:
            log.err('huh, got our own quit message?')
            return
        try:
            del self.users[user.nickname]
        except KeyError:
            log.err("An unexisting user %s just quit!" % (user.nickname,))
            return
        self.runHandlers(ibot.IUserQuitHandler['userQuit'], user, reason)
        for channel in user.channels.values():
            self.runHandlers(ibot.IUserLeftHandler['userLeft'], user, channel)
            del channel.users[user.nickname]

    def irc_PING(self, prefix, server1, server2=None):
        """Reply to a ping from the ircd."""
        # I don't quite understand what the rfc wants me to do here...
        if server2 is not None:
            log.msg('huh, extra data with ping: %s %s' % (server1, server2))
        self.sendLine("PONG %s %s" % (self.user.nickname, server1))

    def irc_PONG(self, prefix, server1, server2=None):
        """The server pongs a ping from us."""
        if not self._tagReplyReceived('ping'):
            log.err('unsolicited pong (%r, %r)' % (server1, server2))

    def irc_NICK(self, prefix, newname):
        """Someone (possibly us) changed nicks."""
        # Must be from a user. Notice this user can be us!
        user = ibot.IUser(prefix)
        oldname = user.nickname
        try:
            del self.users[oldname]
        except KeyError:
            log.err("Unknown user renamed from %s to %s!" %
                    (user.nickname, newname))
            return
        self.users[newname] = user
        user.name = newname
        for channel in user.channels.values():
            del channel.users[oldname]
            channel.users[newname] = user

    def irc_KICK(self, prefix, channel, kicked, message=None):
        """Callback for a kick.

        Figures out if we or someone else got kicked, updates state
        and calls handlers.
        """
        # Must be from a user. Notice the user can be us!
        kicker = ibot.IUser(prefix)
        channel = self.user.channels[channel]
        kicked = self.getUser(kicked)
        self.runHandlers(ibot.IKickedHandler['kicked'],
                         kicker, channel, kicked, message)
        if kicked is self.user:
            self._weLeft(channel)
        else:
            self._userLeft(kicked, channel)

    def irc_TOPIC(self, prefix, channel, topic):
        """Callback for a topic change by us or someone else."""
        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err('topicchange %r ignored' % (channel,))
            return
        self.runHandlers(ibot.ITopicHandler['topicChanged'],
                         prefix, channel, topic)
        channel.topic = topic

    def irc_RPL_TOPIC(self, prefix, channel, topic):
        """Callback with the current topic (onjoin and after TOPIC query."""
        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err('RPL_NOTOPIC %r ignored' % (channel,))
            return
        self.runHandlers(ibot.ITopicHandler['topicState'], channel, topic)
        channel.topic = topic

    def irc_RPL_NOTOPIC(self, prefix, channel, message):
        """Callback for a currently unset topic."""
        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err('RPL_NOTOPIC %r ignored' % (channel,))
            return
        # This is equivalent to '', not None! See the rfc.
        self.runHandlers(ibot.ITopicHandler['topicState'], channel, '')
        channel.topic = ''

    def irc_333(self, prefix, channel, setter, timestamp):
        """Hyperion's RPL_TOPICWHOTIME (sent after joining or TOPIC query)."""
        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err('RPL_TOPICWHOTIME %r ignored' % (channel,))
            return
        self.runHandlers(ibot.ITopicHandler['topicStateExtended'],
                         setter, timestamp)

    def irc_RPL_NAMREPLY(self, prefix, chantype, channel, nicks):
        """Callback for a 'names' request: add users to the channel."""
        # chantype is @ for secret, * for private and = for normal
        channelObj = self.user.channels[channel]
        for nick in nicks.split():
            modes = []
            while nick[0] in self._prefixToMode:
                prefix, nick = nick[0], nick[1:]
                modes.append(self._prefixToMode[prefix])
                # We could "break" here if we are not on a server with
                # CAP multi-prefix, but it seems more likely that servers
                # use multi-prefix unannounced than that they support nicks
                # that start with a prefix char.
            user = self.getUser(nick)
            # We "see" this user now.
            self.users[nick] = user
            channelObj.users[nick] = user
            user.channels[channel] = channelObj
            for mode in modes:
                user.chanmodes[mode][channel] = channelObj

    def irc_RPL_ENDOFNAMES(self, prefix, channel, message):
        """Callback for a 'names' request: ignored."""
        channelObj = self.user.channels[channel]
        channelObj.synced = True

    def irc_RPL_WHOREPLY(self, prefix, channel, username, hostname, server,
                         nickname, herestatus, hopAndRealname):
        """Callback for a 'who' query: set data and call back deferreds."""
        # channel is a channel the target is on
        # herestatus is a H or a G (for here or away)
        # hopAndRealname is '%s %s' % (hopcount, real name)
        if not self._whos:
            log.err('ignoring unrequested WHO')
            return
        pending = self._whos[0]
        # If this is a WHO for a nonvisible user reuse the user object:
        if self.caseFolder(pending[0]) == self.caseFolder(nickname):
            user = pending[2]
        else:
            user = self.getUser(nickname)
        user.username = username
        user.hostname = hostname
        user.server = server
        if herestatus[0] == 'H':
            user.away = False
        elif herestatus[0] == 'G':
            user.away = True
        else:
            log.msg('unknown who status snippet %r' % herestatus)
        try:
            channelObj = self.user.channels[channel]
        except KeyError:
            pass
        else:
            # Is this ever undesirable?
            for mode in self.supported.getFeature('PREFIX'):
                user.chanmodes[mode].pop(channel, None)
            for c in herestatus[1:]:
                mode = self._prefixToMode.get(c)
                if mode is not None:
                    user.chanmodes[mode][channel] = channelObj
                # Like in NAMREPLY we could break here but it seems
                # not doing so will not hurt and may help.
        user.hopcount, user.realname = hopAndRealname.split(' ', 1)

    def irc_RPL_ENDOFWHO(self, prefix, name, message):
        """Callback for a 'who' query."""
        for i, (reqName, d, obj) in enumerate(self._whos):
            if self.caseFolder(reqName) == self.caseFolder(name):
                del self._whos[i]
                d.callback(obj)
                return
        log.err('unrequested ENDOFWHO %r' % (name,))

    def irc_RPL_WHOISUSER(
        self, prefix, nickname, username, hostname, star, realname):
        """Callback for the first response line for a 'whois' query."""
        self.whoisResult = {
            'nickname': nickname,
            'username': username,
            'hostname': hostname,
            'realname': realname,
            }

    def _getWhois(self, nickname, reason, missing_ok=False):
        if self.whoisResult is None:
            if not missing_ok:
                log.msg('No current whoisResult for %s' % (reason,))
            return None
        nickname = self.caseFolder(nickname)
        if self.caseFolder(self.whoisResult['nickname']) != nickname:
            log.msg('Nickname mismatch: %r in %s, %r in whoisResult' %
                    (nickname, reason, self.whoisResult['nickname']))
            return None
        return self.whoisResult

    def irc_RPL_WHOISSERVER(self, prefix, nickname, server, info):
        """Callback for the server part of a whois query."""
        result = self._getWhois(nickname, 'WHOISSERVER')
        if result is not None:
            self.whoisResult['server'] = server
            self.whoisResult['serverinfo'] = info

    def irc_RPL_WHOISIDLE(self, prefix, nickname, *args):
        result = self._getWhois(nickname, 'WHOISIDLE')
        if result is None:
            return

        arglist = list(args)
        # human-readable-message, ignored.
        arglist.pop()
        self.whoisResult['idle'] = int(arglist.pop(0))
        if len(arglist) == 1:
            # Assume this is signon time (it is on seven).
            self.whoisResult['signon'] = datetime.datetime.utcfromtimestamp(
                int(arglist[0]))
        else:
            log.msg('Confusing WHOISIDLE %r' % (args,))

    def irc_RPL_WHOISCHANNELS(self, prefix, nickname, channels):
        """Callback for a channel list of a whois query (may get multiple)."""
        result = self._getWhois(nickname, 'WHOISCHANNELS')
        if result is not None:
            self.whoisResult.setdefault('channels', []).extend(
                channels.split())

    def irc_RPL_AWAY(self, prefix, nickname, message):
        """Callback for the away message."""
        if ibot.IUser.providedBy(prefix):
            # Do not blow up on this for now.
            log.err('huh, RPL_AWAY from non-server')

        result = self._getWhois(nickname, 'RPL_AWAY', True)
        if result is None:
            # We received this from the server because we msgd someone.
            user = self.getUser(nickname)
            user.away = True
            user.awayMessage = message
        else:
            result['away'] = message

    def irc_320(self, prefix, nickname, message):
        """Hyperion's RPL_WHOISIDENTIFIED."""
        result = self._getWhois(nickname, '320/RPL_WHOISIDENTIFIED')
        if result is not None:
            self.getUser(nickname).identified = True
            self.whoisResult['identified'] = message

    def irc_330(self, prefix, nickname, account, message):
        result = self._getWhois(nickname, '330/RPL_WHOISLOGGEDIN')
        if result is not None:
            self.whoisResult['account'] = account

    def irc_378(self, prefix, nickname, host):
        result = self._getWhois(nickname, '378/RPL_WHOISHOST')
        if result is not None:
            self.whoisResult['host'] = host

    def irc_671(self, prefix, nickname, message):
        result = self._getWhois(nickname, '671/RPL_WHOISSECURE')
        if result is not None:
            self.whoisResult['ssl'] = True

    def irc_RPL_ENDOFWHOIS(self, prefix, nickname, message):
        """Callback for the end of a whois reply."""
        result = self._getWhois(nickname, 'RPL_ENDOFWHOIS')
        if result is None:
            return

        target = self._whois.get(nickname)
        if target is None:
            log.msg("I didn't whois %r" % nickname)
            return

        # XXX this code is a little smelly
        self.whoisResult = None
        while target:
            target.pop().callback(result)

    def irc_ERR_NOSUCHNICK(self, prefix, nickname, message):
        """Callback for a (generic) 'nosuchnick' error."""
        # TODO do more with this
        whois = self._whois.pop(nickname, [])
        for d in whois:
            d.errback(ibot.NoSuchUser(nickname))

    def irc_RPL_MYINFO(self, prefix, server, version, usermodes, chanmodes,
                       *args):
        """Essential server info."""
        # TODO: maybe use the other args, which seems to include modes
        # that need an argument.
        # We use this to send "proper" pings
        self.user.server = server
        self.runHandlers(ibot.IStatisticsHandler['myinfo'],
                         server, version, usermodes, chanmodes)

    def irc_RPL_ISUPPORT(self, prefix, *args):
        """Callback for the server's ISUPPORT."""
        # I do not wish to pass the bot into the parser...

        # This is a little clunky, but should cause us to send our CAPABs
        # to the server once if it announces support.
        hadCapab = self.supported.hasFeature('CAPAB')
        userlistmodes = set(self.supported.getFeature('PREFIX'))
        chanmodes = self.supported.getFeature('CHANMODES').copy()

        # Last arg is a human-readable message.
        self.supported.parse(args[:-1])

        if not hadCapab and self.supported.hasFeature('CAPAB'):
            self.sendLine('CAPAB IDENTIFY-MSG')
            self.sendLine('CAPAB IDENTIFY-CTCP')

        mapping = self.supported.getFeature('CASEMAPPING', 'rfc1459')
        try:
            caseFolder = CASE_FOLDERS[mapping]
        except KeyError:
            log.err("unknown casemapping %r" % (mapping,))
        else:
            if self.caseFolder is not caseFolder:
                self.caseFolder = caseFolder
                # Complain if we already have dicts using the old folder.
                # We could go and refold the dicts, but doing so is insane.
                # (we already have us as user at this point, but that's ok)
                if len(self.users) > 1 or self.user.channels:
                    raise NotImplementedError(
                        'Case folding changed while seeing users or channels')
                self.users.refold(self.caseFolder)
                self._whois.refold(self.caseFolder)
                self.user.channels.refold(self.caseFolder)
                for modedict in self.user.chanmodes.itervalues():
                    modedict.refold(self.caseFolder)

        if set(self.supported.getFeature('PREFIX')) != userlistmodes:
            if self.user.channels or len(self.users) > 1:
                raise NotImplementedError('PREFIX changed while connected')
            # We are not in any channels yet, so just recreate this:
            self.user.chanmodes = {}
            for mode in self.supported.getFeature('PREFIX'):
                self.user.chanmodes[mode] = util.PreservingFoldingDict(
                    self.caseFolder)

            self._prefixToMode = dict(
                (prefix, modechar)
                for modechar, (prefix, priority)
                in self.supported.getFeature('PREFIX').iteritems())

        if (self.supported.getFeature('CHANMODES') != chanmodes
            and self.user.channels):
            raise NotImplementedError(
                'CHANMODES changed while seeing channels')

    def irc_290(self, prefix, *args):
        """Hyperion's RPL_CLIENTCAPAB: extra client protocol stuff."""
        for capab in args:
            if "=" in capab:
                capab, val = capab.split("=", 1)
            else:
                val = None
            self.clientCapab[capab] = val

    def irc_RPL_BANLIST(self, prefix, channel, mask, setter, setTime):
        """Callback for a banlist request: add bans to the channel object."""
        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err("RPL_BANLIST for '%s' which I'm not in" % channel)
            return
        channel.modes['b'][mask] = (setter, setTime)

    def irc_RPL_ENDOFBANLIST(self, prefix, channel, message):
        """Callback for the end of a banlist: ignored."""

    def irc_MODE(self, prefix, channel, modes, *args):
        """Callback for a channel- or usermode change.

        Splits it up into changes of a single mode and calls _modeChanged.
        """
        if modes[0] not in '-+':
            log.err('invalid MODE message (no +/-)')
            return
        if self.caseFolder(channel) == self.caseFolder(self.user.nickname):
            # handle user mode change
            for mode in modes:
                if mode == '+':
                    modeSet = True
                elif mode == '-':
                    modeSet = False
                else:
                    self.runHandlers(ibot.IUserModeHandler['userModeChanged'],
                                     self, modeSet, mode)
                    if modeSet:
                        self.modes.add(mode)
                    else:
                        self.modes.remove(mode)
            return

        try:
            channel = self.user.channels[channel]
        except KeyError:
            log.err("got a mode for a channel I'm not in: %r!" % (channel,))
            return

        # turn this tuple into a list so we can pop
        argslist = list(args)
        # list of (modeSet, mode, arg-or-None) tuples
        splitup = []
        # split up the modechange and bail if we can't parse it before
        # changing anything
        for mode in modes:
            chanmodes = self.supported.getFeature('CHANMODES')
            if mode == "+":
                modeSet = True
            elif mode == "-":
                modeSet = False
            elif (mode in self.supported.getFeature('PREFIX') or
                  mode in chanmodes['addressModes'] or
                  mode in chanmodes['param'] or
                  (mode in chanmodes['setParam'] and modeSet)):
                # this one takes a parameter, munge it
                if not argslist:
                    log.err("modechange didn't have enough parameters!")
                    return
                splitup.append((modeSet, mode, argslist.pop(0)))
            elif (mode in chanmodes['noParam'] or
                  mode in chanmodes['setParam']):
                splitup.append((modeSet, mode, None))
            else:
                log.err("mode %s unhandled!" % mode)
                return

        if argslist:
            log.err("leftover args after parsing modechange!")
            return
        # now do the actual applying
        for modeSet, mode, arg in splitup:
            self._modeChanged(prefix, channel, modeSet, mode, arg)

    def irc_PRIVMSG(self, prefix, target, message):
        """Callback for receiving a PRIVMSG."""
        self._handleNoticeOrMsg(prefix, target, message,
                                ibot.ISystemMessageHandler['systemMessage'],
                                ibot.ICTCPQueryHandler['ctcpQuery'],
                                ibot.IMessageHandler['message'])

    def irc_NOTICE(self, prefix, target, message):
        """Callback for receiving a NOTICE."""
        self._handleNoticeOrMsg(prefix, target, message,
                                ibot.ISystemMessageHandler['systemNotice'],
                                ibot.ICTCPReplyHandler['ctcpReply'],
                                ibot.INoticeHandler['notice'])

    def _handleNoticeOrMsg(self, prefix, target, message,
                           systemMethod, ctcpMethod, messageMethod):
        """Handle a NOTICE or PRIVMSG.

        This handles the +/- from IDENTIFY-MSG and parses ctcp stuff.
        prefix, target and message are what is passed to the irc_ method.
        systemMethod, ctcpMethod and messageMethod are z.i IMethod objects.
        """
        user = ibot.IUser(prefix, None)
        if self.caseFolder(target) == self.caseFolder(self.user.nickname):
            source = user
        else:
            source = self.user.channels.get(target)

        # check for server support for IDENTIFY-MSG if we have a user.
        if user is not None and message and (
            'identify-msg' in self.activeCap or (
                'IDENTIFY-MSG' in self.clientCapab and
                'IDENTIFY-CTCP' in self.clientCapab)):
            if message[0] in '+-':
                user.identified = (message[0] == '+')
                message = message[1:]
            else:
                log.err("huh, no +- on message %r" % (message,))
        # We used to complain about missing +- on messages from
        # the server, but:
        # - an "identified server" does not make any sense
        # - at least the server notice for your hostname being cloaked is
        #   sent without a +-
        # so we now expect those to not have a +-. TODO: look out for
        # messages for which user is None that do have the +-.

        # Handle as a system message if the source is not a user or the
        # target is not us or a channel (example: $* for global notices)
        if user is None or source is None:
            self.runHandlers(systemMethod, self, prefix, target, message)
        elif message.startswith(X_DELIM):
            # Just parse as ctcp if it starts with this char. This is
            # *much* simpler than the "ctcpspec" but matches irssi and xchat.
            message = message[1:]
            if message.endswith(X_DELIM):
                message = message[:-1]
            message = ctcpDequote(message)
            if ' ' in message:
                tag, data = message.split(' ', 1)
            else:
                tag, data = message, None
            self.runHandlers(ctcpMethod, user, source, tag, data)
        else:
            self.runHandlers(messageMethod, user, source, message)

    def irc_470(self, prefix, channelRequested, channelJoined, message):
        """Hyperion/seven's ERR_LINKCHANNEL: Forwarded to another channel."""
        # TODO call some handler here.
        # XXX this used to be handled the same way JOIN (of us) was:
        # create a Channel and store it in self.user.channels,
        # then call IJoinHandler.joined.
        # For seven that seems to be wrong: it sends us a regular
        # JOIN after the 470, where we replace the Channel object
        # and call IJoinHandler.joined again.
        # It's possible (but seems unlikely) we need to handle this as join
        # on some other ircd.

    def irc_CAP(self, prefix, nick, subcommand, *args):
        # This is based mainly on what seven actually sends me,
        # partially on www.leeh.co.uk/draft-mitchell-irc-capabilities-02.html

        # I do not see this in the draft rfc but seven sends it.
        # It's "*" if it has not seen our NICK message yet.
        if nick not in ('*', self.user.nickname):
            log.msg('ignoring CAP first arg %r' % (nick,))

        try:
            handler = getattr(self, 'cap_%s' % (subcommand,))
        except AttributeError:
            log.msg('unhandled CAP %r %r %r %r' %
                    (prefix, nick, subcommand, args))
        else:
            handler(args)

    def cap_LS(self, args):
        more, caps = _parseCap(args)

        self.serverCap.update(caps)

        if more:
            return

        supported = self.supportedCap & set(self.serverCap)
        if supported:
            self.sendLine('CAP REQ :%s' % (' '.join(supported),))
        else:
            self.sendLine('CAP END')

    def cap_ACK(self, args):
        more, caps = _parseCap(args)

        # XXX behavior for the more = True case is blatantly wrong
        self.activeCap.update(caps)

        if not more:
            if self.saslMechanisms and 'sasl' in self.activeCap:
                self._startSasl()
            else:
                self.sendLine('CAP END')

    def _startSasl(self):
        mech = self.saslMechanisms[-1]
        self.saslMechanism = SASL_MECHANISMS[mech](self.factory.saslAccount,
                                                   self.factory.saslPassword)
        self.sendLine('AUTHENTICATE %s' % (mech,))

    def cap_NAK(self, args):
        # TODO: handle this better (try with fewer caps?).
        # We should not currently be hitting this: the caps we support
        # are orthogonal and we only request them if the server claims
        # to support them.
        log.msg('server CAP NAK with %r, continuing CAP-less' % (args,))
        self.sendLine('CAP END')

    def irc_AUTHENTICATE(self, prefix, data):
        if data == '+':
            data = ''
        else:
            data = b64decode(data)

        reply = self.saslMechanism.received(data)
        self.sendLine('AUTHENTICATE %s' % (b64encode(reply),))

    def irc_900(self, prefix, mask, account, message):
        # RPL_LOGGEDIN.
        # Only saw this when doing SASL so far.
        nick, sep, rest = mask.partition('!')
        username, rest, hostname = rest.partition('@')
        # TODO: do something useful with the above.

    def irc_903(self, prefix, message):
        # RPL_SASLSUCCESS.
        self.sendLine('CAP END')

    def irc_904(self, prefix, message):
        # RPL_SASLFAIL
        self.saslMechanisms.pop()
        if self.saslMechanisms:
            self._startSasl()
        else:
            self.sendLine('CAP END')

    @defer.inlineCallbacks
    def irc_RPL_WELCOME(self, prefix, line):
        """This is the first thing we receive if our login is accepted."""
        # TODO: consider tweaking the order we do these things in.
        self.loggedIn = True
        self.factory.bot = self
        self.users[self.user.nickname] = self.user
        for handler in self.handlers.get(ibot.ILoginHandler, ()):
            try:
                result = yield handler.loggedIn(self)
            except KeyboardInterrupt:
                raise
            except Exception:
                log.err()
                result = False
            if not result:
                self.quit()
                break
        self.factory.resetDelay()
        self.runHandlers(ibot.IMOTDHandler['welcome'], line)

    def irc_RPL_MOTDSTART(self, prefix, line):
        """Start of message of the day along with a header line of sorts."""
        self.runHandlers(ibot.IMOTDHandler['motdStart'], line)

    def irc_RPL_MOTD(self, prefix, line):
        """Message of the day body."""
        self.runHandlers(ibot.IMOTDHandler['motd'], line)

    def irc_RPL_ENDOFMOTD(self, prefix, line):
        """End of message of the day with an (ignored) footer."""
        self.runHandlers(ibot.IMOTDHandler['motdEnd'])

    def irc_RPL_YOURHOST(self, prefix, line):
        """Human-readable host description."""
        self.runHandlers(ibot.IStatisticsHandler['yourhost'], line)

    def irc_RPL_CREATED(self, prefix, line):
        """Human-readable creation timestamp."""
        self.runHandlers(ibot.IStatisticsHandler['created'], line)

    def irc_RPL_LUSERCLIENT(self, prefix, line):
        """Human-readable network client count."""
        self.runHandlers(ibot.IStatisticsHandler['luserclient'], line)

    def irc_RPL_LUSEROP(self, prefix, count, line):
        """Number of ircops."""
        # Human-readable description tag is ignored.
        try:
            count = int(count)
        except ValueError:
            log.err()
        else:
            self.runHandlers(ibot.IStatisticsHandler['luserop'], count)

    def irc_RPL_LUSERUNKNOWN(self, prefix, count, line):
        """Number of "unknown" connections."""
        # Human-readable description tag is ignored.
        try:
            count = int(count)
        except ValueError:
            log.err()
        else:
            self.runHandlers(ibot.IStatisticsHandler['luserunknown'], count)

    def irc_RPL_LUSERCHANNELS(self, prefix, count, line):
        """Number of channels."""
        # Human-readable description tag is ignored.
        try:
            count = int(count)
        except ValueError:
            log.err()
        else:
            self.runHandlers(ibot.IStatisticsHandler['luserchannels'], count)

    def irc_RPL_LUSERME(self, prefix, line):
        """Human-readable count of clients on this server."""
        self.runHandlers(ibot.IStatisticsHandler['luserclient'], line)

    def irc_250(self, prefix, line):
        """Hyperion's RPL_STATSCONN: highest connection count."""
        self.runHandlers(ibot.IStatisticsHandler['maxConnectionCount'], line)

    def irc_265(self, prefix, *args):
        """Hyperion's RPL_LOCALUSERS."""
        # TODO: newer ircds send not just a human-readable message but also
        # two machine-readable numbers. Use them.
        line = args[-1]
        self.runHandlers(ibot.IStatisticsHandler['localUsers'], line)

    def irc_266(self, prefix, *args):
        """Hyperion's RPL_GLOBALUSERS."""
        # TODO: newer ircds send not just a human-readable message but also
        # two machine-readable numbers. Use them.
        line = args[-1]
        self.runHandlers(ibot.IStatisticsHandler['globalUsers'], line)

    def irc_396(self, prefix, vhost, line):
        # RPL_HOSTHIDDEN, according to seven.
        self.runHandlers(ibot.IHostHiddenHandler['hostHidden'], self, vhost)
        self.user.hostname = vhost

    # stuff called from the above methods.

    def ctcpQuery(self, user, source, tag, data):
        """Handle some very common ctcp queries in the bot itself."""
        if tag == 'ACTION':
            self.runHandlers(ibot.IActionHandler['action'],
                             user, source, data)

    def _weLeft(self, channel):
        """Callback when a channel is left. channel is a Channel object.

        This does maintenance on user and channel objects and dicts.
        """
        self.runHandlers(ibot.ILeftHandler['left'], channel)
        for nick, user in channel.users.iteritems():
            del user.channels[channel.name]
            for modemap in user.chanmodes.itervalues():
                modemap.pop(channel.name, None)
            if not user.channels and user is not self.user:
                # We no longer "see" this user.
                del self.users[nick]

    def ctcpReply(self, user, source, tag, data):
        """Handle some very common ctcp replies in the bot itself."""

    def _modeChanged(self, prefix, channel, modeSet, mode, arg=None):
        """Actually handle a single modechange.

        This calls handlers and updates the channel object.
        """
        self.runHandlers(ibot.IChannelModeHandler['channelModeChanged'],
                         prefix, channel, modeSet, mode, arg)

        if mode in self.supported.getFeature('PREFIX'):
            target = self.users.get(arg)
            if target is None:
                log.err('EEP! got a mode for unknown user %s!' % (arg,))
                return
            if modeSet:
                target.chanmodes[mode][channel.name] = channel
            else:
                target.chanmodes[mode].pop(channel.name, None)
        elif modeSet:
            if mode in self.supported.getFeature('CHANMODES')['addressModes']:
                if ibot.IUser.providedBy(prefix):
                    # username and hostname are set for a prefix
                    setter = '%s!%s@%s' % (
                        prefix.nickname, prefix.username, prefix.hostname)
                else:
                    setter = prefix
                channel.modes[mode][arg] = (setter, time())
            else:
                channel.modes[mode] = arg
        else:
            if mode in self.supported.getFeature('CHANMODES')['addressModes']:
                channel.modes[mode].pop(arg, None)
            else:
                channel.modes.pop(mode, None)

    def _userLeft(self, user, channel):
        """Callback for a user leaving a channel we are on.

        This does maintenance on our user and channel objects and dicts.
        """
        self.runHandlers(ibot.IUserLeftHandler['userLeft'], user, channel)
        del user.channels[channel.name]
        del channel.users[user.nickname]
        if not user.channels:
            # we're not seeing this one, get rid of it
            del self.users[user.nickname]


    ## methods for sending to the ircd

    def away(self, message=''):
        """Set our away message."""
        if message:
            self.sendLine('AWAY :%s' % (message,))
        else:
            self.sendLine('AWAY')

    def mode(self, enabledModes='', disabledModes=''):
        """Set user mode."""
        deferreds = []
        while enabledModes or disabledModes:
            mix = []
            modesPerMessage = self.supported.getFeature('MODES')
            enableBatch = enabledModes[:modesPerMessage]
            enabledModes = enabledModes[modesPerMessage:]
            if enableBatch:
                mix.extend(['+', enableBatch])
            length = modesPerMessage - len(enableBatch)
            disableBatch = disabledModes[:length]
            disabledModes = disabledModes[length:]
            if disableBatch:
                mix.extend(['-', disableBatch])
            deferreds.append(self.sendLine(
                    'MODE %s %s' % (self.user.nickname, ''.join(mix))))
        return defer.gatherResults(deferreds)

    def setNick(self, nickname):
        """Change our nick."""
        self.sendLine("NICK %s" % nickname)

    def quit(self, message=''):
        """Quit from the ircd."""
        self.sendLine("QUIT :%s" % message)

    def join(self, channel, key=None):
        """Join a channel. 'key' is sent to join the channel, if given."""
        if key:
            self.sendLine("JOIN %s %s" % (channel, key))
        else:
            self.sendLine("JOIN %s" % (channel,))

    ## other functions

    # ibot stuff

    def getUser(self, nick):
        """Return a user object from a nick."""
        try:
            return self.users[nick]
        except KeyError:
            user = User(self, nick)
            self.runHandlers(ibot.INewUserHandler['newUser'], user)
            return user

    # non-irc misc

    def runHandlers(self, imethod, *args, **kwargs):
        """Call an IMethod on handlers implementing its interface.

        If one of the handlers failed this logs and continues.
        """
        handlers = []
        if imethod.interface.providedBy(self):
            handlers.append(self)
        handlers.extend(self.handlers.get(imethod.interface, ()))
        for handler in handlers:
            try:
                getattr(handler, imethod.__name__)(*args, **kwargs)
            except KeyboardInterrupt:
                raise
            except Exception:
                log.err()


class IRCBotFactory(protocol.ReconnectingClientFactory):

    """Factory that reconnects and keeps track of handlers.

    'bot' is the bot instance or None if disconnected.
    'handlers' is a mapping handler interface -> list of handlers.
    """

    implements(ibot.IBotFactory)

    protocol = IRCBot

    def __init__(self, nickname, username=None, realname="Twisted Mess",
                 password=None, cap_support=True,
                 saslMechanisms=(), saslAccount=None, saslPassword=None):
        """Initialize.

        name is the bot's nickname.
        realname is the bot's realname.

        saslMechanisms is a sequence of sasl mechanism names.
        The preferred mechanism should be last.
        saslAccount is the sasl account name, defaulting to the nickname.
        saslPassword is the sasl password.
        """
        if saslMechanisms and not cap_support:
            raise ValueError('sasl builds on cap')
        for mech in saslMechanisms:
            if mech not in SASL_MECHANISMS:
                raise ValueError('%r unsupported' % (mech,))
        if saslAccount is None:
            saslAccount = nickname
        if username is None:
            username = nickname
        self.nickname = nickname
        self.username = username
        self.realname = realname
        self.password = password
        self.cap_support = cap_support
        self.saslMechanisms = saslMechanisms
        self.saslAccount = saslAccount
        self.saslPassword = saslPassword
        self.handlers = {}
        self.bot = None

    def addHandler(self, handler):
        """Register a new handler with the bot."""
        for interface in providedBy(handler):
            if interface.extends(ibot.IBotHandler):
                self.handlers.setdefault(interface, []).append(handler)

    def removeHandler(self, handler):
        """Unregister a handler completely."""
        for handlers in self.handlers.itervalues():
            if handler in handlers:
                handlers.remove(handler)
