
"""Database-backed logging."""


import datetime

from zope.interface import implements

from storm import locals as storm, uri

from mzbot import ibot, commands


class Message(object):

    """A single on-channel message."""

    __storm_table__ = 'message'

    id = storm.Int(primary=True)
    # these are bytes to allow people to use zany non-utf8 nicks
    nick = storm.RawStr()
    source = storm.RawStr()
    message = storm.Unicode()
    timestamp = storm.DateTime()

    def __init__(self, nick, source, message):
        self.nick = nick
        self.source = source
        self.message = message


class PrivateMessage(storm.Storm):

    """Single privmsg."""

    __storm_table__ = 'privmsg'

    id = storm.Int(primary=True)
    nick = storm.RawStr()
    message = storm.Unicode()
    timestamp = storm.DateTime()

    def __init__(self, nick, message):
        self.nick = nick
        self.message = message


class ChannelModeChange(storm.Storm):

    """A modechange."""

    __storm_table__ = 'modechange'

    id = storm.Int(primary=True)
    nickname = storm.RawStr()
    username = storm.RawStr()
    hostname = storm.RawStr()
    channel = storm.RawStr()
    modeSet = storm.Bool()
    mode = storm.RawStr()
    arg = storm.RawStr()
    timestamp = storm.DateTime()


class Logger(commands.ConfigurableMixin):

    """Log irc events to a database."""

    # TODO: log more stuff?

    implements(ibot.IMessageHandler, ibot.INoticeHandler, ibot.IActionHandler,
               ibot.IChannelModeHandler, ibot.ILogger)

    def __init__(self):
        self.set_path(u'None')

    def set_path(self, args):
        path, = commands.parseArgs(args, commands.bytestring)
        if path == 'None':
            dburi = uri.URI('sqlite:?timeout=0')
        else:
            dburi = uri.URI('sqlite:%s?timeout=0' % (uri.escape(path),))

        db = storm.Store(storm.create_database(dburi))

        db.execute('CREATE TABLE IF NOT EXISTS [message] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [nick] BLOB NOT NULL,'
                   ' [source] BLOB NOT NULL,'
                   ' [message] TEXT NOT NULL,'
                   ' [timestamp] TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP)')
        db.execute('CREATE INDEX IF NOT EXISTS [message_nick] '
                   'ON [message] ([nick])')
        db.execute('CREATE INDEX IF NOT EXISTS [message_source] '
                   'ON [message] ([source])')
        db.execute('CREATE INDEX IF NOT EXISTS [message_timestamp] '
                   'ON [message] ([timestamp])')

        db.execute('CREATE TABLE IF NOT EXISTS [privmsg] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [nick] BLOB NOT NULL,'
                   ' [message] TEXT NOT NULL,'
                   ' [timestamp] TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP)')
        db.execute('CREATE INDEX IF NOT EXISTS [privmsg_nick] '
                   'ON [privmsg] ([nick])')
        db.execute('CREATE INDEX IF NOT EXISTS [privmsg_timestamp] '
                   'ON [privmsg] ([timestamp])')

        db.execute('CREATE TABLE IF NOT EXISTS [modechange] '
                   '([id] INTEGER PRIMARY KEY,'
                   ' [nickname] BLOB NOT NULL,'
                   ' [username] BLOB,'
                   ' [hostname] BLOB,'
                   ' [channel] BLOB NOT NULL,'
                   ' [modeSet] INTEGER NOT NULL,'
                   ' [mode] BLOB NOT NULL,'
                   ' [arg] BLOB NOT NULL,'
                   ' [timestamp] TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP)')
        db.execute('CREATE INDEX IF NOT EXISTS [modechange_nickname] '
                   'ON [modechange] ([nickname])')
        db.execute('CREATE INDEX IF NOT EXISTS [modechange_channel] '
                   'ON [modechange] ([channel])')
        db.execute('CREATE INDEX IF NOT EXISTS [modechange_timestamp] '
                   'ON [modechange] ([timestamp])')

        self.store = db

    def message(self, user, source, message):
        """Log a message or the like."""
        message = unicode(message, 'utf-8', 'replace')
        if user is source:
            obj = PrivateMessage(nick=user.nickname, message=message)
        else:
            obj = Message(
                nick=user.nickname, source=source.name, message=message)
        self.store.add(obj)
        self.store.flush()

    notice = action = message

    def channelModeChanged(self, prefix, channel, modeSet, mode, arg=None):
        """Log a modechange."""
        user = ibot.IUser(prefix, None)
        obj = ChannelModeChange()
        if user is None:
            obj.nickname = prefix
        else:
            obj.nickname = user.nickname
            obj.username = user.username
            obj.hostname = user.hostname

        obj.channel = channel.name
        obj.modeSet = modeSet
        obj.mode = mode
        obj.arg = arg

        self.store.add(obj)
        self.store.flush()


class SeenHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """Report where someone was and/or is."""

    def __init__(self, **kwargs):
        commands.CommandHandler.__init__(self)
        self.channels = ''

    def _getDelta(self, timestamp):
        """Given a datetime object return a timedelta from now.

        This is separate so the tests can override it.
        """
        return datetime.datetime.utcnow() - timestamp

    @commands.args(commands.bytestring)
    def handle_seen(self, user, source, nick):
        """seen <nick>: when is or was nick seen."""
        loggers = user.bot.handlers.get(ibot.ILogger, ())
        if not loggers:
            source.msgTo(user, u'cannot find logger')
            return
        if len(loggers) > 1:
            source.msgTo(user,
                         u'multiple loggers registered, this is unsupported')
            return
        store = loggers[0].store
        nick = nick.strip()
        if nick in user.bot.users:
            currentChannels = list(
                chan.decode('ascii') for chan in user.bot.users[nick].channels
                if chan in self.channels.split())
            currentChannels.sort()
        else:
            currentChannels = []
        message = store.find(
            Message,
            Message.nick == nick,
            Message.source.is_in(self.channels.split())
            ).order_by(storm.Desc(Message.timestamp)).first()
        nick = nick.decode('ascii')
        if message is None:
            if currentChannels:
                source.msgTo(
                    user, u'%s is currently on %s' % (
                        commands.bold(nick),
                        u', '.join(commands.bold(channel)
                                   for channel in currentChannels)))
            else:
                source.msgTo(user, u"I haven't seen %s" % commands.bold(nick))
            return
        delta = commands.formatTime(self._getDelta(message.timestamp))
        if currentChannels:
            source.msgTo(
                user,
                u"%s%s was last seen %s ago saying '%s' in %s, "
                'and is currently on %s.' % (
                    commands.resetstyle(),
                    commands.bold(nick),
                    commands.underline(delta),
                    commands.colorize(message.message, 'green'),
                    commands.colorize(
                        message.source.decode('ascii'), 'purple'),
                    u', '.join(commands.bold(channel)
                               for channel in currentChannels)))
        else:
            source.msgTo(
                user,
                u"%s%s was last seen %s ago saying '%s' in %s, "
                'and not currently on any channels I know.' %
                (commands.resetstyle(),
                 commands.bold(nick),
                 commands.underline(delta),
                 commands.colorize(message.message, 'green'),
                 commands.colorize(message.source.decode('ascii'), 'purple')))

    def get_channels(self):
        """Get a list of channels we search logs for and users on."""
        return u', '.join(sorted(c.decode('ascii')
                                 for c in self.channels.split()))

    def set_addchannel(self, args):
        """Add a channel to the list of searched channels."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        channels.add(channel)
        self.channels = ' '.join(channels)

    def set_removechannel(self, args):
        """Remove a channel from the list of searched channels."""
        channel, = commands.parseArgs(args, commands.bytestring)
        channels = set(self.channels.split())
        channels.discard(channel)
        self.channels = ' '.join(channels)
