
"""Quotes db."""

from storm import locals as storm, uri

from mzbot import commands

# hackity hackity HACK:
# current versions of storm do not properly escape some sqlite-specific
# reserved words. I found out because I have a column called "index" below.
# This fixes it. It will hopefully be fixed in storm 0.17, but adding
# reserved words twice is a noop, so this will not hurt there.
from storm.databases import sqlite
sqlite.compile.add_reserved_words('''
    abort after analyze attach autoincrement before conflict database detach
    each exclusive explain fail glob if ignore index indexed instead isnull
    limit notnull offset plan pragma query raise regexp reindex release
    rename replace row savepoint temp trigger vacuum virtual
    '''.split())


class Group(storm.Storm):

    """A group of quotes."""

    __storm_table__ = 'group'

    id = storm.Int(primary=True)
    name = storm.Unicode()

    def __init__(self, name):
        self.name = name


class Quote(storm.Storm):

    """A single quote."""

    __storm_table__ = 'quote'
    __storm_primary__ = 'group_id', 'index'

    quote = storm.Unicode()
    source = storm.RawStr()
    group_id = storm.Int()
    group = storm.Reference(group_id, Group.id)
    index = storm.Int()
    timestamp = storm.DateTime()

    def __init__(self, quote, source, group, index):
        self.quote = quote
        self.source = source
        self.group = group
        self.index = index


class QuoteHandler(commands.CommandHandler, commands.ConfigurableMixin):

    """Store and display quotes."""

    configurables = {
        u'group': (u'Group name of quotes we access',
                   'group', u'%s', commands.string),
        }

    def __init__(self):
        commands.CommandHandler.__init__(self)
        self.set_path(u'None')
        self.set_group(u'ungrouped')

    def set_path(self, args):
        path, = commands.parseArgs(args, commands.bytestring)
        if path == 'None':
            dburi = uri.URI('sqlite:?timeout=0')
        else:
            dburi = uri.URI('sqlite:%s?timeout=0' % (uri.escape(path),))

        db = storm.Store(storm.create_database(dburi))
        db.execute('CREATE TABLE IF NOT EXISTS [group] '
                   '([id] INTEGER PRIMARY KEY, [name] TEXT NOT NULL)')
        db.execute('CREATE UNIQUE INDEX IF NOT EXISTS [group_name] '
                   'ON [group] ([name])')
        db.execute('CREATE TABLE IF NOT EXISTS [quote] '
                   '([quote] TEXT NOT NULL,'
                   ' [source] BLOB NOT NULL,'
                   ' [group_id] INTEGER NOT NULL REFERENCES [group] ([id])'
                   '  ON DELETE CASCADE,'
                   ' [index] INTEGER NOT NULL,'
                   ' [timestamp] TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP)')
        # This could do with more indices but meh.
        self.store = db

    def set_group(self, args):
        name, = commands.parseArgs(args, commands.string)
        group = self.store.find(Group, Group.name == name).one()
        if group is None:
            group = self.store.add(Group(name))
        self.group = group

    def get_group(self):
        return self.group.name

    def _query(self, query):
        """Return a list of quotes matching query."""
        return list(
            u'%d: %s' % (quote.index, quote.quote.strip())
            for quote in self.store.find(
                Quote,
                Quote.group == self.group,
                Quote.quote.contains_string(query.strip())))

    @commands.args(commands.longstring)
    def handle_numquotes(self, user, source, query):
        """numquotes <query>: return the number of quotes matching <query>"""
        source.msgTo(user, u'%d quotes match' % (len(self._query(query)),))

    @commands.args(commands.longstring)
    def handle_quotelist(self, user, source, query):
        """quotelist <query>: return a list of all quotes matching <query>"""
        result = self._query(query)
        if len(result) > 10:
            source.msgTo(user, u"more than 10 matches, refusing to display")
        elif not result:
            source.msgTo(user, u"no matches")
        else:
            source.msgTo(user, u"'%s'" % u"', '". join(result))

    @commands.args(commands.longstring)
    def handle_pmquotelist(self, user, source, query):
        """pmquotelist <query>: pm all matching quotes (might take a while)."""
        query = query.strip()
        if not query:
            source.msgTo(user, u"need a query")
            return
        result = self._query(query)
        if not result:
            source.msgTo(user, u"no matches")
        else:
            for quote in result:
                user.msg(quote)

    @commands.args(commands.integer)
    def handle_numquote(self, user, source, num):
        """numquote <int>: return a quote by number."""
        quote = self.store.find(
            Quote, Quote.group == self.group, Quote.index == num).one()
        if quote is None:
            source.msgTo(user, u"I don't have that many quotes")
        else:
            source.msgTo(user, quote.quote)

    @commands.optargs(0, commands.longstring)
    def handle_quote(self, user, source, query=None):
        """Give a random quote, or a quote with the extra args in it."""
        if not query:
            quotes = self.store.find(Quote, Quote.group == self.group)
        else:
            quotes = self.store.find(
                Quote,
                Quote.group == self.group,
                Quote.quote.contains_string(query.strip()))

        quotes.order_by('RANDOM()')
        quote = quotes.first()
        if quote is not None:
            source.msgTo(user, u'%d: %s' % (quote.index, quote.quote.strip()))
        else:
            source.msgTo(
                user, u"No matches, if you wanted to add use 'addquote'")

    @commands.args(commands.longstring)
    def handle_addquote(self, user, source, quote):
        """addquote <quote>: Add a quote to the quotedb."""
        try:
            maxQuote = self.store.find(
                Quote, Quote.group == self.group
                ).order_by(storm.Desc(Quote.index)).first()
            if maxQuote is None:
                index = 1
            else:
                index = maxQuote.index + 1
            quote = self.store.add(Quote(
                quote=quote, source=user.nickname, index=index,
                group=self.group))
        except:
            self.store.rollback()
            raise
        else:
            self.store.commit()

        source.msgTo(user, u'quote %d added' % (quote.index,))
