
"""Various simple handlers used for demonstration purposes."""

import random
import urllib

from zope.interface import implements

from twisted.python import util, reflect, rebuild
from twisted.web import client

from mzbot import commands, ibot, ebot


class PingPongHandler(commands.ConfigurableMixin):

    """Respond to pings and the like."""

    implements(ibot.ICommandHandler)

    def __init__(self):
        self.replies = {}
        self.commandset = set()

    def handle(self, user, source, command, args):
        """Give a reply only if there are no extra args."""
        if args:
            return
        replies = self.replies[command]
        source.msgTo(user, random.choice(replies))

    def get_replies(self):
        """List replies."""
        return u', '.join(
            (u'%s: %s' % (ping, u', '.join(pong for pong in pongs)))
            for ping, pongs in self.replies.iteritems())

    def set_removereply(self, val):
        """removereply <key> <reply text>: remove a reply."""
        ping, pong = commands.parseArgs(
            val, commands.string, commands.longstring)
        try:
            replies = self.replies[ping]
            replies.remove(pong)
        except (KeyError, ValueError):
            raise ebot.InvalidArgument('no such reply')
        # check if we 'lost' a command
        if not replies:
            self.commandset.remove(ping)
            del self.replies[ping]

    def set_addreply(self, val):
        """addreply <key> <reply text>: add a new reply."""
        ping, pong = commands.parseArgs(val,
                                        commands.string, commands.longstring)
        self.replies.setdefault(ping, []).append(pong)
        self.commandset.add(ping)


def _getAndPrintTopic(func):
    """get topic if needed, apply func, then msg the new topic."""
    def _new_f(self, user, channel, *args, **kwargs):
        if channel.name not in self.topic:
            topic = channel.topic
            if topic is None:
                topic = ""
            self.topic[channel.name] = topic.split(
                self.separators.get(channel.name, self.separator))
        value = func(self, user, channel, *args, **kwargs)
        if value is not False:
            user.notice(
                "new topic: %s" %
                ", ".join("(%s: %s)" % (i, part)
                          for i, part in enumerate(self.topic[channel.name])))
    return util.mergeFunctionMetadata(func, _new_f)


class TopicHandler(commands.CommandHandler):

    """Manage the topic."""

    separator = " | "

    def __init__(self):
        commands.CommandHandler.__init__(self)
        self.topic = {}
        self.separators = {}

    @commands.notInPrivmsg
    @commands.args()
    def handle_getseparator(self, user, channel):
        """Get the per-channel separator used in 'get' and 'apply'"""
        separator = self.separators.get(channel.name, self.separator)
        channel.msgTo(user, u'separator is "%s"' % (
                separator.encode('string-escape').decode('ascii'),))

    @commands.notInPrivmsg
    @commands.args(commands.bytestring)
    def handle_setseparator(self, user, channel, separator):
        """Set the per-channel separator used in 'get' and 'apply'"""
        self.separators[channel.name] = separator
        self.handle_getseparator(user, channel, "")

    @commands.notInPrivmsg
    @commands.args()
    def handle_get(self, user, channel):
        """Store the current topic in the bot."""
        if channel.name in self.topic:
            del self.topic[channel.name]
        self.handle_show(user, channel, "")

    @commands.notInPrivmsg
    @commands.args(commands.integer, commands.longbytestring)
    @_getAndPrintTopic
    def handle_insert(self, user, channel, position, text):
        """insert nr text: Insert text into the topic at that position."""
        self.topic[channel.name].insert(position, text)

    @commands.notInPrivmsg
    @commands.args(commands.integer, commands.longbytestring)
    @_getAndPrintTopic
    def handle_set(self, user, channel, position, text):
        """set nr text: set topic part nr to text"""
        try:
            self.topic[channel.name][position] = text
        except IndexError:
            channel.msgTo(user, u"index error")
            return False

    @commands.notInPrivmsg
    @commands.args(commands.integer)
    @_getAndPrintTopic
    def handle_remove(self, user, channel, position):
        """remove nr: remove a part from the topic."""
        try:
            del self.topic[channel.name][position]
        except IndexError:
            channel.msgTo(user, u"index error")
            return False

    @commands.notInPrivmsg
    @commands.args()
    @_getAndPrintTopic
    def handle_show(self, user, channel):
        """Show the current recorded topic."""
        # heh, let the decorators do the actual work
        pass

    @commands.notInPrivmsg
    @commands.args()
    def handle_apply(self, user, channel):
        """Apply the new topic."""
        if channel.name in self.topic:
            channel.setTopic(
                self.separators.get(channel.name, self.separator).join(
                    self.topic[channel.name]))
        else:
            channel.msgTo(user, u"get and modify the topic first")

    @commands.notInPrivmsg
    @commands.kwargs([('position', commands.integer),
                      ('current', commands.bytestring),
                      ('replacement', commands.bytestring)],
                     [('current', commands.bytestring),
                      ('replacement', commands.bytestring)])
    @_getAndPrintTopic
    def handle_replace(
        self, user, channel, current, replacement, position=None):
        """replace [position] <current> <replacement>: search and replace."""
        topic = self.topic[channel.name]
        if position is None:
            for i, piece in enumerate(topic):
                topic[i] = piece.replace(current, replacement)
        else:
            topic[position] = topic[position].replace(current, replacement)


class BackdoorHandler(commands.CommandHandler):

    """rebuild loaded modules."""

    @commands.args(commands.bytestring)
    def handle_rebuild(self, user, source, module):
        """rebuild module: rebuild target module."""
        try:
            moduleObj = reflect.namedModule(module)
        except ImportError, e:
            source.msgTo(user, unicode(e))
        else:
            rebuild.rebuild(moduleObj)
            source.msgTo(user, u'done')

    @commands.args(commands.longstring)
    def handle_eval(self, user, source, expr):
        """Eval <expr>: eval code."""
        def out(put):
            source.msgTo(user, put)
        try:
            eval(compile(expr, '<string>', 'single'), {'out': out}, {})
        except Exception, e:
            source.msgTo(user, u'Miserable failure: %s' % (e,))
        else:
            source.msgTo(user, u'Great success')


class CommandRemapper(commands.ConfigurableMixin):

    """Remap command names."""

    # XXX this is currently not accessible as a config plugin.

    implements(ibot.ICommandHandler, ibot.ICommandHelpProvider)

    def __init__(self, original):
        self.original = original
        self.map = {}
        self.commandset = set()

    def handle(self, user, source, command, args):
        """Pass on to the underlying handler."""
        target = self.map[command]
        if target not in self.original.commandset:
            source.msgTo(
                user, u'remapping error: target command "%s" disappeared' % (
                    target,))
        else:
            return self.original.handle(user, source, target, args)

    def getCommandDescription(self, command):
        """Pass the request on to our underlying handler if possible."""
        original = ibot.ICommandHelpProvider(self.original, None)
        if original is None:
            return 'undocumented'
        originalCommand = self.map[command]
        if originalCommand not in self.original.commandset:
            return 'invalid command'
        docstring = original.getCommandDescription(originalCommand)
        return docstring

    def get_remaps(self):
        """Currently applied remaps."""
        return u', '.join(
            (u'%s: %s' % (key.encode('unicode-escape').decode('ascii'),
                          val.encode('unicode-escape').decode('ascii')))
            for (key, val) in self.map.iteritems())

    def set_remove(self, args):
        """remove <input>: remove a remap."""
        remap, = commands.parseArgs(args, commands.string)
        self.commandset.discard(remap)
        self.map.pop(remap, None)

    def set_add(self, args):
        """add <input> <target>: remap a command."""
        originalCommand, target = commands.parseArgs(
            args, commands.string, commands.string)
        if target not in self.original.commandset:
            raise ebot.InvalidArgument('not a valid target command')
        self.commandset.add(originalCommand)
        self.map[originalCommand] = target


class ShortenHandler(commands.CommandHandler):

    """Use the metamark service to shorten urls."""

    # Tests override this. I doubt there is anything else that needs to.
    url = 'http://metamark.net/api/rest/simple'

    @commands.args(commands.bytestring)
    def handle_shorten(self, user, source, url):
        """shorten <url>: shorten an url."""
        postdata = urllib.urlencode(dict(long_url=url))
        headers = {'content-type': 'application/x-www-form-urlencoded'}
        page = client.getPage(self.url, method='POST',
                              headers=headers, postdata=postdata)
        page.addCallback(self._shortened, user, source)
        return page

    def _shortened(self, result, user, source):
        """Announce the result."""
        # No %r or quotes here, some clients misbehave on copy/paste for those.
        source.msgTo(user, u'Shortened to %s' % (result.decode('ascii'),))
