# do not check docstrings, __init__, missing member
# pylint: disable-msg=C0111, W0232, E1101

"""Tests for the ban manager."""


from zope.interface import verify

from twisted.trial import unittest

from mzbot import banman, testutils, ibot


class BanManagerTest(unittest.TestCase):

    guessedHeader = (u'Say "newincident [incident]" to track as new, '
                     u'"addoffender <incident> to add to existing, "forget" '
                     u'to prune from the db.')
    unknownHeader = (u'Say "newincident <incident>" to track as new, '
                     u'"addoffender <incident> to add to existing, "forget" '
                     u'to prune from the db.')

    def setUp(self):
        testutils.setUp(self)
        self.handler = banman.BanManagementHandler()
        self.handler.set(u'announcechannel', u'empty')
        self.handler.set(u'addchannel', u'normal')

    def testInterfaces(self):
        for interface in (ibot.ICommandHandler,
                          ibot.IConfigurable,
                          ibot.IChannelModeHandler,
                          ibot.IJoinHandler):
            verify.verifyObject(interface, self.handler)

    def testBan(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: me!*@*'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

    def testQuiet(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '%me!*@*')
        self.empty.assertEvents(
            ('notice', u'New quiet: me!*@*'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

    def testForward(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*!#gentoo-ops')
        self.empty.assertEvents(
            ('notice', u'New forward: me!*@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

    def testNoneAffected(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'nobody!*@*!#gentoo-ops')
        self.empty.assertEvents(
            ('notice', u'New forward: nobody!*@* to #gentoo-ops'),
            ('notice', u'No current users affected! ' + self.unknownHeader))

    def testMultipleAffected(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: *!*@*'),
            ('notice', u'3 affected. ' + self.unknownHeader))

    def testBriefUnset(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: *!*@*'),
            ('notice', u'3 affected. ' + self.unknownHeader))
        self.handler.channelModeChanged(self.user, self.normal, False, 'b',
                                        '*!*@*')
        self.empty.assertEvents(('notice', u'Removing brief ban "*!*@*"'))

    def testRedundantUnset(self):
        self.handler.channelModeChanged(self.user, self.normal, False, 'b',
                                        'whatever!is@this')
        self.empty.assertEvents()

    def testIrrelevantMode(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'z',
                                        '*!*@*')
        self.empty.assertEvents()

    def testUntrackedChannel(self):
        self.handler.channelModeChanged(self.user, self.empty, True, 'b',
                                        '*!*@*')
        self.empty.assertEvents()

    def testStringSetter(self):
        self.handler.channelModeChanged('server!', self.normal, True, 'b',
                                        'me!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: me!*@*'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))
        # TODO: verify the Ban object got a sensible setter

    def testBanSetTwice(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: me!*@*'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*')
        self.empty.assertEvents()

    def testNewIncident(self):
        self.handler.handle_newincident(self.user, self.user, '')
        self.user.assertEvents(
            ('msgTo', self.user, u'Give me a name or ban someone.'))
        self.empty.assertEvents()

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*!#gentoo-ops')
        self.empty.assertEvents(
            ('notice', u'New forward: me!*@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))
        self.user.assertEvents()

        self.handler.handle_newincident(self.user, self.user, '')
        self.user.assertEvents(
            ('msgTo', self.user, u'Using name "me" from last ban'))
        self.empty.assertEvents(
            ('notice', u'Incident "me" created'),
            ('notice', u'Offender "me" added to incident "me"'),
            ('notice', u'Ban "me!*@*!#gentoo-ops" added to incident "me"'))

        self.handler.handle_whybanned(self.user, self.user, u'me')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'using mask me!alsome@here from online user'),
            ('msgTo', self.user, u'Incident: "me"'),
            ('msgTo', self.user, u'Offenders: me!alsome@here'),
            ('msgTo', self.user, u'Ban: me!*@*!#gentoo-ops '
             u'(set by me 0 seconds ago)'))

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Matches existing incident "me", say "addtoincident" '
             u'to add this ban to it'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))
        self.user.assertEvents()

        self.handler.handle_newincident(self.user, self.user, '')
        self.user.assertEvents(
            ('msgTo', self.user, u'Using name "me" from last ban'),
            ('msgTo', self.user, u'An incident with that name exists.'))
        self.empty.assertEvents()

        self.handler.handle_addtoincident(self.user, self.user, u'missing')
        self.empty.assertEvents()
        self.user.assertEvents(
            ('msgTo', self.user, u'No incident with that name'))

        self.handler.handle_addtoincident(self.user, self.user, '')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Ban "*!alsome@*!#gentoo-ops" added to incident "me"'))

    def testNewDescriptionMultiBan(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!*@*')
        self.empty.assertEvents(
            ('notice', u'New ban: *!*@*'),
            ('notice', u'3 affected. ' + self.unknownHeader))
        self.handler.handle_newincident(self.user, self.user, '')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'Give me a name or ban exactly one client.'))

    def testNewDescriptionNoOffenderMove(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_newincident(self.user, self.user, u'test')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Incident "test" created'),
            ('notice', u'Offender "me" added to incident "test"'),
            ('notice',
             u'Ban "*!alsome@*!#gentoo-ops" added to incident "test"'))

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New ban: me!*@*'),
            ('notice', u'Matches existing incident "test", '
             'say "addtoincident" to add this ban to it'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_newincident(self.user, self.user, u'')
        self.user.assertEvents(
            ('msgTo', self.user, u'Using name "me" from last ban'))
        self.empty.assertEvents(
            ('notice', u'Incident "me" created'),
            ('notice', u'Offender "me" not moved from incident "test" to "me"'),
            ('notice',
             u'Ban "me!*@*" added to incident "me"'))

    def testAddOffender(self):
        self.handler.handle_newincident(self.user, self.user, u'test')
        self.user.assertEvents()
        self.empty.assertEvents(('notice', u'Incident "test" created'))

        self.handler.handle_addoffender(self.user, self.user, u'test')
        self.user.assertEvents(('msgTo', self.user, u'no offenders to add.'))

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_addoffender(self.user, self.user, u'nottest')
        self.user.assertEvents(('msgTo', self.user, u'No such incident!'))

        self.handler.handle_addoffender(self.user, self.user, u'test')
        self.user.assertEvents(('msgTo', self.user, u'No such incident!'))

    def testWhyBannedTargetSearch(self):
        self.handler.handle_whybanned(self.user, self.user, u'alsome')
        self.empty.assertEvents()
        self.user.assertEvents(('msgTo', self.user, u'Nothing found!'))

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_newincident(self.user, self.user, u'test')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Incident "test" created'),
            ('notice', u'Offender "me" added to incident "test"'),
            ('notice',
             u'Ban "*!alsome@*!#gentoo-ops" added to incident "test"'))

        self.handler.handle_whybanned(self.user, self.user, u'alsome')
        self.empty.assertEvents()
        # XXX this test is racy...
        self.user.assertEvents(
            ('msgTo', self.user, u'Incident: "test"'),
            ('msgTo', self.user, u'Offenders: me!alsome@here'))

    def testBanMatchesOffender(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_newincident(self.user, self.user, u'test')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Incident "test" created'),
            ('notice', u'Offender "me" added to incident "test"'),
            ('notice',
             u'Ban "*!alsome@*!#gentoo-ops" added to incident "test"'))

        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        'me!*@*')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New ban: me!*@*'),
            ('notice', u'Matches existing incident "test", '
             'say "addtoincident" to add this ban to it'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_addtoincident(self.user, self.user, u'')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Ban "me!*@*" added to incident "test"'))

    def testDescription(self):
        self.handler.channelModeChanged(self.user, self.normal, True, 'b',
                                        '*!alsome@*!#gentoo-ops')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'New forward: *!alsome@* to #gentoo-ops'),
            ('notice', u'Affected: me!alsome@here. ' + self.guessedHeader))

        self.handler.handle_newincident(self.user, self.user, u'test')
        self.user.assertEvents()
        self.empty.assertEvents(
            ('notice', u'Incident "test" created'),
            ('notice', u'Offender "me" added to incident "test"'),
            ('notice',
             u'Ban "*!alsome@*!#gentoo-ops" added to incident "test"'))

        self.handler.handle_setdescription(self.user, self.user,
                                           u'"multi missing" stuff')
        self.user.assertEvents(
            ('msgTo', self.user, u'No incident with that name'))

        self.handler.handle_setdescription(self.user, self.user,
                                           u'"test" he is Evil')
        self.user.assertEvents(
            ('msgTo', self.user, u'Description set.'))

        self.handler.handle_whybanned(self.user, self.user, u'me')
        self.user.assertEvents(
            ('msgTo', self.user, u'using mask me!alsome@here from online user'),
            ('msgTo', self.user, u'Incident: "test" (he is Evil)'),
            ('msgTo', self.user, u'Offenders: me!alsome@here'),
            ('msgTo', self.user,
             u'Ban: *!alsome@*!#gentoo-ops (set by me 0 seconds ago)')
            )

    def testConfigurable(self):
        self.assertEquals(u'normal', self.handler.get(u'channels'))
        self.handler.set(u'removechannel', u'normal')
        self.assertEquals(u'', self.handler.get(u'channels'))
        self.handler.set(u'addchannel', u'normal')
        self.assertEquals(u'normal', self.handler.get(u'channels'))

        self.assertEquals(u'14 days', self.handler.get(u'expiry'))
        self.handler.set(u'expiry', u'%d' % (60*60,))
        self.assertEquals(u'1 hour', self.handler.get(u'expiry'))

    def test_unknownbans(self):
        self.handler.handle_unknownbans(self.user, self.user, u'')
        self.empty.assertEvents()
        self.user.assertEvents(
            ('msgTo', self.user, u'You may have to run this one twice.'))
        self.normal.assertEvents(('mode', True, 'b', None))

        self.normal.modes['b']['*!alsome@*'] = ('someone', 42)

        self.handler.handle_unknownbans(self.user, self.user, u'')
        self.empty.assertEvents()
        self.user.assertEvents(
            ('msgTo', self.user, u'You may have to run this one twice.'),
            ('msgTo', self.user, u'normal: *!alsome@*'))
        self.normal.assertEvents(('mode', True, 'b', None))
