# do not check docstrings, __init__, missing members
# pylint: disable-msg=C0111,W0232,E1101

"""Tests for commands.py."""


import datetime

from zope.interface import verify, implements, Interface

from twisted.internet import defer
from twisted.trial import unittest

from mzbot import commands, ibot, ebot, testutils


class MyHandlerException(Exception):
    pass

class MyHandler(commands.CommandHandler):

    """Test handler.

    Docstrings or lack of them is part of the test.
    """

    def __init__(self):
        commands.CommandHandler.__init__(self)

    def handle_return(self, user, channel, args):
        """I return."""
        return self, user, channel, args

    def handle_raise(self, *args):
        """raise: I raise."""
        raise MyHandlerException('raised')

    def handle_succeed(self, user, channel, args):
        """succeed foo bar: I succeed."""
        return defer.succeed(args)

    def handle_fail(self, *args):
        return defer.fail(MyHandlerException('failed'))


class MyConfigurable(MyHandler, commands.ConfigurableMixin):

    """Once again docstrings are part of the test."""

    configurables = {
        u'\N{SNOWMAN}': ('I am bar', 'spork', u'bar is %s', commands.string),
        u'unsettable': ('I am unsettable', 'spork', u'%s', None),
        u'ungettable': ('I am ungettable', 'spork', None, commands.string),
        }

    def __init__(self):
        MyHandler.__init__(self)
        self.spork = None

    def get_foo(self):
        """Get foo."""
        return u'foo!'

    def set_spork(self, value):
        """Set spork."""
        self.spork = value


class CommandHandlerTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = MyHandler()

    def testInterfaces(self):
        verify.verifyObject(ibot.ICommandHandler, self.handler)
        verify.verifyObject(ibot.ICommandHelpProvider, self.handler)

    def testCommandsList(self):
        self.assertEquals(sorted(self.handler.commandset),
                          [u'fail', u'raise', u'return', u'succeed'])

    def testReturn(self):
        self.assertEquals(
            (self.handler, self.user, self.normal, u'bar'),
            self.handler.handle(self.user, self.normal, u'return', u'bar'))

    def testRaise(self):
        self.assertRaises(
            MyHandlerException,
            self.handler.handle, self.user, self.normal, u'raise', u'')

    @defer.inlineCallbacks
    def testSucceed(self):
        result = yield self.handler.handle(self.user, self.normal,
                                           u'succeed', u'bar')
        self.assertEquals(result, u'bar')

    def testFail(self):
        d = self.handler.handle(self.user, self.normal, u'fail', u'')
        self.failUnlessFailure(d, MyHandlerException)
        return d

    def testCommandDescription(self):
        self.assertEquals(
            self.handler.getCommandDescription(u'return'), 'I return.')
        self.assertEquals(
            self.handler.getCommandDescription(u'fail'), 'undocumented')

    def testConfigurable(self):
        handler = MyConfigurable()
        verify.verifyObject(ibot.IConfigurable, handler)
        self.assertEquals({
                u'foo': 'Get foo.',
                u'\N{SNOWMAN}': 'I am bar',
                u'unsettable': 'I am unsettable',
                }, handler.gettables())
        self.assertEquals({
                u'spork': 'Set spork.',
                u'\N{SNOWMAN}': 'I am bar',
                u'ungettable': 'I am ungettable',
                }, handler.settables())
        self.assertEquals(u'foo!', handler.get(u'foo'))
        handler.set(u'spork', u'zomg a spork')
        self.assertEquals(u'zomg a spork', handler.spork)
        self.assertEquals(u'bar is zomg a spork', handler.get(u'\N{SNOWMAN}'))
        handler.set(u'ungettable', u'"not a spork"')
        self.assertEquals(u'not a spork', handler.get(u'unsettable'))
        handler.set(u'\N{SNOWMAN}', u'whatever')
        self.assertEquals(u'whatever', handler.spork)


class SourceRestrictionTest(unittest.TestCase):

    """Tests for decorators that restrict where handlers can be called."""

    setUp = testutils.setUp

    def testNotInPrivmsg(self):
        # ibot says it's a privmsg iff "user is source"
        @commands.notInPrivmsg
        def noop(self, user, source, args):
            source.msgTo(user, u'worked')

        # in-channel message: worked to channel, nothing to user
        noop(None, self.user, self.normal, u'')
        self.normal.assertEvents(('msgTo', self.user, u'worked'))
        self.user.assertEvents()
        # privmsg: something to the user, but not 'worked'
        noop(None, self.user, self.user, u'')
        self.normal.assertEvents()
        self.user.assertEvents(
            ('msgTo', self.user, u"doesn't work in a privmsg"))

    def testOnlyInPrivmsg(self):
        @commands.onlyInPrivmsg
        def noop(self, user, source, args):
            source.msgTo(user, u'worked')

        # privmsg: should work
        noop(None, self.user, self.user, u'')
        self.user.assertEvents(('msgTo', self.user, u'worked'))
        # on-channel msg: should complain
        noop(None, self.user, self.normal, u'')
        self.user.assertEvents()
        self.normal.assertEvents(('msgTo', self.user,
                                  u'do this in a privmsg'))


class DummyAvatar(object):
    implements(ibot.IAvatar)

class ISmarter(Interface):
    """Marker interface."""

class SmarterAvatar(DummyAvatar):
    implements(ISmarter)

class LoginTest(unittest.TestCase):

    """Tests for decorators that check for being logged in."""

    setUp = testutils.setUp

    @commands.maybeLoggedIn(ibot.IAvatar)
    def avatar(self, user, source, avatar, args):
        return avatar

    @commands.maybeLoggedIn(ISmarter)
    def smarter(self, user, source, avatar, args):
        return avatar

    @commands.loggedIn(ISmarter)
    def onlySmarter(self, user, source, avatar, args):
        return avatar

    def testMaybeLoggedInNotLoggedIn(self):
        self.assertIdentical(None, self.avatar(self.user, self.user, u'foo'))
        self.user.assertEvents()
        self.assertIdentical(None, self.smarter(self.user, self.user, u'foo'))
        self.user.assertEvents()

    def testMaybeLoggedInRestricted(self):
        dummy = self.user.data[ibot.IAvatar] = DummyAvatar()
        self.assertIdentical(dummy, self.avatar(self.user, self.user, u'foo'))
        self.user.assertEvents()
        self.assertIdentical(None, self.smarter(self.user, self.user, u'foo'))
        self.user.assertEvents()

    def testMaybeLoggedInSmarter(self):
        dummy = self.user.data[ibot.IAvatar] = SmarterAvatar()
        self.assertIdentical(dummy, self.avatar(self.user, self.user, u'foo'))
        self.user.assertEvents()
        self.assertIdentical(dummy, self.smarter(self.user, self.user, u'foo'))
        self.user.assertEvents()

    def testLoggedInNotLoggedIn(self):
        self.assertIdentical(
            None, self.onlySmarter(self.user, self.user, u'foo'))
        self.user.assertEvents(('msgTo', self.user, u'you are not logged in!'))

    def testLoggedInRestricted(self):
        self.user.data[ibot.IAvatar] = DummyAvatar()
        self.assertEquals(
            None, self.onlySmarter(self.user, self.user, u'foo'))
        self.user.assertEvents(
            ('msgTo', self.user,
             u'you are logged in but do not have the needed access!'))

    def testLoggedInSmarter(self):
        dummy = self.user.data[ibot.IAvatar] = SmarterAvatar()
        self.assertIdentical(
            dummy, self.onlySmarter(self.user, self.user, u'foo'))
        self.user.assertEvents()


class ArgumentsTest(unittest.TestCase):

    """Tests for decorators that do argument parsing for a handler."""

    setUp = testutils.setUp

    def testKeywordArgs(self):
        # based on the docstring example
        @commands.kwargs(
            [("number", commands.integer), ("text", commands.string)],
            [("text", commands.string), ("number", commands.integer)],
            [("number", commands.integer)],
            [("text", commands.string)])
        def ident(self, user, source, number=None, text=None):
            return number, text
        # valid input, should not produce any user events
        for args, number, text in [(u'42', 42, None),
                                   (u'bar', None, u'bar'),
                                   (u'42 bar', 42, u'bar'),
                                   (u'bar 42', 42, u'bar')]:
            self.assertEquals(
                ident(None, self.user, self.user, args), (number, text))
        self.user.assertEvents()
        # failing input
        ident(None, self.user, self.user, u'foo bar')
        self.user.assertEvents(
            ('msgTo', self.user, u'one of the following: '
             u'"bar" is not an integer, '
             u'"foo" is not an integer, too many arguments'))
        ident(None, self.user, self.user, u'')
        self.user.assertEvents(
            ('msgTo', self.user, u'one of the following: not enough arguments'))

        # same as the above but string before int
        @commands.kwargs(
            [("number", commands.integer), ("text", commands.string)],
            [("text", commands.string), ("number", commands.integer)],
            [("text", commands.string)],
            [("number", commands.integer)])
        def identRev(self, user, source, number=None, text=None):
            return number, text
        self.assertEquals(
            identRev(None, self.user, self.user, u'42'), (None, u'42'))

    def testOptArgs(self):
        # based on the docstring example
        @commands.optargs(1, commands.integer, commands.string)
        def ident(self, user, source, number, text=None):
            return number, text
        # valid input, shouldn't cause user events
        self.assertEquals(ident(None, self.user, self.user, u'42'),
                          (42, None))
        self.assertEquals(
            ident(None, self.user, self.user, u'42 bar'), (42, u'bar'))
        self.user.assertEvents()
        # invalid input
        ident(None, self.user, self.user, u'bar')
        self.user.assertEvents(('msgTo', self.user,
                                u'"bar" is not an integer'))

    def testArgs(self):
        # based on the docstring example
        @commands.args(commands.integer, commands.string)
        def ident(self, user, source, number, text):
            return number, text
        # valid input, shouldn't cause user events
        self.assertEquals(
            ident(None, self.user, self.user, u'42 bar'), (42, u'bar'))
        self.user.assertEvents()
        for invalid, message in [
            (u'42', u'not enough arguments'),
            (u'bar', u'"bar" is not an integer'),
            (u'42 bar extra', u'too many arguments'),
            ]:
            ident(None, self.user, self.user, invalid)
            self.user.assertEvents(('msgTo', self.user, message))

    def testFilterRaisingForNoReason(self):
        def stupidFilter(data):
            raise ebot.InvalidArgument()

        @commands.args(stupidFilter)
        def identArgs(self, user, source, arg):
            """Never runs"""
        identArgs(None, self.user, self.user, u'die')
        self.user.assertEvents(('msgTo', self.user, u'invalid argument'))

        @commands.optargs(0, stupidFilter)
        def identOptArgs(self, user, source, arg=None):
            return arg
        self.assertEquals(identOptArgs(None, self.user, self.user, u''), None)
        self.user.assertEvents()
        identOptArgs(None, self.user, self.user, u'die')
        self.user.assertEvents(('msgTo', self.user, u'invalid argument'))

        @commands.kwargs(
            [],
            [('arg', stupidFilter)],
            )
        def identKwargs(self, user, source, arg=None):
            return arg
        self.assertEquals(identKwargs(None, self.user, self.user, u''), None)
        self.user.assertEvents()
        identKwargs(None, self.user, self.user, u'die')
        self.user.assertEvents(('msgTo', self.user, u'one of the following: '
                                u'invalid argument, too many arguments'))


class FilterTest(unittest.TestCase):

    """Test the various command argument filters."""

    # XXX TODO: rename this method (it's now unicode, not str)
    def _assertRaisesStr(self, string, func, *args, **kwargs):
        try:
            func(*args, **kwargs)
        except ebot.InvalidArgument, e:
            self.assertEquals(string, unicode(e))
        else: # pragma: no cover
            self.fail('%s not raised' % (string,))

    def testByteString(self):
        for string, output, remains in [
            (u'foo', 'foo', u''),
            (u'  foo bar ', 'foo', u'bar '),
            (u'"foo bar" baz', 'foo bar', u'baz'),
            (u'"\'"', "'", u''),
            (u'\\200 \200', '\200', u'\200'),
            (u'ba"r', 'ba"r', u''),
            (u'"\\""', '"', u''),
            (u'"\\t"', '\t', u''),
            ]:
            self.assertEquals(commands.bytestring(string), (output, remains))
        for string, message in [
            (u'"', u'""" is not properly quoted'),
            (u"'foo", u'"\'foo" is not properly quoted'),
            (u"'foo'bar", u'"\'foo\'bar" is not properly quoted'),
            (u'', u'argument is empty'),
            (u'\200 \200', u'"\x80 \x80" is not ascii'),
            (u'  ', u"argument is empty")]:
            self._assertRaisesStr(message, commands.bytestring, string)

    def testString(self):
        self.assertEquals(commands.string(u'\u02eb'), (u'\u02eb', u''))
        self.assertEquals(commands.string(u'c \u02eb'), (u'c', u'\u02eb'))

    def testLongString(self):
        for string in [u"foo bar 'baz", u'']:
            self.assertEquals(commands.longstring(string),
                              (string, ''))
        self.assertEquals(
            commands.longstring(unichr(747)),
            (unichr(747), ''))

    def testLongByteString(self):
        for string in [u'foo bar "baz \\xfe']:
            self.assertEquals(
                (string.encode('ascii').decode('string-escape'), u''),
                commands.longbytestring(string))

    def testStringList(self):
        for func in (commands.stringlist, commands.bytestringlist):
            for string, output in [
                (u'foo', ['foo']),
                (u'"f\'oo"  \'b"ar\'', ["f'oo", 'b"ar']),
                (u'ba"r', ['ba"r']),
                ]:
                if func == commands.stringlist:
                    output = [s.decode('ascii') for s in output]
                self.assertEquals((output, ''), func(string))
            for string, message in [
                (u'"', u'""" is not properly quoted'),
                (u"'foo", u'"\'foo" is not properly quoted'),
                (u'  ', u"argument is empty"),
                ]:
                self._assertRaisesStr(message, func, string)

    def testInteger(self):
        for string, output, rest in [
            (u'42 bottles', 42, u'bottles'),
            (u'123456789123456789', 123456789123456789, u''),
            ]:
            self.assertEquals(commands.integer(string), (output, rest))
        for string, message in [
            (u'bottles', u'"bottles" is not an integer'),
            (u'3.1415', u'"3.1415" is not an integer'),
            (u'', u'argument is empty'),
            ]:
            self._assertRaisesStr(message, commands.integer, string)

    def testRangedInteger(self):
        for func, string, message in [
            (commands.rangedInteger(maximum=42), u'43', u'out of range'),
            (commands.rangedInteger(minimum=44), u'43', u'out of range'),
            (commands.rangedInteger(1, 1000000), u'123456789', u'out of range')
            ]:
            self._assertRaisesStr(message, func, string)
        self.assertEquals((42, u''), commands.rangedInteger(maximum=42)(u'42'))
        self.assertEquals((42, u''), commands.rangedInteger(minimum=42)(u'42'))

    def testNumber(self):
        for string, output, rest in [
            (u'-3 deg', float('-3'), u'deg'),
            (u'12.345e6', float('12.345e6'), u''),
            ]:
            self.assertEquals(commands.number(string), (output, rest))
        for string, message in [
            (u'zomg', u'"zomg" is not a number'),
            (u'a7', u'"a7" is not a number'),
            ]:
            self._assertRaisesStr(message, commands.number, string)


class TextEffectsTest(unittest.TestCase):

    """Text modifier stuff."""

    def testColorizer(self):
        self.assertEquals(commands.colorize(u'foo', 'green'), u'\0033foo\017')
        self.assertEquals(commands.colorize(u'foo', 'white'), u'\0030foo\017')
        self.assertEquals(
            commands.colorize(u'foo', 'white', 'blue'), u'\0030,2foo\017')
        self.assertEquals(commands.colorize(u'foo', 1, 2), u'\0031,2foo\017')
        self.assertRaises(ValueError, commands.colorize, u'foo', 'invisible')
        self.assertRaises(ValueError, commands.colorize, u'foo', 'white', 'b')

    def testMisc(self):
        self.assertEquals(u'\x02bold\x02', commands.bold(u'bold'))
        self.assertEquals(u'\x16reverse\x16', commands.reverse(u'reverse'))
        self.assertEquals(u'\x1funderline\x1f',
                          commands.underline(u'underline'))
        self.assertEquals(u'\x0f', commands.resetstyle())


class FormatTimeTest(unittest.TestCase):

    def testFormatTime(self):
        timed = datetime.timedelta
        for delta, string in [
            (timed(days=1, seconds=2, minutes=3, hours=4),
             u'1 day, 4 hours, 3 minutes and 2 seconds'),
            (timed(days=5), u'5 days'),
            (timed(seconds=1, minutes=1), u'1 minute and 1 second'),
            (timed(seconds=1, minutes=2, hours=3),
             u'3 hours, 2 minutes and 1 second'),
            (timed(seconds=0), u'0 seconds'),
            ]:
            self.assertEquals(
                string, commands.formatTime(delta))
