

from __future__ import absolute_import, with_statement, division

import json

from zope.interface import implements

from twisted.trial import unittest
from twisted.internet import defer, protocol, error
from twisted.protocols import basic

from mzbot import conf, coreauth, irc, ibot


class AuthGrammarTest(unittest.TestCase):

    if conf.AuthParser is None: # pragma: no cover
        skip = 'pymeta not available?'

    handlers = {
        'always-true': ('at', coreauth.AlwaysTheSameAuthenticator(True)),
        'always-false': ('af', coreauth.AlwaysTheSameAuthenticator(False)),
        }

    def _assertDesc(self, desc, expr):
        auth = conf.parseAuth(self.handlers, expr)
        self.assertEqual(desc, auth.description())

    def _assertParseError(self, expr):
        self.assertRaises(conf.ParseError,
                          conf.parseAuth, self.handlers, expr)

    def test_one(self):
        self._assertDesc(u'always authenticated', 'always-true')

    def test_and(self):
        self._assertDesc(u'all of (always authenticated, always denied)',
                         'always-true and always-false')

    def test_or(self):
        self._assertDesc(u'one of (always authenticated, always denied)',
                         'always-true or always-false')

    def test_andand(self):
        self._assertDesc(
            u'all of (always authenticated, always denied, always denied)',
            'always-true and always-false and always-false')

    def test_oror(self):
        self._assertDesc(
            u'one of (always authenticated, always denied, always denied)',
            'always-true or always-false or always-false')

    def test_orand(self):
        self._assertDesc(
            u'one of (always authenticated, '
            u'all of (always denied, always denied))',
            'always-true or (always-false and always-false)')

    def test_andor(self):
        self._assertDesc(
            u'all of (always authenticated, '
            u'one of (always denied, always denied))',
            'always-true and (always-false or always-false)')

    def test_missing_parens_andor(self):
        self._assertParseError('always-true and always-false or always-false')

    def test_missing_parens_orand(self):
        self._assertParseError('always-true or always-false and always-false')


class OptionsTest(unittest.TestCase):

    def setUp(self):
        self.options = conf.Options()

    def test_basic_commandline(self):
        self.options.parseOptions(['botconf'])

    def test_explicit_json(self):
        self.options.parseOptions(['-l', 'json', 'botconf'])

    def test_args(self):
        self.options.parseOptions(['-a', '{}'])


class TestProtocol(basic.LineOnlyReceiver):

    def connectionMade(self):
        self.factory.inst = self
        self.received = []
        self.deferred = defer.Deferred()

    def lineReceived(self, line):
        self.received.append(line)
        # XXX HACK
        self.transport.write(':server 001 foonick :hi\r\n')

    def connectionLost(self, reason=None):
        self.deferred.callback(reason)


class LoginHandler(object):

    implements(ibot.ILoginHandler)

    def __init__(self):
        self.deferred = defer.Deferred()

    def loggedIn(self, bot):
        self.deferred.callback(None)
        return defer.succeed(True)


class ServiceMakerTest(unittest.TestCase):

    def setUp(self):
        self.options = conf.Options()

    @defer.inlineCallbacks
    def test_connect(self):
        # Weird high-level test that establishes an actual connection.
        # Stolen from the old service tests more because I wanted to keep
        # the code around than because it is a very useful test.
        serverfactory = protocol.ServerFactory()
        serverfactory.protocol = TestProtocol
        from twisted.internet import reactor
        listeningPort = reactor.listenTCP(0, serverfactory)
        self.addCleanup(listeningPort.stopListening)

        # XXX this could do with a cleaner way of setting
        oldTime = irc.OutQueue.lineTime
        def resetLineTime():
            irc.OutQueue.lineTime = oldTime
        self.addCleanup(resetLineTime)

        config = {
            u'handlers': {},
            u'networks': {
                u'local': {
                    u'handlers': [],
                    u'hostname': u'127.0.0.1',
                    u'nickname': u'foonick',
                    u'port': listeningPort.getHost().port,
                    },
                },
            }
        self.options.parseOptions(['-l', 'json', '-a', json.dumps(config)])
        service = conf.makeService(self.options)

        handler = LoginHandler()

        service.startService()
        botfactory = service.getServiceNamed(u'local').args[2]
        botfactory.continueTrying = False
        botfactory.addHandler(handler)

        yield handler.deferred

        service.stopService()

        self.assertFailure(serverfactory.inst.deferred, error.ConnectionDone)
        yield serverfactory.inst.deferred

        self.assertEqual(
            ['CAP LS', 'NICK foonick', 'USER foonick * * :Twisted Mess'],
            serverfactory.inst.received)

        botfactory.stopTrying()
