# do not check docstrings, __init__, missing member
# pylint: disable-msg=C0111, W0232, E1101

"""Tests for coreauth.py."""


from zope.interface import verify

from twisted.internet import defer, task
from twisted.trial import unittest

from mzbot import coreauth, ibot, testutils


class ChanServ(testutils.User):

    """Hack to respond to msgs to chanserv"""

    def msg(self, msg):
        self.d.callback(True)


class SimpleAuthenticatorTest(unittest.TestCase):

    setUp = testutils.setUp

    def checkIAuth(self, myAuth):
        verify.verifyObject(ibot.IAuthenticator, myAuth)
        self.assertIdentical(ibot.IAuthenticator(myAuth), myAuth)
        self.failUnless(myAuth.description())

    def testModeAuthenticator(self):
        opAuth = coreauth.ModeAuthenticator(channellist=['oper', 'normal'],
                                            mode='o')
        opAuth2 = coreauth.ModeAuthenticator(channellist=['voice', 'normal'],
                                             mode='o')
        self.checkIAuth(opAuth)
        self.assertEquals(u'mode +o in one of voice, normal',
                          opAuth2.description())
        return defer.gatherResults([
                opAuth.auth(self.user).addCallback(self.failUnless),
                opAuth2.auth(self.user).addCallback(self.failIf),
                ])

    @defer.inlineCallbacks
    def testAuthenticatorStack(self):
        allow = coreauth.AlwaysTheSameAuthenticator(status=True)
        deny = coreauth.AlwaysTheSameAuthenticator(status=False)
        for children, all, result in [
            ([allow, allow, allow], True, True),
            ([allow, deny, allow], True, False),
            ([allow, deny, allow], False, True),
            ([deny, deny], False, False),
            ]:
            stack = coreauth.AuthenticatorStack(children=children, all=all)
            self.checkIAuth(stack)
            authed = yield stack.auth(self.user)
            self.assertEquals(authed, result)

    def testAlwaysTheSameAuthenticator(self):
        dList = []
        for const in (True, False):
            constAuth = coreauth.AlwaysTheSameAuthenticator(status=const)
            self.checkIAuth(constAuth)
            d = constAuth.auth(self.user)
            d.addCallback(self.assertEquals, const)
            dList.append(d)
        return defer.gatherResults(dList)

    def testBaseNameAuthenticator(self):
        nameAuth = coreauth.BaseNameAuthenticator(["me!*some@*"])
        self.checkIAuth(nameAuth)
        nameAuth2 = coreauth.BaseNameAuthenticator(["*!other@here"])
        return defer.gatherResults([
                nameAuth.auth(self.user).addCallback(self.failUnless),
                nameAuth.auth(self.other).addCallback(self.failIf),
                nameAuth2.auth(self.other).addCallback(self.failUnless),
                nameAuth2.auth(self.user).addCallback(self.failIf),
                ])

    @defer.inlineCallbacks
    def testNameAuthenticator(self):
        nameAuth = coreauth.NameAuthenticator()
        self.checkIAuth(nameAuth)
        # Just make sure these do not raise exceptions (they did once).
        nameAuth.gettables()
        nameAuth.settables()
        self.assertEqual(u'', nameAuth.get(u'patterns'))
        authed = yield nameAuth.auth(self.user)
        self.failIf(authed)

        nameAuth.set(u'add', u'me!*some@*')
        nameAuth.set(u'remove', u'nonsense')
        authed = yield nameAuth.auth(self.user)
        self.failUnless(authed)

        nameAuth.set(u'remove', u'me!*some@*')
        authed = yield nameAuth.auth(self.user)
        self.failIf(authed)

    def testIdentifiedAuthenticator(self):
        idAuth = coreauth.IdentifiedAuthenticator()
        idAuthUnknown = coreauth.IdentifiedAuthenticator(allowUnknown=True)
        self.checkIAuth(idAuth)
        self.assertEquals(u'an identified user.', idAuth.description())
        self.assertEquals(u"Either identified or 'unknown'",
                          idAuthUnknown.description())

        d = defer.gatherResults([idAuth.auth(self.other),
                                 idAuth.auth(self.user),
                                 idAuthUnknown.auth(self.other),
                                 idAuthUnknown.auth(self.user)])
        def check(result):
            self.assertEquals([True, False, True, True], result)

        d.addCallback(check)
        return d

    @defer.inlineCallbacks
    def testChanServAuthenticator(self):
        chanserv = ChanServ('ChanServ', bot=self.bot)
        self.bot.users['chanserv'] = chanserv
        clock = task.Clock()
        csAuth = coreauth.ChanServAuthenticator(
            channel='test', level=10, clock=clock, maxQueryTime=1.0)
        self.checkIAuth(csAuth)

        for index, (result, data) in enumerate([
                (False, []), # ChanServ isn't home
                (True,  ['Random noise to ignore',
                         '1 10  me',
                         '-- End of list --']), # normal authorized situation
                (False, ['brzaops',
                         '1  5 me',
                         '-- End of list --']), # user is in but level too low
                (False, ['True',
                         '1 20 bar'
                         '2 50  foo',
                         '-- End of list --']), # user is not listed
                (False, ['foo',
                         'bar',
                         'gobbledygook']), # ChanServ is spouting garbage
                (True, ['1 10 me']), # ChanServ misses -- End of list --
                (False, ['stop bothering me']), # more garbage
                (False, ['-- End of list --']), # empty list

                # final test here must end in '-- End of list --' or
                # be invalid or we leave a DelayedCall to fail the
                # auth.
                ]):
            # prime chanserv
            chanserv.d = defer.Deferred()
            # request auth
            d = csAuth.auth(self.user)
            # wait until chanserv is msg'd
            yield chanserv.d
            # Make chanserv talk
            for line in data:
                csAuth.notice(chanserv, chanserv, line)
                # make sure messages from another nick are ignored
                csAuth.notice(self.user, self.user, '1  50  me')
            # Advance the clock past the timeout
            clock.advance(1.1)
            # We should have an auth reply now (if not the test hangs here)
            authed = yield d
            self.assertEquals(
                authed, result, msg='incorrect run %r' % index)

    @defer.inlineCallbacks
    def testMultipleChanServAuths(self):
        lock = defer.DeferredLock()
        chanserv = ChanServ('ChanServ', bot=self.bot)
        self.bot.users['chanserv'] = chanserv
        csAuth = coreauth.ChanServAuthenticator(
            channel='test', level=10, maxQueryTime=.1)
        csAuth.lock = lock
        auth2 = coreauth.ChanServAuthenticator(
            channel='test', level=10, maxQueryTime=.1)
        auth2.lock = lock
        # prime chanserv
        chanserv.d = defer.Deferred()
        # request auth
        d = csAuth.auth(self.user)
        # wait until chanserv is msg'd
        yield chanserv.d
        # prime chanserv for the second message
        chanserv.d = defer.Deferred()

        self.secondAuthed = None
        self.secondCs = None
        def callbackAuth(result):
            self.secondAuthed = result

        def callbackCs(result):
            self.secondCs = result

        # request second auth
        auth2.auth(self.user).addCallback(callbackAuth)
        chanserv.d.addCallback(callbackCs)
        # make sure the second auth is waiting on the first
        self.failIf(self.secondAuthed is not None)
        self.failIf(self.secondCs is not None)
        # release the first auth
        csAuth.notice(chanserv, chanserv, '-- End of list --')
        authed = yield d
        self.failIf(authed)
        # make sure the second one runs only now
        self.failUnless(self.secondCs)
        auth2.notice(chanserv, chanserv, '-- End of list --')
        self.failUnless(self.secondAuthed is False)
