# do not check docstrings, __init__, missing member
# pylint: disable-msg=C0111, W0232, E1101

"""Tests for corehandlers.py."""


import tempfile

from zope.interface import verify, implements

from twisted.python import log
from twisted.internet import defer, task
from twisted.trial import unittest

from mzbot import corehandlers, ibot, ebot, testutils, commands, __version__


# docstrings for these CommandHandlers are part of the test

class IgnoredException(Exception):
    """Special exception we tell the logging system to ignore."""


class FailingTestHandler(commands.CommandHandler):

    def handle_fail(self, user, source, args):
        raise IgnoredException('poof')


class SimpleTestHandler(commands.CommandHandler):

    """I am simple."""

    def __init__(self):
        commands.CommandHandler.__init__(self)

    def handle_foo(self, user, source, args):
        """I am foo."""
        source.msgTo(user, args)

    def handle_bar(self, user, source, args):
        """I am bar."""


class UndocumentedCustomTestHandler(object):

    commandset = set(('baz',))

    def handle(self, user, source, command, args):
        """This may look like documentation but it is not."""


class DocumentedCustomTestHandler(UndocumentedCustomTestHandler):

    """I am custom.

    I have multiple lines.
    """

    implements(ibot.ICommandHelpProvider)

    def getCommandDescription(self, command):
        return {
            'baz': 'I am baz.',
            }[command]


class DummyAuthenticator(object):

    """Return auth'd for the 'other' user."""

    implements(ibot.IAuthenticator)

    def __init__(self, other):
        self.other = other

    def auth(self, user):
        """Compare to our stored user."""
        return defer.succeed(user is self.other)

    def description(self):
        """human-readable description."""
        return self.__doc__


class BotHandlerTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.BotHandler()
        self.handler.ctcpVersion = 'TestBot'
        # intercept log messages
        self.logCatcher = []
        log.addObserver(self.logCatcher.append)

    def tearDown(self):
        log.removeObserver(self.logCatcher.append)

    def testInterfaces(self):
        for interface in (ibot.IUnknownHandler,
                          ibot.ISystemMessageHandler,
                          ibot.ICTCPQueryHandler,
                          ibot.ILoginNickInUseHandler):
            verify.verifyObject(interface, self.handler)

    def testVersion(self):
        handler = corehandlers.BotHandler()
        handler.ctcpQuery(self.user, self.user, 'VERSION', None)
        self.user.assertEvents(
            ('ctcpReply', 'VERSION', 'Twisted Mess %s' % __version__))

    def testCTCP(self):
        self.handler.ctcpQuery(self.user, self.user, 'VERSION', None)
        self.user.assertEvents(('ctcpReply', 'VERSION', 'TestBot'))
        self.failIf(self.logCatcher)
        self.handler.ctcpQuery(self.user, self.user, 'garbage', None)
        self.bot.assertEvents()
        self.assertEquals(len(self.logCatcher), 1)
        self.logCatcher.pop(0)
        self.handler.ctcpQuery(self.user, self.user, 'VERSION', 'garbage')
        self.assertEquals(len(self.logCatcher), 1)
        self.logCatcher.pop(0)

    def testSystemMessageNotice(self):
        self.handler.systemMessage(
            self.bot, self.bot.user.nickname, 'foo', 'message')
        self.handler.systemMessage(
            self.bot, 'bar', 'foo', 'message')
        self.handler.systemNotice(
            self.bot, self.bot.user.nickname, 'foo', 'message')
        self.handler.systemNotice(
            self.bot, 'bar', 'foo', 'message')

    def testUnknown(self):
        self.handler.unknown('123', 'this should be caught')
        self.assertEquals(len(self.logCatcher), 1)
        message = self.logCatcher.pop(0)
        self.assertEquals(
            message['message'],
            ('"received unknown numeric \'123\' '
             '(args: \'this should be caught\')"',))

    def testLoginNickInUse(self):
        self.assertEquals('mzb_', self.handler.loginNickInUse(self.bot, 'mzb'))


class TestWaitForIdentifiedHandler(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.WaitForIdentifiedHandler()
        self.loggedIn = None

    def testInterfaces(self):
        for interface in (ibot.ILoginHandler, ibot.IUserModeHandler,
                          ibot.IConfigurable):
            verify.verifyObject(interface, self.handler)

    def loggedInCallback(self, value):
        self.loggedIn = value

    def testLogin(self):
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.handler.userModeChanged(self.bot, False, 'e')
        self.handler.userModeChanged(self.bot, True, 'j')
        self.assertIdentical(self.loggedIn, None)
        self.handler.userModeChanged(self.bot, True, 'e')
        self.failUnless(self.loggedIn)

    def testMultipleLogins(self):
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.handler.disconnected(self.bot)
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.failUnlessIdentical(self.loggedIn, None)
        self.handler.userModeChanged(self.bot, True, 'e')
        self.failUnless(self.loggedIn)

    def testConfigurable(self):
        self.handler.set(u'mode', u'r')
        self.assertEquals(u'r', self.handler.get(u'mode'))


class TestWaitForCloakHandler(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.WaitForCloakHandler()
        self.loggedIn = None

    def testInterfaces(self):
        for interface in (ibot.ILoginHandler, ibot.IDisconnectHandler,
                          ibot.IHostHiddenHandler):
            verify.verifyObject(interface, self.handler)

    def loggedInCallback(self, value):
        self.loggedIn = value

    def testLogin(self):
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.assertIdentical(self.loggedIn, None)
        self.handler.hostHidden(self.bot, 'secretive/bot')
        self.failUnless(self.loggedIn)

    def testMultipleLogins(self):
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.handler.disconnected(self.bot)
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.failUnlessIdentical(self.loggedIn, None)
        self.handler.hostHidden(self.bot, 'secretive/bot')
        self.failUnless(self.loggedIn)


class SetUserModeHandlerTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.SetUserModeHandler(enabledModes='i')

    def testInterfaces(self):
        for interface in (ibot.ILoginHandler, ibot.IConfigurable):
            verify.verifyObject(interface, self.handler)

    def testLogin(self):
        d = self.handler.loggedIn(self.bot)
        def check(result):
            self.bot.assertEvents(('mode', 'i', ''))
        d.addCallback(check)
        return d

    def testConfigurable(self):
        self.handler.set(u'disabledmodes', u'oe')
        self.assertEquals(u"'oe'", self.handler.get(u'disabledmodes'))
        self.handler.set(u'enabledmodes', u'ba')
        self.assertEquals(u"'ba'", self.handler.get(u'enabledmodes'))


class TestAutoJoinHandler(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.loggedIn = False
        self.handler = corehandlers.AutoJoinHandler()

    def testInterfaces(self):
        for interface in (ibot.ILoginHandler, ibot.ICommandHandler,
                          ibot.IConfigurable):
            verify.verifyObject(interface, self.handler)

    def loggedInCallback(self, ignored):
        self.loggedIn = True

    def testConfigurable(self):
        self.assertEquals(u'', self.handler.get(u'channels'))
        self.handler.set(u'add', u'empty\\xff')
        self.handler.set(u'add', u'normal sekrit')
        self.assertIn(self.handler.get(u'channels'), (
                u'empty\\xff, normal', u'normal, empty\\xff'))
        self.handler.set(u'remove', u'empty\\xff')
        self.assertEquals(u'normal', self.handler.get(u'channels'))

    def testLogin(self):
        self.handler.set(u'add', u'empty')
        self.handler.set(u'add', u'normal sekrit')
        self.bot.user.channels = {}
        self.handler.loggedIn(self.bot).addCallback(self.loggedInCallback)
        self.failUnless(self.loggedIn)
        self.assertEquals(
            sorted(self.bot.user.channels), ['empty', 'normal'])

    def testCommands(self):
        self.handler.handle_part(self.user, self.user, u'empty')
        self.empty.assertEvents(('part', None))
        self.user.assertEvents()

        del self.bot.user.channels['empty']
        self.handler.handle_part(self.user, self.user, u'empty')
        self.user.assertEvents(
            ('msgTo', self.user, u"I'm not in that channel!"))

        self.handler.handle_join(self.user, self.user, u'empty key')
        self.bot.assertEvents(('join', 'empty', 'key'))

        self.handler.handle_part(self.user, self.user, u'')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'give me an argument or do this in channel'))

        self.bot.user.channels['empty'] = self.empty
        self.handler.handle_part(self.user, self.empty, u'')
        self.empty.assertEvents(('part', None))


class TestPrivMsgLogger(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.PrivMsgLogger()
        # intercept log messages
        self.logCatcher = []
        log.addObserver(self.logCatcher.append)

    def tearDown(self):
        log.removeObserver(self.logCatcher.append)

    def testInterfaces(self):
        for interface in (ibot.IMessageHandler,
                          ibot.INoticeHandler,
                          ibot.IActionHandler):
            verify.verifyObject(interface, self.handler)

    def testLogging(self):
        for methodName in ('message', 'notice', 'action'):
            method = getattr(self.handler, methodName)
            method(self.user, self.normal, 'foo')
            self.failIf(self.logCatcher)
            method(self.user, self.user, 'bar')
            self.assertEquals(len(self.logCatcher), 1)
            message = self.logCatcher.pop()
            self.assertEquals(message['message'], ("user 'me' said 'bar'",))


class FloodIgnoreTest(unittest.TestCase):

    def setUp(self):
        self.clock = task.Clock()
        # Necessary because we expect 0 to be "forever ago"
        self.clock.advance(9000)
        self.ignore = corehandlers.FloodIgnore(_clock=self.clock)
        testutils.setUp(self)

    def testInterfaces(self):
        verify.verifyObject(ibot.IIgnore, self.ignore)

    def testFlood(self):
        for i in range(5):
            self.failIf(self.ignore.ignored(self.other, self.other))
        self.failUnless(self.ignore.ignored(self.other, self.other))
        self.other.assertEvents(
            ('msgTo', self.other,
             u"you're talking too much, I'm ignoring you for a bit"))
        self.failUnless(self.ignore.ignored(self.other, self.other))
        self.other.assertEvents()
        # timelapse!
        self.clock.advance(self.ignore.ignoreTime + 1)
        self.failIf(self.ignore.ignored(self.other, self.other))

    def testFloodMessagesDecrease(self):
        """Does decreasing floodMessages after messages are received work?"""
        self.failIf(self.ignore.ignored(self.other, self.other))
        self.ignore.floodMessages = 1
        self.failUnless(self.ignore.ignored(self.other, self.other))

    def testFloodMessagesIncrease(self):
        """Does increasing floodMessages after messages are received work?"""
        self.ignore.floodMessages = 1
        self.failIf(self.ignore.ignored(self.other, self.other))
        self.ignore.floodMessages = 10
        self.failIf(self.ignore.ignored(self.other, self.other))

    def testConfigurable(self):
        self.ignore.set(u'floodtime', u'7.2')
        self.ignore.set(u'floodmessages', u'3')
        self.ignore.set(u'ignoretime', u'42.7')
        self.assertEquals(u'7 seconds', self.ignore.get(u'floodtime'))
        self.assertEquals(u'3 messages', self.ignore.get(u'floodmessages'))
        self.assertEquals(u'43 seconds', self.ignore.get(u'ignoretime'))


class CommandHandlerHolderBaseTest(unittest.TestCase):

    def testStubs(self):
        # Silly test, more for coverage than anything useful.
        holder = corehandlers.CommandHandlerHolderBase()
        self.assertEqual((), holder.getGlobalCommands())
        self.assertEqual((), holder.getRestrictedCommands())
        self.assertEqual((), holder.getGlobalAuths())
        self.assertEqual((), holder.getRestrictedAuths())


class AlwaysIgnore(object):

    implements(ibot.IIgnore)

    def ignored(self, user, source):
        return True


def setUpCommandHandlerHolder(self):
    self.h1 = SimpleTestHandler()
    self.h2 = UndocumentedCustomTestHandler()
    self.h3 = DocumentedCustomTestHandler()
    self.auth = DummyAuthenticator(self.other)
    self.holder = corehandlers.CommandHandlerHolder()
    self.holder.addHandler(u'h1', self.h1)
    self.holder.addAuth(u'h1', self.auth, commands=[u'foo'], channel=None)
    self.holder.addHandler(u'h2', self.h2, auth=self.auth, channel=None)
    self.holder.addHandler(u'h3', self.h3)


class TestCommandHandlerHolder(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        setUpCommandHandlerHolder(self)

    def testInterfaces(self):
        for interface in (ibot.IMessageHandler,):
            verify.verifyObject(interface, self.holder)

    def testSetUp(self):
        self.assertEquals(self.holder.globalCommands, {
                u'h1': self.h1,
                u'h3': self.h3,
                })
        self.assertEquals(
            self.holder._restrictedCommands, {None: {u'h2': self.h2}})
        self.assertEquals(self.holder.globalAuths, {})
        self.assertEquals(self.holder._restrictedAuths, {
                None: {
                    u'h1': {u'foo': self.auth},
                    u'h2': {None: self.auth},
                    },
                })
        self.assertEquals(self.holder.restrictedCommands, {None: {
                    u'h1': self.h1,
                    u'h2': self.h2,
                    u'h3': self.h3,
                    }})
        self.assertEquals(self.holder.restrictedAuths, {None: {
                    (u'h1', u'foo'): self.auth,
                    (u'h2', None): self.auth,
                    }})

    def testIgnore(self):
        self.bot.handlers[ibot.IIgnore] = [AlwaysIgnore()]
        self.holder.message(self.other, self.other, 'foo')
        self.other.assertEvents()

    def testAddAuth(self):
        self.holder.addAuth(u'h3', self.auth)
        self.assertEquals(
            self.holder.restrictedAuths[None], {(u'h1', u'foo'): self.auth,
                                                (u'h2', None): self.auth,
                                                (u'h3', None): self.auth})
        self.assertEquals(self.auth, self.holder.globalAuths[u'h3', None])
        self.holder.addAuth(u'h3', None, channel='#foo')
        self.assertEquals({}, self.holder.restrictedAuths['#foo'])

    def testAddHandler(self):
        # test "hiding" a handler
        self.holder.addHandler(u'h1', None, '#foo')
        self.assertEquals(
            self.holder._restrictedCommands['#foo'], {u'h1': None})
        self.assertEquals({u'h3': self.h3},
                          self.holder.restrictedCommands['#foo'])

    def testNoReplyNeeded(self):
        self.holder.message(self.user, self.normal, 'foo')
        self.holder.message(self.user, self.normal, 'testbot: hi there')
        self.holder.message(self.user, self.user, 'testbot: foo')
        self.holder.message(self.user, self.user, 'hi there')
        self.user.assertEvents()
        self.normal.assertEvents()

    def testOnChannelMessage(self):
        self.holder.message(self.user, self.normal, 'testbot: foo')
        self.normal.assertEvents(('msgTo', self.user, ''))
        self.holder.message(self.user, self.normal, 'testbot: foo bar')
        self.normal.assertEvents(('msgTo', self.user, 'bar'))
        self.holder.message(self.user, self.normal, 'testbot: h1 foo')
        self.normal.assertEvents(('msgTo', self.user, ''))
        self.holder.message(self.user, self.normal, 'testbot: h1 foo bar')
        self.normal.assertEvents(('msgTo', self.user, 'bar'))

    def testPrivateMessage(self):
        self.holder.message(self.other, self.other, 'foo')
        self.other.assertEvents(('msgTo', self.other, ''))
        self.holder.message(self.other, self.other, 'foo bar')
        self.other.assertEvents(('msgTo', self.other, 'bar'))
        self.holder.message(self.other, self.other, 'h1 foo')
        self.other.assertEvents(('msgTo', self.other, ''))
        self.holder.message(self.other, self.other, 'h1 foo bar')
        self.other.assertEvents(('msgTo', self.other, 'bar'))

    def testUnauthorized(self):
        self.holder.message(self.user, self.user, 'foo')
        self.user.assertEvents(
            ('msgTo', self.user, u'Sorry, you are not allowed to do that.'))

    def testMultipleMatches(self):
        self.holder.message(self.other, self.other, 'baz')
        self.other.assertEvents((
                'msgTo', self.other,
                u"multiple matches for 'baz', try one of 'h2 baz', 'h3 baz'"))

    def testPrefix(self):
        self.holder.prefix = '!'
        self.holder.message(self.user, self.normal, '!foo')
        self.holder.message(self.user, self.normal, '!foo bar')
        self.holder.message(self.user, self.normal, '! foo')
        self.normal.assertEvents(('msgTo', self.user, ''),
                                 ('msgTo', self.user, 'bar'))
        self.holder.message(self.user, self.user, '!foo')
        # (makes sure it strips prefix in privmsg)
        self.user.assertEvents(('msgTo', self.user,
                                u'Sorry, you are not allowed to do that.'))
        # check the prefix doesn't mess with normal targetted messages
        self.testNoReplyNeeded()
        self.testOnChannelMessage()

    def testFailure(self):
        self.holder.addHandler(u'fail', FailingTestHandler())
        self.holder.message(self.user, self.user, 'fail')
        self.user.assertEvents(('msgTo', self.user,
                                u'That made me traceback, '
                                'tell my owner to fix me'))
        self.flushLoggedErrors(IgnoredException)

    def testSimpleConfigurables(self):
        self.holder.set(u'prefix', u'\\xff')
        self.assertEquals(ur'\xff', self.holder.get(u'prefix'))
        self.assertEquals('\xff', self.holder.prefix)
        self.holder.set(u'prefix', u'None')
        self.assertEquals(u'None', self.holder.get(u'prefix'))
        self.assertIdentical(None, self.holder.prefix)

    def testBrokenEncoding(self):
        self.holder.message(self.user, self.user, '\xff')
        self.user.assertEvents(
            ('msgTo', self.user, u"'\\xff' is not utf-8"))


class TestHelpHandler(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        setUpCommandHandlerHolder(self)
        self.setUpHelpHandler()

    def setUpHelpHandler(self):
        self.handler = corehandlers.HelpHandler(self.holder)

    def testInterfaces(self):
        for interface in (ibot.ICommandHandler,):
            verify.verifyObject(interface, self.handler)

    def testModuleList(self):
        self.handler.handle_help(self.user, self.user, u'')
        self.user.assertEvents(
            ('msg', u'available modules:'),
            ('msg', u'\x02h1\x02: I am simple.'),
            ('msg', u'\x02h2\x02'),
            ('msg', u'\x02h3\x02: I am custom.'),
            ('msg', u"for further help, 'help modulename'"))

    def testModuleHelp(self):
        self.handler.handle_help(self.user, self.user, u'h1')
        self.user.assertEvents(
            ('msg', u'\x02bar\x02: I am bar.'),
            ('msg', u'\x02foo\x02: I am foo.'))

    def testNormalCommandHelp(self):
        self.handler.handle_help(self.user, self.user, u'foo')
        self.user.assertEvents(
            ('msg', u'\x02bar\x02: I am bar.'),
            ('msg', u'\x02foo\x02: I am foo.'))

    def testUndocumentedCommandHelp(self):
        self.handler.handle_help(self.user, self.user, u'h2')
        self.user.assertEvents(('msg', u'undocumented, commands: baz'))

    def testCustomCommandHelp(self):
        self.handler.handle_help(self.user, self.user, u'h3')
        self.user.assertEvents(('msg', u'\x02baz\x02: I am baz.'))

    def testDuplicateCommand(self):
        self.handler.handle_help(self.user, self.user, u'baz')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'that command is offered by the following modules: h2, h3. '
             u'You can ask for further help on those.'))

    def testUnknownCommand(self):
        self.handler.handle_help(self.user, self.user, u'garbage')
        self.user.assertEvents((
                'msgTo', self.user,
                u"I don't have that module or command, try 'help'"))

    def testOnChannelReplies(self):
        for query in (u'', u'h1'):
            self.handler.handle_help(self.user, self.normal, query)
        self.normal.assertEvents(
            ('msgTo', self.user, u"module list is being /msg'd"),
            ('msgTo', self.user, u"help on module 'h1' is being /msg'd"))


class TestAuthenticatedHelpHandler(TestHelpHandler):

    def setUpHelpHandler(self):
        self.handler = corehandlers.HelpHandler(self.holder, self.auth)

    def testModuleAuthentication(self):
        self.handler.handle_help(self.other, self.other, u'h2')
        self.other.assertEvents(
            ('msg', u'undocumented, commands: baz'),
            ('msg',
             u'entire module needs authentication: '
             u"Return auth'd for the 'other' user."))

    def testCommandAuthentication(self):
        self.handler.handle_help(self.other, self.other, u'h1')
        self.other.assertEvents(
            ('msg', u'\x02bar\x02: I am bar.'),
            ('msg', u'\x02foo\x02: I am foo. '
             u"(authentication: Return auth'd for the 'other' user.)"))


class CommandHandlerTest(unittest.TestCase):

    setUp = testutils.setUp

    @defer.inlineCallbacks
    def testPasswordAuthHandler(self):
        handler = corehandlers.PasswordAuthenticatorHandler({u'foo': u'bar'})
        verify.verifyObject(ibot.ICommandHandler, handler)
        verify.verifyObject(ibot.IAuthenticator, handler)
        self.failUnless(handler.description())
        self.assertIdentical(ibot.IAuthenticator(handler), handler)
        authed = yield handler.auth(self.user)
        self.failIf(authed)
        # this should not identify, since it's on a channel
        handler.handle(self.user, self.normal, u'identify', u'foo bar')
        # this should not identify, wrong password
        handler.handle(self.user, self.user, u'identify', u'foo baz')
        # this should work
        handler.handle(self.other, self.other, u'identify', u'foo bar')
        authed = yield handler.auth(self.user)
        self.failIf(authed)
        authed = yield handler.auth(self.other)
        self.failUnless(authed)


class MyConfigurable(commands.ConfigurableMixin):

    """Once again docstrings are part of the test."""

    configurables = {
        u'\N{SNOWMAN}': ('I am bar', 'spork', u'bar is %s', commands.string),
        u'unsettable': ('I am unsettable', 'spork', u'%s', None),
        u'ungettable': ('I am ungettable', 'spork', None, commands.string),
        }

    def __init__(self):
        self.spork = None

    def set_spork(self, value):
        """Set spork."""
        raise ebot.InvalidArgument(u'nope')


class AdminHandler(corehandlers.AdminHandlerBase):

    def __init__(self):
        corehandlers.AdminHandlerBase.__init__(self)
        self.handler = MyConfigurable()

    def _getConfigurable(self, user, source, handlername):
        return {u'config': self.handler}[handlername]

class AdminHandlerBaseTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = AdminHandler()

    def testStuff(self):
        self.handler.handle(self.user, self.user, u'gettables', u'config')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'\N{SNOWMAN}: I am bar, unsettable: I am unsettable'))
        self.handler.handle(self.user, self.user, u'settables', u'config')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'\N{SNOWMAN}: I am bar, spork: Set spork., '
             u'ungettable: I am ungettable'))

        self.handler.handle(self.user, self.user, u'getconfig', u'config no')
        self.user.assertEvents(('msgTo', self.user, u'no such config key'))
        self.handler.handle(self.user, self.user, u'setconfig', u'config n 1')
        self.user.assertEvents(('msgTo', self.user, u'no such config key'))

        self.handler.handle(self.user, self.user,
                            u'getconfig', u'config \N{SNOWMAN}')
        self.user.assertEvents(('msgTo', self.user, u'bar is None'))

        self.handler.handle(self.user, self.user,
                            u'setconfig', u'config \N{SNOWMAN} 42')
        self.user.assertEvents(('msgTo', self.user, u'set'))
        self.assertEqual(u'42', self.handler.handler.spork)

        self.handler.handle(self.user, self.user,
                            u'setconfig', u'config spork whatever')
        self.user.assertEvents(('msgTo', self.user, u'nope'))


class SpecialFeaturesHandlerTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.handler = corehandlers.SpecialFeaturesHandler()

    def tearDown(self):
        del self.handler

    def testStuff(self):
        self.handler.handle_say(self.user, self.user, u'slartibartfast hi')
        self.user.assertEvents(('msgTo', self.user, u'no such channel'))
        self.handler.handle_say(self.user, self.user, u'normal hi there')
        self.user.assertEvents()
        self.normal.assertEvents(('msg', u'hi there'))

        self.handler.handle_do(self.user, self.user, u'slartibartfast hi')
        self.user.assertEvents(('msgTo', self.user, u'no such channel'))
        self.handler.handle_do(self.user, self.user, u'normal hi there')
        self.user.assertEvents()
        self.normal.assertEvents(('me', u'hi there'))

        self.handler.handle_remusr(self.user, self.user, u'spork bar')
        self.user.assertEvents(('msgTo', self.user, u'not enough arguments'))
        self.handler.handle_remusr(self.user, self.user, u'slarti me bye')
        self.user.assertEvents(('msgTo', self.user, u'no such channel'))
        self.handler.handle_remusr(self.user, self.user, u'normal slarti bye')
        self.normal.assertEvents()
        self.user.assertEvents(('msgTo', self.user, u'no such user'))
        self.handler.handle_remusr(self.user, self.user, u'normal me bye')
        self.normal.assertEvents(('remove', self.user, 'bye'))
        self.user.assertEvents()


class RawLogFileTest(unittest.TestCase):

    setUp = testutils.setUp

    def testRawLog(self):
        handler = corehandlers.RawLogFile()
        with tempfile.NamedTemporaryFile() as f:
            handler.set(u'filename', f.name.decode('ascii'))
            handler.rawlog(self.bot, True, 'outbound \0')
            handler.rawlog(self.bot, False, 'inbound \xff')
            outbound, inbound = list(f)
            self.failUnless(outbound.endswith(" >>> 'outbound \\x00'\n"))
            self.failUnless(inbound.endswith(" <<< 'inbound \\xff'\n"))


class TestChanServOpHandler(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.chanserv = testutils.User('ChanServ', self.bot)
        self.bot.users['chanserv'] = self.chanserv
        self.handler = corehandlers.ChanServOpHandler(channels='empty normal')

    def testInterfaces(self):
        for interface in (ibot.IJoinHandler, ibot.ICommandHandler,
                          ibot.IConfigurable):
            verify.verifyObject(interface, self.handler)

    def testJoin(self):
        self.handler.joined(self.empty)
        self.chanserv.assertEvents(('msg', u'op empty'))
        self.handler.joined(self.voice)
        self.chanserv.assertEvents()

    def testCommands(self):
        self.handler.handle_op(self.user, self.user, u'foo')
        self.user.assertEvents(
            ('msgTo', self.user, u"doesn't work in a privmsg"))
        self.handler.handle_op(self.user, self.normal, u'')
        self.chanserv.assertEvents(('msg', u'op normal'))
        self.handler.handle_op(self.user, self.normal, u'other')
        self.chanserv.assertEvents(('msg', u'op normal other'))

        self.handler.handle_deop(self.user, self.user, u'foo')
        self.user.assertEvents(
            ('msgTo', self.user, u"doesn't work in a privmsg"))
        self.handler.handle_deop(self.user, self.normal, '')
        self.chanserv.assertEvents(('msg', u'op normal -testbot'))

        self.handler.handle_giveop(self.user, self.normal, u'foo')
        self.normal.assertEvents(('mode', True, 'o', 'foo'))

    def testConfigurable(self):
        self.assertEquals(set([u'channels']),
                          set(self.handler.gettables().keys()))
        self.assertEquals(set([u'addchannel', u'removechannel']),
                          set(self.handler.settables().keys()))
        self.handler.set(u'removechannel', u'empty')
        self.assertRaises(
            ebot.InvalidArgument, self.handler.set, u'removechannel', u'empty')
        self.assertEquals(u'normal', self.handler.get(u'channels'))
        self.handler.set(u'addchannel', u'foo\\xff')
        self.assertIn(self.handler.get(u'channels'), (
                ur'normal, foo\xff', ur'foo\xff, normal'))


class IgnoreCommandHandlerTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.clock = task.Clock()
        self.handler = corehandlers.IgnoreCommandHandler(_clock=self.clock)

    def testInterfaces(self):
        verify.verifyObject(ibot.IIgnore, self.handler)

    def testConfigurable(self):
        self.assertEquals(u'600 seconds', self.handler.get(u'ignoretime'))
        self.handler.set(u'ignoretime', u'7')
        self.assertEquals(u'7 seconds', self.handler.get(u'ignoretime'))

    def testIgnoring(self):
        self.failIf(self.handler.ignored(self.other))
        self.handler.handle_ignore(self.user, self.user, u'other')
        self.user.assertEvents(('msgTo', self.user, u'done.'))
        self.failUnless(self.handler.ignored(self.other))
        self.handler.handle_unignore(self.user, self.user, u'other')
        self.user.assertEvents(('msgTo', self.user, u'done.'))
        self.failIf(self.handler.ignored(self.other))

    def testExpiring(self):
        self.handler.handle_ignore(self.user, self.user, u'other')
        self.failUnless(self.handler.ignored(self.other))
        self.clock.advance(self.handler.ignoreTime + 1)
        self.failIf(self.handler.ignored(self.other))

    def testErrors(self):
        self.handler.handle_ignore(self.user, self.user, u'everyone')
        self.handler.handle_unignore(self.user, self.user, u'everyone')
        self.handler.handle_unignore(self.user, self.user, u'other')
        self.user.assertEvents(
            ('msgTo', self.user, u'I do not see that user!'),
            ('msgTo', self.user, u'I do not see that user!'),
            ('msgTo', self.user, u'I was not ignoring that user!'))
