# do not check docstrings and __init__
# pylint: disable-msg=C0111, W0232

"""Tests for irc.py."""


from zope.interface import verify, implements

from twisted.internet import defer, protocol, task
from twisted.trial import unittest
from twisted.test import proto_helpers

from mzbot import irc, ibot


# pairs raw, lowquoted, ctcpquoted
STRING_SUBJECTS = [
    ('Hello, this is a nice string with no complications.',
     'Hello, this is a nice string with no complications.',
     'Hello, this is a nice string with no complications.'),
    ('xargs\0might\0like\0this',
     'xargs\x100might\x100like\x100this',
     'xargs\0might\0like\0this'),
    ('embedded\rnewline\r\nFUN\n',
     'embedded\x10rnewline\x10r\x10nFUN\x10n',
     'embedded\rnewline\r\nFUN\n'),
    ('escape!\\ escape!\x10 \\\x01 \x100',
     'escape!\\ escape!\x10\x10 \\\x01 \x10\x100',
     'escape!\\\\ escape!\x10 \\\\\\a \x100'),
    ]


class QuotingTest(unittest.TestCase):

    def testQuoteSanity(self):
        for raw, lowQuoted, ctcpQuoted in STRING_SUBJECTS:
            self.failUnlessEqual(raw, irc.lowDequote(lowQuoted))
            self.failUnlessEqual(lowQuoted, irc.lowQuote(raw))
            self.failUnlessEqual(raw, irc.ctcpDequote(ctcpQuoted))
            self.failUnlessEqual(ctcpQuoted, irc.ctcpQuote(raw))


class FakeFactory(object):

    """Minimal fake factory to make IRCBot happy."""

    rawlog = None
    password = None
    nickname = 'foonick'
    username = 'foouser'
    realname = 'real name'
    cap_support = False
    saslMechanisms = ()

    def resetDelay(self):
        """Noop."""


class LoginHandler(object):

    implements(ibot.ILoginHandler)

    def __init__(self):
        self.d = defer.Deferred()

    def loggedIn(self, bot):
        self.d.callback(True)
        return defer.succeed(True)


class LoginTest(unittest.TestCase):

    @defer.inlineCallbacks
    def testLogin(self):
        loginHandler = LoginHandler()
        bot = irc.IRCBot()
        bot.factory = FakeFactory()
        bot.factory.handlers = {ibot.ILoginHandler: [loginHandler]}
        # make sure nothing is sent immediately...
        bot._outQueue.buildUpTime = 0
        # but keep the tests speedy
        bot._outQueue.lineTime = .01
        transport = proto_helpers.StringTransport()
        bot.makeConnection(transport)
        bot.dataReceived(':server 001 foonick :hi\r\n')
        yield loginHandler.d
        self.assertEquals(
            transport.value(),
            'NICK foonick\r\n'
            'USER foouser * * :real name\r\n'
            )
        bot.connectionLost()

    @defer.inlineCallbacks
    def testPassword(self):
        loginHandler = LoginHandler()
        bot = irc.IRCBot()
        bot.factory = FakeFactory()
        bot.factory.handlers = {ibot.ILoginHandler: [loginHandler]}
        bot.factory.password = 'bar'
        # make sure nothing is sent immediately...
        bot._outQueue.buildUpTime = 0
        # but keep the tests speedy
        bot._outQueue.lineTime = .01
        transport = proto_helpers.StringTransport()
        bot.makeConnection(transport)
        bot.dataReceived(':server 001 foonick :hi\r\n')
        yield loginHandler.d
        self.assertEquals(
            transport.value(),
            'PASS bar\r\n'
            'NICK foonick\r\n'
            'USER foouser * * :real name\r\n' # XXX makes no sense
            )
        bot.connectionLost()


class LoggedInTests(unittest.TestCase):

    def _clearTransport(self, result):
        self.transport.clear()

    def setUp(self):
        loginHandler = LoginHandler()
        self.bot = irc.IRCBot()
        self.bot.factory = FakeFactory()
        self.bot.factory.handlers = {ibot.ILoginHandler: [loginHandler]}
        # make sure nothing is sent immediately...
        self.bot._outQueue.buildUpTime = 0
        # but keep the tests speedy
        self.bot._outQueue.lineTime = 0
        self.transport = proto_helpers.StringTransport()
        self.bot.makeConnection(self.transport)
        self.bot.dataReceived(':server 001 foonick :hi\r\n')
        loginHandler.d.addCallback(self._clearTransport)
        return loginHandler.d

    def assertSent(self, value):
        """Check if the transport received value and clear it."""
        self.assertEquals(self.transport.value(), value)
        self.transport.clear()

    def tearDown(self):
        return self.bot.connectionLost()

    def testInterfaces(self):
        verify.verifyObject(ibot.IBot, self.bot)
        verify.verifyObject(ibot.IUser,
                            irc.User(self.bot, 'testuser'))
        verify.verifyObject(ibot.IChannel,
                            irc.Channel(self.bot, 'testchannel'))

    def testMsg(self):
        user = self.bot.getUser('somebody')
        user.msg('hi')
        self.assertSent('PRIVMSG somebody :hi\r\n')
        user.msgTo(user, 'again')
        self.assertSent('PRIVMSG somebody :again\r\n')

    def testAway(self):
        self.bot.away('gone with the wind')
        self.assertSent('AWAY :gone with the wind\r\n')
        self.bot.away()
        self.assertSent('AWAY\r\n')

    def testNick(self):
        self.bot.setNick('panda')
        self.assertSent('NICK panda\r\n')

    def testQuit(self):
        self.bot.quit('bye bye')
        self.assertSent('QUIT :bye bye\r\n')

    def testJoinEmpty(self):
        self.bot.join('mychannel')
        self.assertSent('JOIN mychannel\r\n')
        self.bot.dataReceived(
            ':foonick!foonick@somewher JOIN :mychannel\r\n'
            ':server 353 foonick = mychannel :foonick\r\n'
            ':server 366 foonick mychannel :End of /NAMES list.\r\n'
            )
        self.assertIn('mychannel', self.bot.user.channels)
        channel = self.bot.user.channels.get('mychannel')
        self.assertEquals(1, len(channel.users))
        self.assertIdentical(self.bot.user, channel.users.get('foonick'))

    def testJoinPopulated(self):
        self.bot.join('mychannel')
        self.assertSent('JOIN mychannel\r\n')
        # No topic set
        self.bot.dataReceived(
            ':foonick!foonick@somewher JOIN :mychannel\r\n'
            ':server 353 foonick = mychannel :foonick normal +voice @oper\r\n'
            ':server 366 foonick mychannel :End of /NAMES list.\r\n'
            )
        channel = self.bot.user.channels.get('mychannel')
        self.assertIdentical(None, channel.topic)
        self.assertEquals(4, len(channel.users))
        self.assertNotIn('mychannel', channel.users['normal'].chanmodes['v'])
        self.assertIn('mychannel', channel.users['voice'].chanmodes['v'])
        self.assertIn('mychannel', channel.users['oper'].chanmodes['o'])

    def testJoinTopic(self):
        self.bot.join('mychannel')
        self.assertSent('JOIN mychannel\r\n')
        # Includes dancer/hyperion's 333 numeric
        self.bot.dataReceived(
            ':foonick!foonick@somewher JOIN :mychannel\r\n'
            ':server 332 foonick mychannel :whatever\r\n'
            ':server 333 foonick mychannel marienz 1146013913\r\n'
            ':server 353 foonick = mychannel :foonick\r\n'
            ':server 366 foonick mychannel :End of /NAMES list.\r\n'
            )
        channel = self.bot.user.channels.get('mychannel')
        self.assertEquals('whatever', channel.topic)


class ServerFactory(protocol.ServerFactory):

    """String proto factory keeping track of its proto."""

    def __init__(self, lines):
        self.lines = lines

    def buildProtocol(self, addr):
        self.theProto = proto_helpers.LineSendingProtocol(self.lines)
        return self.theProto


class FactoryTest(unittest.TestCase):

    def setUp(self):
        self.factory = irc.IRCBotFactory('foobot')

    def tearDown(self):
        del self.factory

    def testInterfaces(self):
        verify.verifyObject(ibot.IBotFactory, self.factory)

    def testConnecting(self):
        loginHandler = LoginHandler()
        self.factory.addHandler(loginHandler)
        serverFactory = ServerFactory([':server 001 bot :hi\r\n'])
        unixAddress = self.mktemp()
        from twisted.internet import reactor
        port = reactor.listenUNIX(unixAddress, serverFactory)
        connector = reactor.connectUNIX(unixAddress, self.factory)
        def check(result):
            self.assertEquals(['CAP LS',
                               'NICK foobot',
                               'USER foobot * * :Twisted Mess'],
                              serverFactory.theProto.response)
            connector.disconnect()
            port.stopListening()
            # Ok, this is annoying. This runs (sort of intentionally)
            # *before* the resetDelay call in irc_RPL_WELCOME, which
            # sets continueTrying to 1. And then in the next reactor
            # iteration we start a new reconnect. So we either need to
            # prevent that resetDelay call or we need to call
            # stopTrying() *after* the disconnect gets processed.
            # Currently we do the latter...
            return task.deferLater(reactor, 0.1, self.factory.stopTrying)
        loginHandler.d.addCallback(check)
        return loginHandler.d
