# do not check docstrings and __init__
# pylint: disable-msg=C0111, W0232

"""Tests for logging."""


import os.path as osp
import datetime

from zope.interface import verify

from twisted.trial import unittest

from mzbot import logging, ibot, testutils


class RealLoggerTest(unittest.TestCase):

    def testRealLogger(self):
        path = self.mktemp()
        logger = logging.Logger()
        self.failIf(osp.exists(path))
        logger.set(u'path', path)
        self.failUnless(osp.exists(path))


class LoggingTest(unittest.TestCase):

    def setUp(self):
        self.logger = logging.Logger()
        testutils.setUp(self)

    def testInterface(self):
        verify.verifyObject(ibot.ILogger, self.logger)

    def testLogMessage(self):
        self.logger.message(
            self.user, self.normal,
            u'test \N{COPYRIGHT SIGN}'.encode('utf8'))
        self.assertEquals(
            u'test \N{COPYRIGHT SIGN}',
            self.logger.store.find(logging.Message).one().message)

    def testLogPrivMsg(self):
        self.logger.message(
            self.user, self.user,
            u'test \N{COPYRIGHT SIGN}'.encode('utf8'))
        self.assertEquals(
            u'test \N{COPYRIGHT SIGN}',
            self.logger.store.find(logging.PrivateMessage).one().message)

    def testLogModeChange(self):
        self.logger.channelModeChanged(
            self.user, self.normal, True, 'q', 'foo!*@*')
        message = self.logger.store.find(logging.ChannelModeChange).one()
        self.assertEquals('me', message.nickname)
        self.assertEquals('here', message.hostname)

    def testLogSystemModeChange(self):
        self.logger.channelModeChanged(
            'ircd', self.normal, True, 'q', 'foo!*@*')
        message = self.logger.store.find(logging.ChannelModeChange).one()
        self.assertEquals('ircd', message.nickname)
        self.assertIdentical(None, message.hostname)


class StandaloneSeenTest(unittest.TestCase):

    setUp = testutils.setUp

    def testNoLogger(self):
        handler = logging.SeenHandler()
        handler.handle(self.user, self.user, u'seen', u'somebody')
        self.user.assertEvents(('msgTo', self.user, u'cannot find logger'))

    def testInterface(self):
        handler = logging.SeenHandler()
        verify.verifyObject(ibot.ICommandHandler, handler)
        verify.verifyObject(ibot.IConfigurable, handler)

    def testGetDelta(self):
        # Hacky test to see if the non-test _getDelta works at all.
        # Only works if this finishes in less than a second.
        handler = logging.SeenHandler()
        delta = handler._getDelta(datetime.datetime.utcnow())
        self.failUnless(delta < datetime.timedelta(seconds=1))

    def testConfigurable(self):
        handler = logging.SeenHandler()
        self.assertEquals(u'', handler.get(u'channels'))
        handler.set(u'addchannel', u'test')
        self.assertEquals(u'test', handler.get(u'channels'))
        handler.set(u'removechannel', u'nottest')
        self.assertEquals(u'test', handler.get(u'channels'))
        handler.set(u'removechannel', u'test')
        self.assertEquals(u'', handler.get(u'channels'))


class TestSeenHandler(logging.SeenHandler):
    def _getDelta(self, timestamp):
        """We saw everyone 7 seconds ago."""
        return datetime.timedelta(seconds=7)


class SeenHandlerTest(unittest.TestCase):

    def setUp(self):
        self.logger = logging.Logger()
        self.handler = TestSeenHandler()
        self.handler.set(u'addchannel', u'normal')
        self.handler.set(u'addchannel', u'oper')
        testutils.setUp(self)
        self.bot.handlers[ibot.ILogger] = [self.logger]
        self.logger.message(
            self.other, self.normal,
            u'hi \N{COPYRIGHT SIGN}'.encode('utf-8'))
        self.logger.message(self.user, self.voice, 'sekrit')

    def testSeenHandler(self):
        verify.verifyObject(ibot.ICommandHandler, self.handler)
        for args, output in [
            (u'nobody', u"I haven't seen \x02nobody\x02"),
            (u'me', u'\x02me\x02 is currently on \x02normal\x02, \x02oper\x02')]:
            self.handler.handle(self.user, self.user, u'seen', args)
            self.user.assertEvents(('msgTo', self.user, output))

        self.handler.handle(self.user, self.user, u'seen', u'other')
        self.user.assertEvents(
            ('msgTo', self.user,
             u"\x0f\x02other\x02 was last seen \x1f7 seconds\x1f ago "
             u"saying '\x033hi \N{COPYRIGHT SIGN}\x0f' in \x036normal\x0f, "
             u'and not currently on any channels I know.'))
        self.normal.users['other'] = self.other
        self.other.channels['normal'] = self.normal
        self.handler.handle(self.user, self.user, u'seen', u'other')
        self.user.assertEvents(
            ('msgTo', self.user,
             u"\x0f\x02other\x02 was last seen \x1f7 seconds\x1f ago "
             u"saying '\x033hi \N{COPYRIGHT SIGN}\x0f' in "
             u'\x036normal\x0f, and is currently on \x02normal\x02.'))

    def testMultipleSeenHandlers(self):
        logger2 = logging.Logger()
        self.bot.handlers[ibot.ILogger].append(logger2)
        self.handler.handle(self.user, self.user, u'seen', u'me')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'multiple loggers registered, this is unsupported'))
