# do not check docstrings, __init__, missing member
# pylint: disable-msg=C0111, W0232, E1101

"""Tests for simplehandlers.py."""


from zope.interface import verify

from twisted.trial import unittest
from twisted.internet import reactor

from twisted.web import resource, server

from mzbot import simplehandlers, ibot, ebot, testutils, commands


class CommandHandlerTest(unittest.TestCase):

    setUp = testutils.setUp

    def testPingPongHandler(self):
        handler = simplehandlers.PingPongHandler()
        verify.verifyObject(ibot.ICommandHandler, handler)
        handler.set(u'addreply', u'ping\xff pong\xff')
        self.assertEquals(set([u'ping\xff']), handler.commandset)
        handler.handle(self.user, self.user, u'ping\xff', u'')
        handler.handle(self.other, self.other, u'ping\xff', u'pong')
        self.user.assertEvents(('msgTo', self.user, u'pong\xff'))
        self.other.assertEvents()
        # Configurable checks
        self.assertEquals(u'ping\xff: pong\xff', handler.get(u'replies'))
        self.assertRaises(
            ebot.InvalidArgument, handler.set, u'removereply', u'pong ping')
        handler.set(u'removereply', u'ping\xff pong\xff')
        self.failIf(handler.commandset)
        handler.set(u'addreply', u'pong\xff ping\xff pong\xff')
        handler.handle(self.user, self.user, u'pong\xff', u'')
        self.user.assertEvents(('msgTo', self.user, u'ping\xff pong\xff'))

    def testTopicHandler(self):
        handler = simplehandlers.TopicHandler()
        verify.verifyObject(ibot.ICommandHandler, handler)
        self.normal.setTopic('test | foo\xff')
        self.normal.events = []
        for command, args, output in [
            (u'getseparator', u'', []),
            (u'apply', u'', []),
            (u'get', u'', ['new topic: (0: test), (1: foo\xff)']),
            (u'setseparator', u'" :\\xff: "', []),
            (u'insert', u'1 bar',
             ['new topic: (0: test), (1: bar), (2: foo\xff)']),
            (u'insert', u'42 bar', [
                    'new topic: (0: test), (1: bar), (2: foo\xff), (3: bar)']),
            (u'remove', u'3', ['new topic: (0: test), (1: bar), (2: foo\xff)']),
            (u'set', u'0 foo', ['new topic: (0: foo), (1: bar), (2: foo\xff)']),
            (u'set', u'42 bar', []),
            (u'remove', u'2', ['new topic: (0: foo), (1: bar)']),
            (u'remove', u'42', []),
            (u'show', u'', ['new topic: (0: foo), (1: bar)']),
            (u'apply', u'', []),
            (u'replace', u'0 oo a\\xfft', ['new topic: (0: fa\xfft), (1: bar)']),
            (u'replace', u'\\xff \\xfa', ['new topic: (0: fa\xfat), (1: bar)']),
            (u'replace', u'a i', ['new topic: (0: fi\xfat), (1: bir)'])
            ]:
            handler.handle(self.user, self.normal, command, args)
            self.user.assertEvents(*list(('notice', out) for out in output))
        self.normal.assertEvents(
            ('msgTo', self.user, u'separator is " | "'),
            ('msgTo', self.user, u'get and modify the topic first'),
            ('msgTo', self.user, u'separator is " :\\xff: "'),
            ('msgTo', self.user, u'index error'),
            ('msgTo', self.user, u'index error'),
            ('topic', 'foo :\xff: bar'))
        # test both regrabbing and None-handling
        self.normal.setTopic(None)
        handler.handle(self.user, self.normal, u'get', u'')
        self.user.assertEvents(('notice', 'new topic: (0: )'))

    def testBackdoorHandler(self):
        handler = simplehandlers.BackdoorHandler()
        for command, args, output in [
            (u'eval', u'1/0',
             u'Miserable failure: integer division or modulo by zero'),
            (u'eval', u'0/1', u'Great success'),
            (u'rebuild', u'mzbot.missing', u'No module named missing'),
            (u'rebuild', u'mzbot.simplehandlers', u'done'),
            ]:
            handler.handle(self.user, self.user, command, args)
            self.user.assertEvents(('msgTo', self.user, output))
        handler.handle(self.user, self.user, u'eval', u'out(u"yo")')
        self.user.assertEvents(('msgTo', self.user, u'yo'),
                               ('msgTo', self.user, u'Great success'))


class RemappedHandler(commands.CommandHandler):
    def handle_foo(self, user, source, args):
        """foo: this docstring is part of the test."""
        return 42


class UndocumentedHandler(object):
    # This does not bother to actually implement ICommandHandler.
    commandset = set((u'foo',))


class CommandRemapperTest(unittest.TestCase):

    def setUp(self):
        self.handler = RemappedHandler()
        self.remapper = simplehandlers.CommandRemapper(self.handler)

    def testInterfaces(self):
        for interface in (
            ibot.ICommandHandler, ibot.ICommandHelpProvider,
            ibot.IConfigurable):
            verify.verifyObject(interface, self.remapper)

    def testConfigurable(self):
        self.assertEquals(u'', self.remapper.get(u'remaps'))
        self.assertRaises(
            ebot.InvalidArgument, self.remapper.set, u'add', u'foo bar')
        self.remapper.set(u'add', u'bar foo')
        self.remapper.set(u'add', u'ba\xff foo')
        self.assertEquals(set([u'bar', u'ba\xff']), self.remapper.commandset)
        self.assertIn(
            self.remapper.get(u'remaps'),
            (ur'bar: foo, ba\xff: foo', ur'ba\xff: foo, bar: foo'))
        # this should be a no-op
        self.remapper.set(u'remove', u'bax')
        self.remapper.set(u'remove', u'ba\xff')
        self.assertEquals(u'bar: foo', self.remapper.get(u'remaps'))

    def testHandle(self):
        self.remapper.set(u'add', u'bar foo')
        self.assertEquals(42, self.remapper.handle(None, None, u'bar', u''))

    def testCommandDescription(self):
        self.remapper.set(u'add', u'bar foo')
        self.assertEquals(
            ': this docstring is part of the test.',
            self.remapper.getCommandDescription(u'bar'))
        self.handler.commandset.remove(u'foo')
        self.assertEquals('invalid command',
                          self.remapper.getCommandDescription(u'bar'))
        handler = UndocumentedHandler()
        remapper = simplehandlers.CommandRemapper(handler)
        self.assertEquals('undocumented',
                          remapper.getCommandDescription(u'undoc'))

    def testCommandDisappears(self):
        self.remapper.set(u'add', u'bar foo')
        self.handler.commandset.remove(u'foo')
        user = testutils.User('test')
        self.remapper.handle(user, user, u'bar', u'')
        user.assertEvents(
            ('msgTo', user,
             u'remapping error: target command "foo" disappeared'))


class MetamarkResource(resource.Resource):

    isLeaf = True

    def __init__(self):
        resource.Resource.__init__(self)
        self.requests = []

    def render_POST(self, request):
        self.requests.append((request.uri, request.args))
        return 'barf'


class ShortenHandlerTest(unittest.TestCase):

    def setUp(self):
        self.resource = MetamarkResource()
        site = server.Site(self.resource)
        self.port = reactor.listenTCP(0, site)
        self.handler = simplehandlers.ShortenHandler()
        portnr = self.port.getHost().port
        self.handler.url = 'http://127.0.0.1:%s/' % (portnr,)
        testutils.setUp(self)

    def tearDown(self):
        del self.resource
        del self.handler
        self.port.stopListening()
        del self.port

    def testShorten(self):
        d = self.handler.handle_shorten(self.user, self.user, u'spork')
        d.addCallback(self._testShorten)
        return d

    def _testShorten(self, result):
        self.user.assertEvents(('msgTo', self.user, u"Shortened to barf"))
        self.assertEquals([('/', {'long_url': ['spork']})],
                          self.resource.requests)
