# do not check docstrings, __init__, missing member
# pylint: disable-msg=C0111, W0232, E1101

"""Tests for spellhandler."""


from zope.interface import verify

from twisted.trial import unittest

from mzbot import ibot, ebot, testutils

# Make sure we only swallow ImportError for enchant, not for spellhandler
try:
    import enchant
    enchant # shut up pyflakes
except ImportError: # pragma: no cover
    enchant = None
else:
    from mzbot import spellhandler


class FakeDict(object):

    def __init__(self, corrections):
        """corrections is a mapping of misspellings to corrections."""
        self.corrections = corrections
        self.pwl = set()

    def check(self, word):
        return word in self.pwl or word not in self.corrections

    def suggest(self, word):
        return self.corrections.get(word, [])

    def add_to_pwl(self, word):
        self.pwl.add(word)


class SpellHandlerTest(unittest.TestCase):

    if enchant is None: # pragma: no cover
        skip = 'Enchant not available'

    def setUp(self):
        self.handler = spellhandler.SpellHandler()
        self.handler.dict = FakeDict({
                u'teh': [u'the', u'Te'],
                u'blotz': [],
                u'self-explainatory': [],
                })
        testutils.setUp(self)

    def tearDown(self):
        del self.handler

    def test_interfaces(self):
        for iface in (ibot.ICommandHandler,
                      ibot.IConfigurable):
            verify.verifyObject(iface, self.handler)

    def test_configurable(self):
        self.assertEquals(u'5', self.handler.get(u'maxsuggestions'))
        self.handler.set(u'maxsuggestions', u'7')
        self.assertEquals(u'7', self.handler.get(u'maxsuggestions'))
        self.assertEquals(u'en', self.handler.get(u'language'))
        # Invalid language
        self.assertRaises(ebot.InvalidArgument,
                          self.handler.set, u'language', u'spork')
        self.handler.set(u'language', u'en')

    def test_spelling(self):
        self.handler.handle(self.user, self.user, u'spell', u'foo')
        self.user.assertEvents(
            ('msgTo', self.user, u'"foo" is spelled correctly.'))

        self.handler.handle(self.user, self.user, u'spell', u'teh')
        self.user.assertEvents(('msgTo', self.user, u'teh: the, Te'))

        self.handler.handle(self.user, self.user, u'spell', u'self-explainatory')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'You just made "self-explainatory" up, right?'))

        self.handler.handle(self.user, self.user, u'spell', u'foo bar')
        self.user.assertEvents(('msgTo', self.user, u'All spelled correctly.'))

        self.handler.handle(self.user, self.user, u'spell', u'foo bar teh')
        self.user.assertEvents(('msgTo', self.user, u'teh: the, Te'))

        self.handler.handle(self.user, self.user, u'spell',
                            u'foo self-explainatory')
        self.user.assertEvents(
            ('msgTo', self.user,
             u'You just made "self-explainatory" up, right?'))

        self.handler.handle(self.user, self.user, u'spell', u'teh blotz')
        self.user.assertEvents(('msgTo', self.user,
                                u'teh, blotz are misspelled, '
                                u'use "spell <word>" for suggestions'))

        self.handler.handle(self.user, self.user, u'spell', u'')
        self.user.assertEvents(('msgTo', self.user, u'not enough arguments'))

    def test_message(self):
        self.handler.message(self.user, self.user, '(sp?) foo(sp?) bar (sp?)')
        self.user.assertEvents(('msgTo', self.user, u'All spelled correctly.'))
        self.handler.message(self.user, self.user, '\xff (sp?)')
        self.user.assertEvents(('msgTo', self.user,
                                u'I can only spell-check utf-8'))
        self.handler.message(self.user, self.user, '\xff')
        self.user.assertEvents()

    def test_addword(self):
        self.handler.handle(self.user, self.user, u'spell', u'blotz')
        self.user.assertEvents(('msgTo', self.user,
                                u'You just made "blotz" up, right?'))
        self.handler.handle(self.user, self.user, u'addword', u'blotz')
        self.user.assertEvents(('msgTo', self.user, u'added'))
        self.handler.handle(self.user, self.user, u'spell', u'blotz')
        self.user.assertEvents(('msgTo', self.user,
                                u'"blotz" is spelled correctly.'))
