
from zope.interface import verify

from twisted.trial import unittest
from twisted.internet import task

from mzbot import wolf, testutils, ibot, commands


B = commands.bold


class TestWolfHandler(wolf.WolfBotHandler):

    _force_roles = True

    def _userlist(self, user, source=None):
        user.msg(u'userlist goes here')

    def randint(self, lower, upper):
        return self.nextrand


class WolfTest(unittest.TestCase):

    def setUp(self):
        testutils.setUp(self)
        self.clock = task.Clock()
        self.handler = TestWolfHandler(_clock=self.clock)
        self.handler.set(u'channel', u'normal')
        self.handler.joined(self.normal)
        # Add a bunch of users to use as players
        self.players = []
        for i in xrange(self.handler.maxplayers + 1):
            name = 'p%d' % (i,)
            player = testutils.User(name, bot=self.bot)
            self.bot.users[name] = player
            self.normal.users[name] = player
            player.channels['normal'] = self.normal
            self.players.append(player)

    def testInterfaces(self):
        verify.verifyObject(ibot.IJoinHandler, self.handler)
        verify.verifyObject(ibot.ILeftHandler, self.handler)
        verify.verifyObject(ibot.IUserLeftHandler, self.handler)
        verify.verifyObject(ibot.IMessageHandler, self.handler)

    def testGoat(self):
        self.handler.handle(self.user, self.user, u'goat', u'that-guy')
        self.user.assertEvents()
        self.normal.assertEvents(
            ('msg', B(u'that-guy has been kicked by a goat. (by me)')))

    def testLeaveChannel(self):
        self.handler.left(self.normal)
        self.handler.handle(self.user, self.user, u'goat', u'that-guy')
        self.user.assertEvents(('msgTo', self.user,
                                u'I am not on my control channel!'))
        self.normal.assertEvents()

    def testPreGame(self):
        self.handler.handle_join(self.other, self.other, u'')
        self.other.assertEvents(
            ('msgTo', self.other,
             u'You are not in the correct channel. Join normal first.'))
        self.normal.assertEvents()
        self.handler.handle_join(self.user, self.normal, u'')
        self.user.assertEvents()
        self.normal.assertEvents(('msg', B(wolf.STARTED.format(u'me'))),
                                 ('mode', True, 'v', 'me'))
        self.handler.handle_join(self.user, self.normal, u'')
        self.normal.assertEvents(('msgTo', self.user,
                                  u'you are already in the game'))
        self.handler.handle_join(self.other, self.other, u'')
        self.other.assertEvents(('msgTo', self.other,
                                 u'You are not in the correct channel. '
                                 u'Join normal first.'))

        self.handler.handle_fjoin(self.user, self.normal, u'other')
        self.normal.assertEvents(('msgTo', self.user,
                                  u'other is not in the correct channel. '
                                  u'Ask them to join normal first.'))
        self.handler.handle_fjoin(self.user, self.normal, u'p0')
        self.normal.assertEvents(
            ('msgTo', self.players[0],
             u'you have been joined to the game by {0}'.format(u'me')),
            ('mode', True, 'v', 'p0'))
        self.handler.handle_fjoin(self.user, self.normal, u'p0')
        self.normal.assertEvents(('msgTo', self.user,
                                  u'p0 is already in the game.'))

        self.handler.handle_start(self.other, self.other, u'')
        self.other.assertEvents(('msgTo', self.other,
                                 u'You are not playing.'))
        self.normal.assertEvents()

        self.clock.advance(self.handler.minstarttime)

        self.handler.handle_start(self.user, self.user, u'')
        self.normal.assertEvents()
        self.user.assertEvents(('msgTo', self.user, u'Not enough players!'))

        self.handler.handle_del(self.user, self.user, u'')
        self.normal.assertEvents(('mode', False, 'v', 'me'))
        self.user.assertEvents()

        self.handler.handle_fdel(self.user, self.user, self.user.uname)
        self.normal.assertEvents()
        self.user.assertEvents(('msgTo', self.user,
                                wolf.BAD_TARGET.format(u'removing')))
        self.handler.handle_fdel(self.user, self.user, u'p0')
        self.normal.assertEvents(('mode', False, 'v', 'p0'))
        self.user.assertEvents()

        self.handler.handle_join(self.user, self.normal, u'')
        self.user.assertEvents()
        self.normal.assertEvents(('msg', B(wolf.STARTED.format(u'me'))),
                                 ('mode', True, 'v', 'me'))
        self.handler.handle_join(self.players[0], self.players[0], u'')
        self.normal.assertEvents(('mode', True, 'v', 'p0'))
        self.handler.handle_join(self.players[1], self.players[1], u'')
        self.normal.assertEvents(('mode', True, 'v', 'p1'))

        self.handler.userLeft(self.user, self.normal)
        self.normal.assertEvents()

        self.handler.handle_fdel(self.players[0], self.players[0], u'me')
        self.players[0].assertEvents(('msgTo', self.players[0],
                                      wolf.BAD_TARGET.format(u'removing')))

        # TODO: this should probably send a message to the channel
        # so the removee knows who removed them.
        self.handler.handle_fdel(self.players[0], self.players[0], u'p1')
        self.normal.assertEvents(('mode', False, 'v', 'p1'))

        self.handler.handle_del(self.players[0], self.players[0], u'')
        self.normal.assertEvents(('mode', False, 'v', 'p0'))

        self.players[25].away = True
        self.handler.handle_ping(self.user, self.user, u'')
        self.user.assertEvents()
        self.normal.assertEvents(
            ('who',),
            ('msg', B(u'{0} says PING!!! {1}'.format(
                        u'me', u' '.join(sorted(u'p{0}'.format(i)
                                                for i in xrange(25)))))))

        self.handler.handle_ping(self.user, self.user, u'')
        self.normal.assertEvents()
        self.user.assertEvents(('msgTo', self.user, wolf.PING_TOO_FAST))

        self.handler.handle_pong(self.players[0], self.players[0], u'')
        self.players[0].assertEvents()
        self.normal.assertEvents(
            ('msgTo', self.user,
             u'{0} says PONG!!!'.format(self.players[0].uname)))

        self.handler.userLeft(self.user, self.normal)
        self.handler.handle_pong(self.players[0], self.players[0], u'')
        self.normal.assertEvents()


    def _startSmallGame(self):
        for i in xrange(4):
            self.handler.handle_join(self.players[i], self.normal, u'')
        self.normal.assertEvents(
            ('msg', B(wolf.STARTED.format(u'p0'))),
            ('mode', True, 'v', 'p0'),
            ('mode', True, 'v', 'p1'),
            ('mode', True, 'v', 'p2'),
            ('mode', True, 'v', 'p3'),
            )
        self.handler.handle_start(self.players[0], self.normal, u'')
        self.normal.assertEvents(
            ('msgTo', self.players[0], wolf.WAIT_FOR_PLAYERS.format(
                    self.handler.minstarttime + 10)))
        self.clock.advance(self.handler.minstarttime)
        self.handler.handle_start(self.players[0], self.normal, u'')
        self.normal.assertEvents(
            ('mode', True, 'm', None),
            ('msg', B(u'Game is now starting.')),
            ('msg', B(wolf.NOW_NIGHT)))

        self.assertEqual(1, len(self.handler.wolves))
        wolfie = list(self.handler.wolves)[0]
        self.assertEqual(1, len(self.handler.seers))
        seer = list(self.handler.seers)[0]
        self.assertEqual(0, len(self.handler.bullets))
        p1, p2 = list(set(self.players[:4]) - set([wolfie, seer]))
        p1.assertEvents()
        p2.assertEvents()
        seer.assertEvents(('msg', wolf.YOU_ARE_SEER),
                          ('msg', wolf.SEER_SEE),
                          ('msg', u'userlist goes here'))

        wolfie.assertEvents(('msg', wolf.YOU_ARE_WOLF),
                            ('msg', wolf.WOLF_KILL),
                            ('msg', u'userlist goes here'))
        return wolfie, seer, p1, p2

    def testUserList(self):
        wolfie, seer, p1, p2 = self._startSmallGame()
        wolf.WolfBotHandler._userlist(self.handler, seer)
        seer.assertEvents(('msgTo', seer, u'4 players: p0, p1, p2, p3'))
        wolf.WolfBotHandler._userlist(self.handler, wolfie)
        wolfie.assertEvents(('msgTo', wolfie, u'4 players: ' + {
                    self.players[0]: u'(w) p0, p1, p2, p3',
                    self.players[1]: u'p0, (w) p1, p2, p3',
                    self.players[2]: u'p0, p1, (w) p2, p3',
                    self.players[3]: u'p0, p1, p2, (w) p3',
                    }[wolfie]))

    def testSmallGameWolves(self):
        wolfie, seer, p1, p2 = self._startSmallGame()

        # Check a user who is not playing leaving is ignored.
        self.handler.userLeft(self.user, self.normal)
        self.normal.assertEvents()

        # Test the night timeout.
        self.clock.advance(120)

        for idle in [wolfie, seer, p1, p2]:
            idle.assertEvents()
        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_KILLED.format(u'a young penguin', u'pet'))),
            ('msg', B(wolf.NOW_LYNCH)))

        # XXX is there a day timeout to test?
        self.handler.handle_lynch(wolfie, self.normal, self.user.uname)
        self.normal.assertEvents(('msgTo', wolfie,
                                  wolf.BAD_TARGET.format(u'lynching')))
        self.handler.handle_lynch(wolfie, self.normal, p1.uname)
        def _assertVote(final=False):
            self.assertEqual('msg', self.normal.events[0][0])
            self.failUnless(u' votes for ' in self.normal.events[0][1])
            self.normal.events.pop(0)
            if not final:
                self.normal.assertEvents()
        _assertVote()
        self.handler.handle_lynch(seer, self.normal, p1.uname)
        _assertVote()
        self.handler.handle_lynch(p1, self.normal, wolfie.uname)
        _assertVote()
        self.handler.handle_lynch(p2, self.normal, p1.uname)
        _assertVote(True)
        self.assertEqual('msg', self.normal.events[0][0])
        self.failUnless(self.normal.events[0][1].startswith(
                B(u'Against their protests, the villagers lynch ')[:-1]))
        self.normal.events.pop(0)
        self.normal.assertEvents(('mode', False, 'v', p1.name),
                                 ('msg', B(wolf.NOW_NIGHT)))

        p1.assertEvents()
        p2.assertEvents()
        wolfie.assertEvents(('msg', wolf.WOLF_KILL),
                            ('msg', u'userlist goes here'))
        seer.assertEvents(('msg', wolf.SEER_SEE),
                          ('msg', u'userlist goes here'))

        self.handler.handle_kill(seer, seer, u'p1')
        seer.assertEvents()
        self.handler.handle_see(wolfie, wolfie, u'p1')
        wolfie.assertEvents()
        self.handler.handle_visit(wolfie, wolfie, u'p1')
        wolfie.assertEvents()

        self.handler.handle_kill(wolfie, wolfie, u'everyone')
        wolfie.assertEvents(('msgTo', wolfie,
                             wolf.BAD_TARGET.format(u'killing')))
        self.handler.handle_see(seer, seer, u'everyone')
        seer.assertEvents(('msgTo', seer,
                           wolf.BAD_TARGET.format(u'seeing')))

        self.handler.handle_see(seer, seer, wolfie.uname)
        seer.assertEvents(('msgTo', seer,
                           u'{0} is a wolf'.format(wolfie.uname)))
        self.handler.handle_see(seer, seer, wolfie.uname)
        seer.assertEvents(('msgTo', seer, wolf.CANNOT_MULTI_SEE))

        self.handler.handle_kill(wolfie, wolfie, seer.uname)
        wolfie.assertEvents(('msgTo', wolfie,
                             u'{0} is selected.'.format(seer.uname)))

        self.normal.events.pop() # -v modechange
        self.normal.events.pop() # -v modechange
        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_KILLED.format(seer.uname, u'seer'))),
            ('mode', False, 'v', seer.name),
            ('msg', B(wolf.WIN_WOLVES)),
            ('msg', B(self.handler.reveal)),
            ('mode', False, 'm', None))

    def testSmallGameVillagers(self):
        wolfie, seer, p1, p2 = self._startSmallGame()

        self.handler.handle_kill(wolfie, wolfie, seer.uname)
        wolfie.assertEvents(('msgTo', wolfie,
                             u'{0} is selected.'.format(seer.uname)))
        self.handler.handle_see(seer, seer, wolfie.uname)
        seer.assertEvents(('msgTo', seer,
                           u'{0} is a wolf'.format(wolfie.uname)))

        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_KILLED.format(seer.uname, u'seer'))),
            ('mode', False, 'v', seer.name),
            ('msg', B(wolf.NOW_LYNCH)))
        # Ignore lynch orders from a non-player.
        self.handler.handle_lynch(self.user, self.user, wolfie.uname)
        self.user.assertEvents()
        self.normal.assertEvents()
        self.handler.handle_lynch(p1, p1, wolfie.uname)
        self.handler.handle_lynch(p2, p2, wolfie.uname)
        p1.assertEvents()
        p2.assertEvents()
        self.normal.events.pop(0) # confirmation of lynch
        self.normal.events.pop(0) # confirmation of lynch
        self.failUnless(self.normal.events[0][1].startswith(
                B(u'Against their protests, the villagers lynch ')[:-1]))
        self.normal.events.pop(0)
        self.normal.events.pop() # -v modechange
        self.normal.events.pop() # -v modechange
        self.normal.assertEvents(('mode', False, 'v', wolfie.name),
                                 ('msg', B(wolf.WIN_VILLAGERS)),
                                 ('msg', B(self.handler.reveal)),
                                 ('mode', False, 'm', None))

    def testLynchTimeout(self):
        wolfie, seer, p1, p2 = self._startSmallGame()

        self.handler.handle_kill(wolfie, wolfie, seer.uname)
        wolfie.assertEvents(('msgTo', wolfie,
                             u'{0} is selected.'.format(seer.uname)))
        self.handler.handle_see(seer, seer, wolfie.uname)
        seer.assertEvents(('msgTo', seer,
                           u'{0} is a wolf'.format(wolfie.uname)))

        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_KILLED.format(seer.uname, u'seer'))),
            ('mode', False, 'v', seer.name),
            ('msg', B(wolf.NOW_LYNCH)))

        self.handler.handle_lynch(p1, self.normal, wolfie.uname)
        self.handler.message(p1, self.normal, 'lynch him')
        self.normal.events.pop(0)
        self.clock.advance(self.handler.lynch_ping_timeout)
        self.normal.assertEvents(('msg', B(wolf.PING_ALIVE.format(
                        u', '.join(sorted(u.uname for u in [wolfie, p2]))))))

        self.handler.handle_lynch(wolfie, self.normal, p1.uname)
        self.handler.message(wolfie, self.normal, 'lynch him')
        self.normal.events.pop(0)
        self.clock.advance(self.handler.lynch_del_timeout)
        # Game over, drop all the messages about that:
        self.assertEqual(7, len(self.normal.events))
        del self.normal.events[2:]
        self.normal.assertEvents(('msg', B(wolf.KILL_IDLE.format(p2.uname))),
                                 ('mode', False, 'v', p2.name))

    def testNoLynchTimeout(self):
        wolfie, seer, p1, p2 = self._startSmallGame()

        self.handler.handle_kill(wolfie, wolfie, seer.uname)
        wolfie.assertEvents(('msgTo', wolfie,
                             u'{0} is selected.'.format(seer.uname)))
        self.handler.handle_see(seer, seer, wolfie.uname)
        seer.assertEvents(('msgTo', seer,
                           u'{0} is a wolf'.format(wolfie.uname)))

        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_KILLED.format(seer.uname, u'seer'))),
            ('mode', False, 'v', seer.name),
            ('msg', B(wolf.NOW_LYNCH)))

        for player in [p1, p2, wolfie]:
            self.handler.message(player, self.normal, 'lynch him')
        self.clock.advance(self.handler.lynch_ping_timeout)
        self.normal.assertEvents()

    def testTooManyPlayers(self):
        for i in xrange(self.handler.maxplayers):
            self.handler.handle_join(self.players[i], self.players[i], u'')
        maxplayer = self.players[self.handler.maxplayers]
        self.handler.handle_join(maxplayer, maxplayer, u'')
        maxplayer.assertEvents(
            ('msgTo', maxplayer,
             u'Sorry, but there are too many players already.'))

    def _startLargerGame(self):
        for i in xrange(8):
            self.handler.handle_join(self.players[i], self.normal, u'')
        self.normal.assertEvents(
            ('msg', B(wolf.STARTED.format(u'p0'))),
            ('mode', True, 'v', 'p0'),
            ('mode', True, 'v', 'p1'),
            ('mode', True, 'v', 'p2'),
            ('mode', True, 'v', 'p3'),
            ('mode', True, 'v', 'p4'),
            ('mode', True, 'v', 'p5'),
            ('mode', True, 'v', 'p6'),
            ('mode', True, 'v', 'p7'),
            )
        self.handler.handle_fstart(self.players[0], self.normal, u'')
        self.normal.assertEvents(
            ('mode', True, 'm', None),
            ('msg', B(u'Game is now starting.')),
            ('msg', B(wolf.NOW_NIGHT)))

        self.assertEqual(2, len(self.handler.wolves))
        wolf1, wolf2 = list(self.handler.wolves)
        self.assertEqual(1, len(self.handler.seers))
        seer = list(self.handler.seers)[0]
        self.assertEqual(1, len(self.handler.bullets))
        shooter = list(self.handler.bullets)[0]
        self.assertEqual(1, len(self.handler.harlots))
        harlot = list(self.handler.harlots)[0]
        p1, p2, p3 = list(set(self.players[:8]) -
                          set([wolf1, wolf2, seer, shooter, harlot]))
        p1.assertEvents()
        p2.assertEvents()
        p3.assertEvents()
        seer.assertEvents(('msg', wolf.YOU_ARE_SEER),
                          ('msg', wolf.SEER_SEE),
                          ('msg', u'userlist goes here'))

        harlot.assertEvents(('msg', wolf.YOU_ARE_HARLOT),
                            ('msg', wolf.HARLOT_VISIT),
                            ('msg', u'userlist goes here'))

        for wolfie in wolf1, wolf2:
            wolfie.assertEvents(('msg', wolf.YOU_ARE_WOLF),
                                ('msg', wolf.WOLF_KILL),
                                ('msg', u'userlist goes here'))
        return wolf1, wolf2, seer, shooter, harlot, p1, p2, p3

    def testLargerGame(self):
        (wolf1, wolf2, seer, shooter, harlot,
         p1, p2, p3) = self._startLargerGame()

        self.handler.handle_see(seer, seer, harlot.uname)
        seer.assertEvents(('msgTo', seer,
                           u'{0} is a harlot'.format(harlot.uname)))

        for wolfie in wolf1, wolf2:
            self.handler.handle_kill(wolfie, wolfie, harlot.uname)
            wolfie.assertEvents(('msgTo', wolfie,
                                 u'{0} is selected.'.format(harlot.uname)))

        self.handler.message(wolf1, wolf1, 'kill')
        self.handler.message(wolf1, self.normal, 'live')
        wolf2.assertEvents(('msg', u'{0} says: kill'.format(wolf1.uname)))

        # Harlot goes last, to test the game waits for them.
        self.normal.assertEvents()
        self.handler.handle_visit(wolf1, wolf1, harlot.uname)
        wolf1.assertEvents()
        self.handler.handle_visit(harlot, harlot, harlot.uname)
        harlot.assertEvents(('msgTo', harlot, wolf.HARLOT_VISIT_SELF))
        self.handler.handle_visit(harlot, harlot, u'me')
        harlot.assertEvents(('msgTo', harlot,
                             wolf.BAD_TARGET.format(u'visiting')))
        self.handler.handle_visit(harlot, harlot, p1.uname)
        harlot.assertEvents(('msgTo', harlot,
                             wolf.HARLOT_VISITS_TARGET.format(p1.uname)))
        p1.assertEvents(('msg', wolf.HARLOT_VISITS_YOU.format(harlot.uname)))

        self.handler.handle_visit(harlot, harlot, p2.uname)
        harlot.assertEvents(('msgTo', harlot, wolf.HARLOT_VISITS_ALREADY))

        self.normal.assertEvents(
            ('msg', B(wolf.WOLVES_MISSED_HARLOT)),
            ('msg', B(wolf.WOLVES_KILLED.format(u'a young penguin', u'pet'))),
            ('msg', B(wolf.NOW_LYNCH)))

        shooter.assertEvents(
            ('msg',
             wolf.YOU_ARE_SHOOTER.format(self.handler.bullets[shooter])))

        self.handler.handle_shoot(seer, seer, shooter.uname)
        seer.assertEvents()

        self.handler.handle_lynch(p3, p3, shooter.uname)
        self.normal.assertEvents(('msg', B(u'{0} votes for {1}'.format(
                        p3.uname, shooter.uname))))

        self.handler.handle_shoot(shooter, shooter, u'me')
        shooter.assertEvents(('msgTo', shooter,
                              wolf.BAD_TARGET.format(u'shooting')))
        self.handler.nextrand = 70
        self.handler.handle_shoot(shooter, shooter, p3.uname)
        shooter.assertEvents()
        self.normal.assertEvents(
            ('msg', B(wolf.SHOOTER_SHOT.format(shooter.uname, p3.uname))),
            ('msg', B(wolf.SHOOTER_SHOT_VILLAGER.format(p3.uname,
                                                        shooter.uname,))))

        self.handler.handle_stats(shooter, shooter, u'')
        # The real _userlist is smarter than this (uses msgTo)
        shooter.assertEvents(
            ('msg', u'userlist goes here'),
            ('msgTo', shooter, wolf.STATS.format(
                    wolves=2, seers=1, harlots=1)))

        self.handler.handle_votes(shooter, shooter, u'')
        shooter.assertEvents(('msgTo', shooter, u'Votes:'),
                             ('msgTo', shooter, wolf.VOTES.format(8, 4, 7)))

        self.handler.handle_lynch(shooter, shooter, wolf1.uname)
        shooter.assertEvents()
        self.normal.assertEvents(('msg', B(u'{0} votes for {1}'.format(
                        shooter.uname, wolf1.uname))))

        self.handler.handle_votes(shooter, shooter, u'')
        shooter.assertEvents(
            ('msgTo', shooter, u'Votes: {0}: 1 ({1})'.format(wolf1.uname,
                                                             shooter.uname)),
            ('msgTo', shooter, wolf.VOTES.format(8, 4, 7)))
