# do not check docstrings
# pylint: disable-msg=C0111

"""Some things to help write unittests."""


from zope.interface import implements

from twisted.internet import defer
from twisted.trial import unittest

from mzbot import ibot


class _EventHelperMixin(object):

    def assertEvents(self, *events):
        """Verify self.events equals the passed in value and reset it."""
        failures = []
        for i, (expected, got) in enumerate(zip(events, self.events)):
            if expected != got:
                failures.append('line %s: %s != %s' % (i + 1, expected, got))
        for line in events[len(self.events):]:
            failures.append('missing %s' % (line,))
        for line in self.events[len(events):]:
            failures.append('extra %s' % (line,))
        if failures:
            raise unittest.FailTest('\n'.join(failures))
        del self.events[:]


class Bot(_EventHelperMixin):

    """An IBot that appends everything done to it to self.events."""

    implements(ibot.IBot)

    def __init__(self, nickname='testbot'):
        self.user = User(nickname, self)
        self.users = {nickname: self.user}
        self.events = []
        self.handlers = {}

    def getUser(self, nick):
        try:
            return self.users[nick.lower()]
        except KeyError:
            return User(nick, bot=self)

    def caseFolder(self, msg):
        return msg.lower()

    def quit(self, reason=None):
        self.events.append(('quit', reason))

    def join(self, channel, key=None):
        self.events.append(('join', channel, key))
        self.user.channels[channel.lower()] = Channel(channel, bot=self)

    def away(self, message=''):
        self.events.append(('away', message))

    def mode(self, enabledModes='', disabledModes=''):
        self.events.append(('mode', enabledModes, disabledModes))

    def setNick(self, nick):
        self.events.append(('setNick', nick))
        del self.users[self.user.nickname]
        self.users[nick] = self.user
        self.user.name = nick


# XXX Yes, typechecks suck, but I really want to avoid the
# implicit conversion between unicode and str below...

class Source(_EventHelperMixin):

    """An ISource that appends everything done to it to self.events."""

    implements(ibot.ISource)

    def __init__(self, name, bot=None):
        self.name = name
        self.data = {}
        self.bot = bot
        self.events = []

    def notice(self, message):
        self.events.append(('notice', message))

    def msg(self, message):
        if isinstance(message, str): # pragma: no cover
            raise TypeError('Pass unicode, not str')
        self.events.append(('msg', message))

    def msgTo(self, target, message):
        if isinstance(message, str): # pragma: no cover
            raise TypeError('Pass unicode, not str')
        self.events.append(('msgTo', ibot.IUser(target), message))

    def me(self, message):
        if isinstance(message, str): # pragma: no cover
            raise TypeError('Pass unicode, not str')
        self.events.append(('me', message))

    def ctcpReply(self, tag, data=None):
        self.events.append(('ctcpReply', tag, data))

    def ctcpQuery(self, tag, data=None):
        self.events.append(('ctcpQuery', tag, data))

    @property
    def uname(self):
        return self.name.decode('ascii')


class Channel(Source):

    """An IChannel that appends everything done to it to self.events."""

    implements(ibot.IChannel)

    def __init__(self, name, bot=None):
        Source.__init__(self, name, bot)
        self.users = {}
        self.modes = {'b': {}}
        self.topic = None

    def setTopic(self, topic):
        self.topic = topic
        self.events.append(('topic', topic))

    def mode(self, mode, enable, arg=None):
        self.events.append(('mode', mode, enable, arg))

    def kick(self, user, reason=None):
        self.events.append(('kick', user, reason))

    def remove(self, user, reason=None):
        self.events.append(('remove', user, reason))

    def part(self, reason=None):
        self.events.append(('part', reason))

    def who(self):
        self.events.append(('who',))
        return defer.succeed(None)


class User(Source):

    """An IUser that appends everything done to it to self.events."""

    implements(ibot.IUser)

    def __init__(self, name, bot=None, realname='Test User',
                 username='testuser', hostname='testhost'):
        Source.__init__(self, name, bot)
        self.server = 'someserver'
        self.realname = realname
        self.username = username
        self.hostname = hostname
        self.channels = {}
        self.chanmodes = {'o': {}, 'v': {}}
        self.identified = None
        self.away = None
        self.awayMessage = None

    def who(self):
        return defer.succeed(self)

    def hostmask(self):
        return defer.succeed((self.nickname, self.username, self.hostname))

    @property
    def nickname(self):
        return self.name


def setUp(test):
    """A method to call from setUp that sets up users and channels.

    self.bot is a bot.
    self.user is user 'me'
    self.other is identified user 'other'
    self.empty is an empty channel
    self.normal is a channel with 'me' in it
    self.voice is a channel with 'me' in it and voiced
    self.oper is a channel with 'me' in it as oper
    """
    test.bot = Bot()
    test.user = User('me', bot=test.bot, username='alsome', hostname='here')
    test.other = User(
        'other', bot=test.bot, username='other', hostname='here')
    test.other.identified = True
    test.bot.users['me'] = test.user
    test.bot.users['other'] = test.other

    test.empty = Channel('empty', bot=test.bot)
    test.bot.user.channels['empty'] = test.empty
    for name in ['normal', 'voice', 'oper']:
        chan = Channel(name, bot=test.bot)
        setattr(test, name, chan)
        test.user.channels[name] = chan
        test.bot.user.channels[name] = chan
        chan.users['me'] = test.user
    test.user.chanmodes['v']['voice'] = test.voice
    test.user.chanmodes['o']['oper'] = test.oper
